/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "Simikon8052.hh"
#include "LeakTracker.hh"

// MISC_INCLUDES
#include <iostream.h>

/*===========================================================================*/
/* INSTRUCTIONS IMPLEMENTATION                                               */
/*===========================================================================*/

void Simikon8052::nop( Simikon8052 *c )
{
  c->PC++;
}

void Simikon8052::ajmp( Simikon8052 *c )
{
  c->PC = c->ajmp_PC( c );
}

uint Simikon8052::ajmp_PC( Simikon8052 *c )
{
  uint PC = c->PC + 2;

  PC &= 0xF800;
  PC |= (( c->activeCode[ c->PC ].opCode & 0xE0 ) << 3 );
  PC |= c->activeCode[ c->PC ].b1;

  return PC;
}

void Simikon8052::ljmp( Simikon8052 *c )
{
  c->PC = c->ljmp_PC( c );
}

uint Simikon8052::ljmp_PC( Simikon8052 *c )
{
  return ( ( c->activeCode[ c->PC ].b1 << 8 ) | c->activeCode[ c->PC ].b2 );
}

void Simikon8052::rr( Simikon8052 *c )
{
  uint RMBit = ( c->iram[ ACC ] & 0x0001 ) << 7;

  c->iram[ ACC ] >>= 1;
  c->iram[ ACC ] |= RMBit;

  c->PC++;
  c->updateIRAM( ACC, c->iram[ ACC ] );
}

void Simikon8052::inc_A( Simikon8052 *c )
{
  if ( ++c->iram[ ACC ] > 0xFF ) c->iram[ ACC ] = 0;

  c->PC++;
  c->updateIRAM( ACC, c->iram[ ACC ]  );
}

void Simikon8052::inc_IRAM( Simikon8052 *c )
{
  uint address = c->activeCode[ c->PC ].b1;

  if ( ++c->iram[ address ] > 0x00FF ) c->iram[ address ] = 0;

  c->PC += 2;
  c->updateIRAM( address, c->iram[ address ] );
}

void Simikon8052::inc_INDR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x06;

  uint address = c->GetRXvalue( r );
  uint value = c->iram[ address ];
 
  if ( ++value > 0x00FF ) value = 0;

  c->PC++;
  c->updateIRAM( address, value );
}

void Simikon8052::inc_DIRR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x08;

  uint address = c->GetRXaddress( r );
  uint value = c->GetRXvalue( r );

  if ( ++value > 0x00FF ) value = 0;

  c->PC++;
  c->updateIRAM( address, value );
}

void Simikon8052::jbc( Simikon8052 *c )
{
  uint bitAddress = c->activeCode[ c->PC ].b1;

  c->PC = c->jbc_PC( c );

  if ( c->getBit( bitAddress ) )
    c->setBit( bitAddress, false );
}

uint Simikon8052::jbc_PC( Simikon8052 *c )
{
  uint PC = c->PC + 3;

  if ( c->getBit( c->activeCode[ c->PC ].b1 ) ) {
    PC += ( c->activeCode[ c->PC ].b2 < 128 ) ? 
      c->activeCode[ c->PC ].b2 : - ( 0xFF - c->activeCode[ c->PC ].b2 + 1 );
  }
  
  return PC;
}

void Simikon8052::acall( Simikon8052 *c )
{
  uint nextInstruction = c->PC + 2;

  c->Push( nextInstruction & 0x00FF ); // LSB
  c->Push( ( nextInstruction & 0xFF00 ) >> 8 ); // MSB

  c->PC = c->acall_PC( c );
}

uint Simikon8052::acall_PC( Simikon8052 *c )
{
  /* 
   * The destination address is obtained by successively concatenating
   * the five high-order bits of the incremented PC, op code bits 7-5,
   * and the second byte of the instruction.
   */

  uint PC = c->PC + 2;

  PC &= 0xF800;
  PC |= (( c->activeCode[ c->PC ].opCode & 0xE0 ) << 3);
  PC |= c->activeCode[ c->PC ].b1;

  return PC;
}

void Simikon8052::lcall( Simikon8052 *c )
{
  uint nextInstruction = c->PC + 3;

  c->Push( nextInstruction & 0x00FF ); // LSB
  c->Push( ( nextInstruction & 0xFF00 ) >> 8 ); // MSB

  c->PC = c->lcall_PC( c );
}

uint Simikon8052::lcall_PC( Simikon8052 *c )
{
  return ( ( c->activeCode[ c->PC ].b1 << 8 ) | c->activeCode[ c->PC ].b2 );
}

void Simikon8052::rrc( Simikon8052 *c )
{
  bool RMBit = ( c->iram[ ACC ] & 0x0001 );
 
  c->iram[ ACC ] >>= 1;
  c->iram[ ACC ] |= ( c->getBit( C ) << 7 );
  c->setBit( C, RMBit );

  c->PC++;
  c->updateIRAM( ACC, c->iram[ ACC ] );
}

void Simikon8052::dec_A( Simikon8052 *c )
{
  if ( c->iram[ ACC ]-- == 0 ) c->iram[ ACC ] = 0x00FF;

  c->PC += c->activeCode[ c->PC ].bytes;
  c->updateIRAM( ACC, c->iram[ ACC ]  );
}

void Simikon8052::dec_IRAM( Simikon8052 *c )
{
  uint address = c->activeCode[ c->PC ].b1;

  if ( c->iram[ address ]-- == 0 ) c->iram[ address ] = 0x00FF;
  
  c->PC += c->activeCode[ c->PC ].bytes;
  c->updateIRAM( address, c->iram[ address ] );
}

void Simikon8052::dec_INDR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x16;
  uint address = c->GetRXvalue( r );
  uint value = c->iram[ address ];

  if ( value-- == 0 ) value = 0x00FF;

  c->PC += c->activeCode[ c->PC ].bytes;
  c->updateIRAM( address, value );
}

void Simikon8052::dec_DIRR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x18;
  uint address = c->GetRXaddress( r );
  uint value = c->GetRXvalue( r );

  if ( value-- == 0 ) value = 0x00FF;

  c->PC += c->activeCode[ c->PC ].bytes;
  c->updateIRAM( address, value );
}

void Simikon8052::jb( Simikon8052 *c )
{
  c->PC = c->jb_PC( c );
}

uint
Simikon8052::jb_PC( Simikon8052 *c )
{
  uint PC = c->PC + 3;

  if ( c->getBit( c->activeCode[ c->PC ].b1 ) ) {
    PC += ( c->activeCode[ c->PC ].b2 < 128 ) ? 
      c->activeCode[ c->PC ].b2 : - ( 0xFF - c->activeCode[ c->PC ].b2 + 1 );
  }

  return PC;
}

void Simikon8052::ret( Simikon8052 *c )
{
  c->PC = c->ret_PC( c );

  c->Pop();
  c->Pop();
}

uint Simikon8052::ret_PC( Simikon8052 *c )
{
  if ( c->stack->count() >= 2 ) {
    uint msb = *c->stack->pop();
    uint lsb = *c->stack->pop();

    c->stack->push( new uint( lsb ) );
    c->stack->push( new uint( msb ) );
  
    return ( ( msb << 8 ) | lsb );
  }
  
  return 0;
}

void Simikon8052::rl( Simikon8052 *c )
{
  uint LMBit = ( c->iram[ ACC ] & 0x0080 ) >> 7;

  c->iram[ ACC ] <<= 1;
  c->iram[ ACC ] |= LMBit;

  c->PC += c->activeCode[ c->PC ].bytes;
  c->updateIRAM( ACC, c->iram[ ACC ] );
}

// FIXME: overflow (OV) flag ne proracunava tocno!!!

#define SET_ADD_A_FLAGS() do { \
  bool carry = ( acc > 0xFF ) ? 1 : 0; \
  bool auxcarry = ( ( acc > 0x0F ) && !carry ) ? 1 : 0; \
  bool overflow = 0; \
  c->setBit( C, carry ); \
  c->setBit( AC, auxcarry ); \
  c->setBit( OV, overflow ); \
} while ( 0 )

void Simikon8052::add_DATA( Simikon8052 *c )
{
  uint acc = c->iram[ ACC ] + c->activeCode[ c->PC ].b1;

  SET_ADD_A_FLAGS();

  c->PC += c->activeCode[ c->PC ].bytes;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::add_IRAM( Simikon8052 *c )
{
  uint acc = c->iram[ ACC ] + c->iram[ c->activeCode[ c->PC ].b1 ];

  SET_ADD_A_FLAGS();

  c->PC += c->activeCode[ c->PC ].bytes;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::add_INDR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x26;
  uint address = c->GetRXvalue( r );
 
  uint acc = c->iram[ ACC ] + c->iram[ address ];

  SET_ADD_A_FLAGS();

  c->PC++;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::add_DIRR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x28;
  uint acc = c->iram[ ACC ] + c->GetRXvalue( r );

  SET_ADD_A_FLAGS();

  c->PC++;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::jnb( Simikon8052 *c )
{  
  c->PC = c->jnb_PC( c );
}

uint
Simikon8052::jnb_PC( Simikon8052 *c )
{
  uint PC = c->PC + 3;

  if ( !c->getBit( c->activeCode[ c->PC ].b1 ) ) {
    PC += ( c->activeCode[ c->PC ].b2 < 128 ) ? 
      c->activeCode[ c->PC ].b2 : - ( 0xFF - c->activeCode[ c->PC ].b2 + 1);
  }

  return PC;
}

void Simikon8052::reti( Simikon8052 *c )
{
  c->PC = c->ret_PC( c );

  c->Pop();
  c->Pop();

  if ( c->highPriorityInterruptExecuting )
    c->highPriorityInterruptExecuting = false;
  else
    c->lowPriorityInterruptExecuting = false;
}

void Simikon8052::rlc( Simikon8052 *c )
{
  bool LMBit = ( c->iram[ ACC ] & 0x0080 );
  
  c->iram[ ACC ] <<= 1;
  c->iram[ ACC ] |= c->getBit( C );
  c->setBit( C, LMBit );

  c->PC++;
  c->updateIRAM( ACC, c->iram[ ACC ] );
}

void Simikon8052::addc_DATA( Simikon8052 *c )
{
  uint acc = c->iram[ ACC ] +
           c->activeCode[ c->PC ].b1 + c->getBit( C );

  SET_ADD_A_FLAGS();

  c->PC += 2;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::addc_IRAM( Simikon8052 *c )
{
  uint acc = c->iram[ ACC ] + 
             c->iram[ c->activeCode[ c->PC ].b1 ] + c->getBit( C );

  SET_ADD_A_FLAGS();

  c->PC += 2;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::addc_INDR( Simikon8052 *c )
{
  uint r   = c->activeCode[ c->PC ].opCode - 0x36;
  uint acc = c->iram[ ACC ] + 
             c->iram[ c->GetRXvalue( r ) ] + 
             c->getBit( C );

  SET_ADD_A_FLAGS();

  c->PC++;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::addc_DIRR( Simikon8052 *c )
{
  uint r   = c->activeCode[ c->PC ].opCode - 0x38;
  uint acc = c->iram[ ACC ] +
             c->GetRXvalue( r ) +
             c->getBit( C );

  SET_ADD_A_FLAGS();

  c->PC++;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::jc( Simikon8052 *c )
{
  c->PC = c->jc_PC( c );
}

uint Simikon8052::jc_PC( Simikon8052 *c )
{
  uint PC = c->PC + 2;

  if ( c->getBit( C ) ) {
    PC += ( c->activeCode[ c->PC ].b1 < 128 ) ?
      c->activeCode[ c->PC ].b1 : -( 0xFF - c->activeCode[ c->PC ].b1 + 1 );
  }

  return PC;
}

void Simikon8052::orl_IRAM_A( Simikon8052 *c )
{
  uint address = c->activeCode[ c->PC ].b1;
  uint value   = c->iram[ address ] | c->iram[ ACC ];

  c->PC += 2;
  c->updateIRAM( address, value );
}

void Simikon8052::orl_IRAM_DATA( Simikon8052 *c )
{
  uint address = c->activeCode[ c->PC ].b1;
  uint value   = c->iram[ address ] | c->activeCode[ c->PC ].b2;

  c->PC += 3;
  c->updateIRAM( address, value );
}

void Simikon8052::orl_DATA( Simikon8052 *c )
{
  uint acc = c->iram[ ACC ] | c->activeCode[ c->PC ].b1;

  c->PC += 2;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::orl_IRAM( Simikon8052 *c )
{
  uint acc = c->iram[ ACC ] | c->iram[ c->activeCode[ c->PC ].b1 ];

  c->PC += 2;
  c->updateIRAM( ACC, acc );
}

void Simikon8052::orl_INDR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x46;
  uint address = c->GetRXvalue( r );

  c->PC++;
  c->updateIRAM( ACC, c->iram[ ACC ] | c->iram[ address ] );
}

void Simikon8052::orl_DIRR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x48;
  uint value = c->GetRXvalue( r );

  c->PC++;
  c->updateIRAM( ACC, c->iram[ ACC ] | value );
}

void Simikon8052::jnc( Simikon8052 *c )
{
  c->PC = c->jnc_PC( c );
}

uint Simikon8052::jnc_PC( Simikon8052 *c )
{
  uint PC = c->PC + 2;

  if ( !c->getBit( C ) ) {
    PC += ( c->activeCode[ c->PC ].b1 < 128 ) ?
      c->activeCode[ c->PC ].b1 : -( 0xFF - c->activeCode[ c->PC ].b1 + 1 );
  }

  return PC;
}

void Simikon8052::anl_IRAM_A( Simikon8052 * )
{
  cout << "Simikon8052::anl_IRAM_A(): not implemented!\n";
}

void Simikon8052::anl_IRAM_DATA( Simikon8052 * )
{
  cout << "Simikon8052::anl_IRAM_DATA(): not implemented!\n";
}

void Simikon8052::anl_DATA( Simikon8052 *c )
{
  c->iram[ ACC ] &= c->activeCode[ c->PC ].b1;

  c->PC += 2;
  c->updateIRAM( ACC, c->iram[ ACC ] );
}

void Simikon8052::anl_IRAM( Simikon8052 * )
{
  cout << "Simikon8052::anl_IRAM(): not implemented!\n";
}

void Simikon8052::anl_INDR( Simikon8052 * )
{
  cout << "Simikon8052::anl_INDR(): not implemented!\n";
}

void Simikon8052::anl_DIRR( Simikon8052 * )
{
  cout << "Simikon8052::anl_DIRR(): not implemented!\n";
}

void Simikon8052::jz( Simikon8052 *c )
{
  c->PC = c->jz_PC( c );
}

uint Simikon8052::jz_PC( Simikon8052 *c )
{
  uint PC = c->PC + 2;

  if ( c->iram[ ACC ] == 0 ) {
    PC += ( c->activeCode[ c->PC ].b1 < 128 ) ?
      c->activeCode[ c->PC ].b1 : -( 0xFF - c->activeCode[ c->PC ].b1 + 1 );
  
  }
  
  return PC;
}

void Simikon8052::xrl_IRAM_A( Simikon8052 *c )
{
  uint address = c->activeCode[ c->PC ].b1;
  uint value   = c->iram[ address ] ^ c->iram[ ACC ];

  c->PC += 2;
  c->updateIRAM( address, value );
}

void Simikon8052::xrl_IRAM_DATA( Simikon8052 *c )
{
  uint address = c->activeCode[ c->PC ].b1;
  uint value   = c->iram[ address ] ^ c->activeCode[ c->PC ].b2;

  c->PC += 3;
  c->updateIRAM( address, value );
}

void Simikon8052::xrl_DATA( Simikon8052 *c )
{
  uint value = c->iram[ ACC ] ^ c->activeCode[ c->PC ].b1;

  c->PC += 2;
  c->updateIRAM( ACC, value );
}

void Simikon8052::xrl_IRAM( Simikon8052 *c )
{
  uint value = c->iram[ ACC ] ^ c->iram[ c->activeCode[ c->PC ].b1 ];

  c->PC += 2;
  c->updateIRAM( ACC, value );
}

void Simikon8052::xrl_INDR( Simikon8052 *c )
{
  uint r       = c->activeCode[ c->PC ].opCode - 0x66;
  uint address = c->GetRXvalue( r );
  uint value   = c->iram[ ACC ] ^ c->iram[ address ];

  c->PC++;
  c->updateIRAM( ACC, value );
}

void Simikon8052::xrl_DIRR( Simikon8052 *c )
{
  uint r       = c->activeCode[ c->PC ].opCode - 0x68;
  uint value   = c->iram[ ACC ] ^ c->GetRXvalue( r );

  c->PC++;
  c->updateIRAM( ACC, value );
}

void Simikon8052::jnz( Simikon8052 *c )
{
  c->PC = c->jnz_PC( c );
}

uint Simikon8052::jnz_PC( Simikon8052 *c )
{
  uint PC = c->PC + 2;

  if ( c->iram[ ACC ] ) { // if any bit of the accumulator is set
    PC += ( c->activeCode[ c->PC ].b1 < 128 ) ?
      c->activeCode[ c->PC ].b1 : -( 0xFF - c->activeCode[ c->PC ].b1 + 1 );
  }

  return PC;
}

void Simikon8052::orl_C_BADDR( Simikon8052 * )
{
  cout << "Simikon8052::orl_C_BADDR(): not implemented!\n";
}

void Simikon8052::jmp( Simikon8052 * )
{
  cout << "Simikon8052::jmp(): not implemented!\n";
}

void Simikon8052::mov_A_DATA( Simikon8052 *c )
{
  c->PC += 2;
  c->updateIRAM( ACC, c->activeCode[ c->PC - 2 ].b1 );
}

void Simikon8052::mov_IRAM_DATA( Simikon8052 *c )
{
  c->PC += 3;
  c->updateIRAM( c->activeCode[ c->PC - 3 ].b1, c->activeCode[ c->PC - 3 ].b2 );
}

void Simikon8052::mov_INDR_DATA( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x76;
  uint address = c->GetRXvalue( r );

  c->PC += 2;
  c->updateIRAM( address, c->activeCode[ c->PC - 2 ].b1 );
}

void Simikon8052::mov_DIRR_DATA( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x78;
  uint address = c->GetRXaddress( r );

  c->PC += 2;
  c->updateIRAM( address, c->activeCode[ c->PC - 2 ].b1 );
}

void Simikon8052::sjmp( Simikon8052 *c )
{
  c->PC = c->sjmp_PC( c );
}

uint Simikon8052::sjmp_PC( Simikon8052 *c )
{
  uint PC = c->PC + 2;

  PC += ( c->activeCode[ c->PC ].b1 < 128 ) ? 
    c->activeCode[ c->PC ].b1 : - ( 0xFF - c->activeCode[ c->PC ].b1 + 1 );

  return PC;
}

void Simikon8052::anl_C_BADDR( Simikon8052 * )
{
  cout << "Simikon8052::anl_C_BADDR(): not implemented!\n";
}

void Simikon8052::movc_PC( Simikon8052 * )
{
  cout << "Simikon8052::movc_PC(): not implemented!\n";
}

void Simikon8052::div( Simikon8052 *c )
{
  c->PC++;

  c->setBit( C, false );

  if ( c->iram[ B ] ) {
    uint quotient  = c->iram[ ACC ] / c->iram[ B ];
    uint remainder = c->iram[ ACC ] % c->iram[ B ];

    c->setBit( OV, false );

    c->updateIRAM( ACC, quotient );
    c->updateIRAM( B, remainder );
  } else {
    c->setBit( OV, true );
  }
}

void Simikon8052::mov_IRAM_IRAM( Simikon8052 *c )
{
  c->PC += 3;
  c->updateIRAM( c->activeCode[ c->PC - 3 ].b2, 
                 c->iram[ c->activeCode[ c->PC - 3 ].b1 ] );
}

void Simikon8052::mov_IRAM_INDR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x86;
  uint dataAddress = c->GetRXvalue( r );

  c->PC += 2;
  c->updateIRAM( c->activeCode[ c->PC - 2 ].b1, c->iram[ dataAddress ] );
}

void Simikon8052::mov_IRAM_DIRR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0x88;
  uint data = c->GetRXvalue( r );

  c->PC += 2;
  c->updateIRAM( c->activeCode[ c->PC - 2 ].b1, data );
}

void Simikon8052::mov_DPTR_DATA( Simikon8052 *c )
{
  c->PC += 3;
  c->updateIRAM( DPL, c->activeCode[ c->PC - 3 ].b2 );
  c->updateIRAM( DPH, c->activeCode[ c->PC - 3 ].b1 );
}

void Simikon8052::mov_BADDR_C( Simikon8052 *c )
{
  c->PC += 2;
  c->setBit( c->activeCode[ c->PC - 2 ].b1, c->getBit( C ) );
}

void Simikon8052::movc_DPTR( Simikon8052 *c )
{
  uint codeAddress = ( ( c->iram[ DPH ] << 8 ) | c->iram[ DPL ] ) + c->iram[ ACC ];

  c->PC++;
  c->updateIRAM( ACC, c->activeCode[ codeAddress ].opCode );
}

void Simikon8052::subb_DATA( Simikon8052 * )
{
  cout << "Simikon8052::subb_DATA(): not implemented!\n";
}

void Simikon8052::subb_IRAM( Simikon8052 * )
{
  cout << "Simikon8052::subb_IRAM(): not implemented!\n";
}

void Simikon8052::subb_INDR( Simikon8052 * )
{
  cout << "Simikon8052::subb_INDR(): not implemented!\n";
}

void Simikon8052::subb_DIRR( Simikon8052 * )
{
  cout << "Simikon8052::subb_DIRR(): not implemented!\n";
}

void Simikon8052::orl_C_bADDR( Simikon8052 * )
{
  cout << "Simikon8052::orl_C_bADDR(): not implemented!\n";
}

void Simikon8052::mov_C_BADDR( Simikon8052 *c )
{
  c->PC += 2;
  c->setBit( C, c->getBit( c->activeCode[ c->PC - 2 ].b1 ) );
}

void Simikon8052::inc_DPTR( Simikon8052 *c )
{
  c->PC++;

  if ( ++c->iram[ DPL ] > 0xFF ) {
    c->iram[ DPL ] = 0;

    if ( ++c->iram[ DPH ] > 0xFF )
      c->iram[ DPH ] = 0;
    c->updateIRAM( DPH, c->iram[ DPH ] );
  }

  c->updateIRAM( DPL, c->iram[ DPL ] );
}

void Simikon8052::mul_AB( Simikon8052 *c )
{
  uint product = c->iram[ ACC ] * c->iram[ B ];
  
  c->iram[ ACC ] = product & 0x00FF;
  c->iram[ B ] = ( product & 0xFF00 ) >> 8;

  c->setBit( OV, ( product > 0x00FF ) );
  c->setBit( C, false ); // carry flag is always cleared

  c->PC++;

  c->updateIRAM( ACC, c->iram[ ACC ] );
  c->updateIRAM( B, c->iram[ B ] );
}

void Simikon8052::undefind( Simikon8052 * )
{
  cout << "WARNING: Pokusavate izvrsiti nedefiniranu instrukciju!\n";
}

void Simikon8052::mov_INDR_IRAM( Simikon8052 * )
{
  cout << "Simikon8052::mov_INDR_IRAM(): not implemented!\n";
}

void Simikon8052::mov_DIRR_IRAM( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0xA8;
  uint address = c->GetRXaddress( r );

  c->PC += 2;
  c->updateIRAM( address, c->iram[ c->activeCode[ c->PC - 2 ].b1 ] );
}

void Simikon8052::anl_C_bADDR( Simikon8052 * )
{
  cout << "Simikon8052::anl_C_bADDR(): not implemented!\n";
}

void Simikon8052::cpl_BADDR( Simikon8052 *c )
{
  uint bitAddress = c->activeCode[ c->PC ].b1;

  c->PC += 2;
  c->setBit( bitAddress, !c->getBit( bitAddress ) );
}

void Simikon8052::cpl_C( Simikon8052 * )
{
  cout << "Simikon8052::cpl_C(): not implemented!\n";
}

void Simikon8052::cjne_A_DATA( Simikon8052 *c )
{
  c->tmp_carry = c->getBit( C );

  c->PC = c->cjne_A_DATA_PC( c );

  if ( c->tmp_carry != c->getBit( C ) )
    c->setBit( C, c->tmp_carry );
}

uint Simikon8052::cjne_A_DATA_PC( Simikon8052 *c )
{
  uint PC = c->PC + 3;

  if ( c->iram[ ACC ] != c->activeCode[ c->PC ].b1 ) {
    PC += ( c->activeCode[ c->PC ].b2 < 128 ) ? 
      c->activeCode[ c->PC ].b2 : - ( 0xFF - c->activeCode[ c->PC ].b2 + 1 );

    c->tmp_carry = ( c->iram[ ACC ] < c->activeCode[ c->PC ].b1 );
  }

  return PC;
}

void Simikon8052::cjne_A_IRAM( Simikon8052 *c )
{
  c->tmp_carry = c->getBit( C );

  c->PC = c->cjne_A_IRAM_PC( c );

  if ( c->tmp_carry != c->getBit( C ) )
    c->setBit( C, c->tmp_carry ); 
}

uint Simikon8052::cjne_A_IRAM_PC( Simikon8052 *c )
{
  uint PC = c->PC + 3;

  if ( c->iram[ ACC ] != c->iram[ c->activeCode[ c->PC ].b1 ] ) {
    PC += ( c->activeCode[ c->PC ].b2 < 128 ) ? 
      c->activeCode[ c->PC ].b2 : - ( 0xFF - c->activeCode[ c->PC ].b2 + 1 );

    c->tmp_carry = ( c->iram[ ACC ] < c->iram[ c->activeCode[ c->PC ].b1 ] );
  }

  return PC;
}

void Simikon8052::cjne_INDR_DATA( Simikon8052 * )
{
  cout << "Simikon8052::cjne_INDR_DATA(): not implemented!\n";
}

void Simikon8052::cjne_DIRR_DATA( Simikon8052 *c )
{
  c->tmp_carry = c->getBit( C );
  c->PC = c->cjne_DIRR_DATA_PC( c );

  if ( c->tmp_carry != c->getBit( C ) )
    c->setBit( C, c->tmp_carry );
}

uint Simikon8052::cjne_DIRR_DATA_PC( Simikon8052 *c )
{
  uint PC = c->PC + 3;

  uint r = c->activeCode[ c->PC ].opCode - 0xB8;
  uint rdata = c->GetRXvalue( r );

  if ( rdata != c->activeCode[ c->PC ].b1 ) {
    PC += ( c->activeCode[ c->PC ].b2 < 128 ) ? 
      c->activeCode[ c->PC ].b2 : - ( 0xFF - c->activeCode[ c->PC ].b2 + 1 );

    c->tmp_carry = ( rdata < c->activeCode[ c->PC ].b1 );
  }

  return PC;
}

void Simikon8052::push( Simikon8052 *c )
{
  c->Push( c->iram[ c->activeCode[ c->PC ].b1 ] );
  c->PC += c->activeCode[ c->PC ].bytes;
}

void Simikon8052::clr_BADDR( Simikon8052 *c )
{
  c->PC += 2;
  c->setBit( c->activeCode[ c->PC - 2 ].b1, false );
}

void Simikon8052::clr_C( Simikon8052 *c )
{
  c->PC++;
  c->setBit( C, false );
}

void Simikon8052::swap( Simikon8052 * )
{
  cout << "Simikon8052::swap(): not implemented!\n";
}

void Simikon8052::xch_IRAM( Simikon8052 * )
{
  cout << "Simikon8052::xch_IRAM(): not implemented!\n";
}

void Simikon8052::xch_A_INDR( Simikon8052 * )
{
  cout << "Simikon8052::xch_A_INDR(): not implemented!\n";
}

void Simikon8052::xch_A_DIRR( Simikon8052 *c )
{
  uint r   = c->activeCode[ c->PC ].opCode - 0xC8;
  uint adr = c->GetRXaddress( r );

  c->PC++;

  uint tmp = c->iram[ ACC ];
  c->iram[ ACC ] = c->iram[ adr ];
  c->iram[ adr ] = tmp;

  c->updateIRAM( ACC, c->iram[ ACC ] );
  c->updateIRAM( adr, c->iram[ adr ] );
}

void Simikon8052::pop( Simikon8052 *c )
{
  c->PC += 2;
  c->updateIRAM( c->activeCode[ c->PC - 2 ].b1, c->Pop() );
}

void Simikon8052::setb_BADDR( Simikon8052 *c )
{
  c->PC += 2;
  c->setBit( c->activeCode[ c->PC - 2 ].b1, true );
}

void Simikon8052::setb_C( Simikon8052 *c )
{
  c->PC += c->activeCode[ c->PC ].bytes;
  c->setBit( C, true );
}

void Simikon8052::da( Simikon8052 * )
{
  cout << "Simikon8052::da(): not implemented!\n";
}

void Simikon8052::djnz_IRAM( Simikon8052 * )
{
  cout << "Simikon8052::djnz_IRAM(): not implemented!\n";
}

void Simikon8052::xchd( Simikon8052 * )
{
  cout << "Simikon8052::xchd(): not implemented!\n";
}

void Simikon8052::djnz_DIRR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0xD8;
  uint address = c->GetRXaddress( r );

  c->PC = c->djnz_DIRR_PC( c );

  if ( c->iram[ address ]-- == 0 ) {
    c->iram[ address ] = 0xFF;
  }

  c->updateIRAM( address, c->iram[ address ] );
}

uint Simikon8052::djnz_DIRR_PC( Simikon8052 *c )
{
  uint PC = c->PC + 2;

  uint r = c->activeCode[ c->PC ].opCode - 0xD8;
  uint address = c->GetRXaddress( r );

  if ( c->iram[ address ] - 1 != 0 )
    PC += ( c->activeCode[ c->PC ].b1 < 128 ) ? 
      c->activeCode[ c->PC ].b1 : - ( 0xFF - c->activeCode[ c->PC ].b1 + 1 );

  return PC;
}

void Simikon8052::movx_A_INDDPTR( Simikon8052 * )
{
  cout << "Simikon8052::movx_A_INDDPTR(): not implemented!\n";
}

void Simikon8052::movx_A_INDR( Simikon8052 * )
{
  cout << "Simikon8052::movx_A_INDR(): not implemented!\n";
}

void Simikon8052::clr_A( Simikon8052 *c )
{
  c->PC++;
  c->updateIRAM( ACC, 0 );
}

void Simikon8052::mov_A_IRAM( Simikon8052 *c )
{
  c->PC += 2;
  c->updateIRAM( ACC, c->iram[ c->activeCode[ c->PC - 2 ].b1 ] );
}

void Simikon8052::mov_A_INDR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0xE6;
  uint dataAddress = c->GetRXvalue( r );

  c->PC++;
  c->updateIRAM( ACC, c->iram[ dataAddress ] );
}

void Simikon8052::mov_A_DIRR( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0xE8;
  uint data = c->GetRXvalue( r );

  c->PC++;
  c->updateIRAM( ACC, data );
}

void Simikon8052::movx_INDDPTR_A( Simikon8052 * )
{
  cout << "Simikon8052::movx_INDDPTR_A(): not implemented!\n";
}

void Simikon8052::movx_INDR_A( Simikon8052 * )
{
  cout << "Simikon8052::movx_INDR_A(): not implemented!\n";
}

void Simikon8052::cpl_A( Simikon8052 *c )
{
  c->PC++;
  c->iram[ ACC ] = ~c->iram[ ACC ];

  c->updateIRAM( ACC, c->iram[ ACC ] );
}

void Simikon8052::mov_IRAM_A( Simikon8052 *c )
{
  c->PC += 2;
  c->updateIRAM( c->activeCode[ c->PC - 2].b1, c->iram[ ACC ] );
}

void Simikon8052::mov_INDR_A( Simikon8052 *c )
{
  uint r = c->activeCode[ c->PC ].opCode - 0xF6;
  uint address = c->GetRXvalue( r );

  c->PC++;
  c->updateIRAM( address, c->iram[ ACC ] );
}

void Simikon8052::mov_DIRR_A( Simikon8052 * )
{
  cout << "Simikon8052::mov_DIRR_A(): not implemented!\n";
}
