/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "Simikon8052.hh"
#include "SimikonProject.hh"
#include "SimikonSettings.hh"
#include "SimikonActions.hh"
#include "SimikonConsole.hh"
#include "SimikonUtils.hh"

#include "LeakTracker.hh"

// QT_INCLUDES
#include <qtimer.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qlistview.h>
#include <qprogressbar.h>

// MISC_INCLUDES
#include <iostream.h>

// DEFINES
#define MAX_LINE_SIZE 256
#define DATA_RECORD 0
#define EOF_RECORD 1

typedef struct {
  const char *name;
  uint        value;
  uint        address;
} Register;

static const Register gpr_registers[] = {
  { "R0", 0x00, R0 },
  { "R1", 0x00, R1 },  
  { "R2", 0x00, R2 },
  { "R3", 0x00, R3 },
  { "R4", 0x00, R4 },
  { "R5", 0x00, R5 },
  { "R6", 0x00, R6 },
  { "R7", 0x00, R7 }
};

#define NO_OF_GPR sizeof( gpr_registers ) / sizeof( gpr_registers[0] )

static const Register sfr_registers[] = {
  { "PC",   0x00, 0    },
  { "ACC",  0x00, ACC  },
  { "B",    0x00, B    },
  { "SP",   0x07, SP   },
  { "DPL",  0x00, DPL  },
  { "DPH",  0x00, DPH  },
  { "PCON", 0x00, PCON },
  { "SCON", 0x00, SCON },
  { "TCON", 0x00, TCON },
  { "TMOD", 0x00, TMOD },
  { "TL0",  0x00, TL0  },
  { "TL1",  0x00, TL1  },
  { "TH0",  0x00, TH0  },
  { "TH1",  0x00, TH1  },
  { "P0",   0xFF, P0   },
  { "P1",   0xFF, P1   },
  { "P2",   0xFF, P2   },
  { "P3",   0xFF, P3   },
  { "SBUF", 0x00, SBUF },
  { "IE",   0x00, IE   },
  { "IP",   0x00, IP   },
  { "PSW",  0x00, PSW  }
};

#define NO_OF_SFR sizeof( sfr_registers ) / sizeof( sfr_registers[0] )

// INTERRUPT STUFF

enum {
  X0_HANDLER_VECTOR = 0x0003,
  T0_HANDLER_VECTOR = 0x000B,
  X1_HANDLER_VECTOR = 0x0013,
  T1_HANDLER_VECTOR = 0x001B,
   S_HANDLER_VECTOR = 0x0023
};

typedef struct {
  uint        bit;
  const char *name;
  uint        address;
} SFR;

static const SFR ie_sfr[] = {
  { 0, "EX0", EX0 },
  { 1, "ET0", ET0 },
  { 2, "EX1", EX1 },
  { 3, "ET1", ET1 },
  { 4, "ES",  ES  },
  { 5, "-",   0xAD },
  { 6, "-",   0xAE },
  { 7, "EA",  EA }
};

static const SFR ip_sfr[] = {
  { 0, "PX0", PX0  },
  { 1, "PT0", PT0  },
  { 2, "PX1", PX1  },
  { 3, "PT1", PT1  },
  { 4, "PS",  PS   },
  { 5, "-",   0xBD },
  { 6, "-",   0xBE },
  { 7, "-",   0xBF }
};

static const SFR psw_sfr[] = {
  { 0, "P",   P   },
  { 1, "-",   F0  },
  { 2, "OV",  OV  },
  { 3, "RS0", RS0 },
  { 4, "RS1", RS1 },
  { 5, "-",   F1  },
  { 6, "AC",  AC  },
  { 7, "C",   C   }
};

static const SFR scon_sfr[] = {
  { 0, "RI",  RI  },
  { 1, "TI",  TI  },
  { 2, "RB8", RB8 },
  { 3, "TB8", TB8 },
  { 4, "REN", REN },
  { 5, "SM2", SM2 },
  { 6, "SM1", SM1 }, 
  { 7, "SM0", SM0 }
};

static const SFR pcon_sfr[] = {
  { 0, "IDLE", PCON   },
  { 1, "PD",   PCON+1 },
  { 2, "GF0",  PCON+2 },
  { 3, "GF1",  PCON+3 },
  { 4, "-",    PCON+4 },
  { 5, "-",    PCON+5 },
  { 6, "-",    PCON+6 },
  { 7, "SMOD", PCON+7 }
};

static const SFR tcon_sfr[] = {
  { 0, "",    TCON   },
  { 1, "",    TCON+1 },
  { 2, "",    TCON+2 },
  { 3, "",    TCON+3 },
  { 4, "TR0", TCON+4 },
  { 5, "TF0", TCON+5 },
  { 6, "TR1", TCON+6 },
  { 7, "TF1", TCON+7 }
};

static const SFR tmod_sfr[] = {
  { 0, "T0M0",  TMOD   },
  { 1, "T0M1",  TMOD+1 },
  { 2, "C/T0",  TMOD+2 },
  { 3, "GATE0", TMOD+3 },
  { 4, "T1M0",  TMOD+4 },
  { 5, "T1M1",  TMOD+5 },
  { 6, "C/T1",  TMOD+6 },
  { 7, "GATE1", TMOD+7 }
};

static const SFR px_sfr[ 4 ][ 8 ] = {
  { { 0, "P0.0", P0   },
    { 1, "P0.1", P0+1 },
    { 2, "P0.2", P0+2 },
    { 3, "P0.3", P0+3 },
    { 4, "P0.4", P0+4 },
    { 5, "P0.5", P0+5 },
    { 6, "P0.6", P0+6 },
    { 7, "P0.7", P0+7 } },
  { { 0, "P1.0", P1   },
    { 1, "P1.1", P1+1 },
    { 2, "P1.2", P1+2 },
    { 3, "P1.3", P1+3 },
    { 4, "P1.4", P1+4 },
    { 5, "P1.5", P1+5 },
    { 6, "P1.6", P1+6 },
    { 7, "P1.7", P1+7 } },
  { { 0, "P2.0", P2   },
    { 1, "P2.1", P2+1 },
    { 2, "P2.2", P2+2 },
    { 3, "P2.3", P2+3 },
    { 4, "P2.4", P2+4 },
    { 5, "P2.5", P2+5 },
    { 6, "P2.6", P2+6 },
    { 7, "P2.7", P2+7 } },
  { { 0, "P3.0", P3   },
    { 1, "P3.1", P3+1 },
    { 2, "P3.2", P3+2 },
    { 3, "P3.3", P3+3 },
    { 4, "P3.4", P3+4 },
    { 5, "P3.5", P3+5 },
    { 6, "P3.6", P3+6 },
    { 7, "P3.7", P3+7 } }
};

enum {
  PC_I, ACC_I, B_I, SP_I, DPL_I, DPH_I, PCON_I, SCON_I, TCON_I, TMOD_I,
  TL0_I, TL1_I, TH0_I, TH1_I, P0_I, P1_I, P2_I, P3_I, SBUF_I, IE_I, IP_I, PSW_I
};

typedef enum {
  OP_CODE = 1,
  ARG1 = 2,
  ARG2 = 3
} InstructionByte; 

// GLOBAL_OBJECTS
extern QProgressBar    *progress;
extern SimikonActions  *simikonActions;
extern SimikonConsole  *simikonConsole;
extern SimikonSettings *simikonSettings;

// FUNCTION_DECLARATIONS
uint calculateNumberOfStepsForHEX( const QString &hexFile );
uint calculateNumberOfStepsForASM( const QString &asmFile );

/*===========================================================================*/
/* class: Simikon8052                                                        */
/*===========================================================================*/

Simikon8052::Simikon8052()
{
  iram = new uint[ 0xFF ];

  stack = new QStack<uint>;
  stack->setAutoDelete( true );

  timer = new QTimer( this, "Clock" );
  connect( timer, SIGNAL( timeout() ), this, SLOT( SlotSimulatorClock( void ) ) );

  depth             = 0;
  programLength     = 0;

  lowPriorityInterruptExecuting  = false;
  highPriorityInterruptExecuting = false;

  debuggerMode    = true;
  running         = false;
  steppingOut     = false;
  resetInProgress = false;

  Timer0mode = 0;
  Timer1mode = 0;
}

Simikon8052::~Simikon8052()
{
  delete [] iram;
  delete stack;
}

// PUBLIC_MEMBER_FUNCTIONS

void
Simikon8052::interrupt( Interrupt i )
{
  if ( highPriorityInterruptExecuting ) return;
  if ( lowPriorityInterruptExecuting && !getIPbit( i ) ) return;

  if ( getIPbit( i ) ) {
    highPriorityInterruptExecuting = true;
  } else {
    lowPriorityInterruptExecuting = true;
  }

  // LSB
  Push( PC & 0x00FF );
  Push( ( PC & 0xFF00 ) >> 8 );

  switch( i ) {

    case INT_X0:
      InterruptX0();
      break;

    case INT_X1:
      InterruptX1();
      break;

    case INT_T0:
      InterruptT0();
      break;

    case INT_T1:
      InterruptT1();
      break;

    case INT_S:
      InterruptS();
      break;
  }

  depth++;

  emit signalUpdateSFR( PC_I, uintToHexStr16( PC ) );

  if ( debuggerMode && !running ) {
    refreshActions(); 
    markNextInstruction();
  }
}

INSTALL_RETTYPE
Simikon8052::install( Project *prj, bool reloadASM )
{
  activeCode = prj->program;
  activeData = &prj->pData;

  // reset program memory 
  for ( uint i = 0; i <= programLength; i++ )
    emit ( this->*signalUpdateProgramMemory )( i, 0 ); 

  INSTALL_RETTYPE retval = Parse( prj->pData.hexFile, prj->pData.asmFile,
                                  prj->pData.dir, reloadASM );
  
  ParseListing();

  if ( retval == OK || retval == ASM_ERROR )
    prj->programLength = programLength;

  reset();

  return retval;
}

INSTALL_RETTYPE
Simikon8052::Parse( const QString &hexFile, const QString &asmFile,
                    const QString &dir, bool reloadASM )
{
  QFile fpHEX( FFP( hexFile, dir ) );
  QFile fpASM( FFP( asmFile, dir ) );

  asmFileOK = true;
  hexFileOK = true;

  INSTALL_RETTYPE retval = OK;

  if ( !fpHEX.open( IO_ReadOnly ) ) {
    cout << "WARNING: Simikon8052::Parse(): "
         << "Can't open HEX file for reading! \n";

    hexFileOK = false;
    retval = HEX_ERROR;

    emit signalInvalidHEX();
  }

  if ( !fpASM.open( IO_ReadOnly ) ) {
    cout << "WARNING: Simikon8052::Parse(): "
         << "Can't open ASM file for reading! \n";

    asmFileOK = false;
    retval = ASM_ERROR;

    emit signalInvalidASM();
  }
  
  if ( !asmFileOK && !hexFileOK ) {
    retval = FUCKED;
    return retval;   // nemamo sto trazit dalje :)
  } 

  if ( asmFileOK && hexFileOK )
    emit signalValidALL();

  QTextStream tfpHEX( &fpHEX );
  QTextStream tfpASM( &fpASM );

  uint        lineNoHEX = 0;
  uint        lineNoASM = 0;

  uint opCodeAddress   = 0xA5;
  uint instructionByte = OP_CODE;
  bool instructionDone = false;
  bool programCodeDone = false;

  QString mnemonicStr;
  QString opCodeStr;

  uint address = 0;
  uint bytes = 0;

  if ( hexFileOK ) {
    uint noOfSteps = calculateNumberOfStepsForHEX( fullFilePath( hexFile, dir ) );
    progress->setTotalSteps( noOfSteps );
 
    while ( !tfpHEX.eof() ) {
      uint   byte;
      uint   recType;
      int    recLength;
      char   dataBytes[ MAX_LINE_SIZE ];
      char  *pdata;

      QString lineStrHEX( tfpHEX.readLine() );

      sscanf( lineStrHEX.ascii(), ":%2x%4x%2x%s", 
              &recLength, &address, &recType, dataBytes );

      pdata = dataBytes;

      switch ( recType ) {
      
      case DATA_RECORD:

        for ( ; recLength > 0; recLength-- ) {
        
          sscanf( pdata, "%2x", &byte );
          pdata += 2;
          bytes++;

          // update progress bar
          progress->setProgress( progress->progress() + 1 );

          emit ( this->*signalUpdateProgramMemory )( address, byte );

          programLength = address;

          if ( !programCodeDone ) {
          
            switch ( instructionByte ) {
            
            case OP_CODE:
            
              if ( byte != 0xA5 ) {
                activeCode[ address ].opCode  = byte;
                activeCode[ address ].cycles  = instructionTable[ byte ].cycles;
                activeCode[ address ].bytes   = instructionTable[ byte ].bytes;
                activeCode[ address ].execute = instructionTable[ byte ].execute;
                activeCode[ address ].PC      = instructionTable[ byte ].PC;
                activeCode[ address ].lineHEX = lineNoHEX++;
                activeCode[ address ].isValid = true;
                
                opCodeStr   = "0x" + uintToHexStr8( byte );
                mnemonicStr = instructionTable[ byte ].mnemonic;
                
                switch ( byte ) {
                  // ajmp
                case 0x01:
                case 0x21:
                case 0x41:
                case 0x61:
                case 0x81:
                case 0xA1:
                case 0xC1:
                case 0xE1:
                  // acall
                case 0x11:
                case 0x31:
                case 0x51:
                case 0x71:
                case 0x91:
                case 0xB1:
                case 0xD1:
                case 0xF1:

                  uint addr = address + 2;
                  addr &= 0xF800;
                  addr |= (( byte & 0xE0 ) << 3 );
                  addr >>= 8;

                  mnemonicStr = mnemonicStr.arg( uintToHexStr8( addr ) );
                  break;
                }

                opCodeAddress = address;

                address++;

              } else {
                programCodeDone = true;
              }

              break;
          
            case ARG1:
              activeCode[ opCodeAddress ].b1 = byte;

              opCodeStr.append( " 0x" );
              opCodeStr.append( uintToHexStr8( byte ) );
              
              mnemonicStr = mnemonicStr.arg( uintToHexStr8( byte ) );
              address++;
              break;

            case ARG2:
              activeCode[ opCodeAddress ].b2 = byte;

              opCodeStr.append( " 0x" );
              opCodeStr.append( uintToHexStr8( byte ) );
              
              mnemonicStr = mnemonicStr.arg( uintToHexStr8( byte ) );
              address++;
              break;
            }
          
            if ( !programCodeDone ) {

              if ( activeCode[ opCodeAddress ].bytes == instructionByte ) {
                instructionDone = true;
              } else 
                instructionByte++;

              if ( instructionDone ) {
                emit signalAddInstruction( "0x" + uintToHexStr16( opCodeAddress ),
                          mnemonicStr, opCodeStr,
                          QString::number( activeCode[ opCodeAddress ].cycles ) );
              
                if ( asmFileOK ) {
                  
                  QString m( instructionTable[ activeCode[ opCodeAddress ].opCode ].mnemonic );
                  m.truncate( m.find( ' ' ) );


                  while ( !tfpASM.eof() ) {
                    QString lineStrASM = tfpASM.readLine();

                    if ( reloadASM ) 
                      emit signalAddEditorLine( lineStrASM );

                    lineStrASM = lineStrASM.lower();
                    lineNoASM++;

                    int i_comment = lineStrASM.find( ";" );
                    int i_inst    = lineStrASM.find( m );

                    /* We check to see if found mnemonic string isn't inside of
                     * comment, because we'll get wrong instruction line number.
                     */

                    if ( i_inst != -1 )
                      if ( i_comment == -1 || ( i_comment > i_inst ) )
                          break;
                  }
                  
                  activeCode[ opCodeAddress ].lineASM = lineNoASM - 1;
                }

                address = opCodeAddress + activeCode[ opCodeAddress ].bytes;
                activeData->address = address;

                instructionDone = false;
                instructionByte = OP_CODE;

                mnemonicStr = "";
                opCodeStr = "";
              }
            } else {
              activeCode[ address ].opCode = byte;
              activeCode[ address ].bytes  = 1;
              address++;
            }
          } else {
            activeCode[ address ].opCode = byte;
            activeCode[ address ].bytes  = 1;
            address++;

            if ( byte == 0xA5 )
              programCodeDone = false;
          }
        }
        
        break;
      
      case EOF_RECORD:
        break;
        
      }
    }

    if ( programCodeDone ) { // procitaj do kraja asm file
      if ( asmFileOK && reloadASM ) {
        QString lineStr;

        while ( !tfpASM.eof() ) {
          lineStr = tfpASM.readLine();
          emit signalAddEditorLine( lineStr );
        }
      }
    }

    fpHEX.close();
    fpASM.close();

  } else if ( asmFileOK ) {

    if ( reloadASM ) {

      uint noOfSteps = calculateNumberOfStepsForASM( FFP( asmFile, dir ) );
      progress->setTotalSteps( noOfSteps );

      while ( !tfpASM.eof() ) {
        QString lineStrASM = tfpASM.readLine();
        emit signalAddEditorLine( lineStrASM );
        progress->setProgress( progress->progress() + 1 );
      }
    }

    fpASM.close();
  }

  progress->reset();

  return retval;
}

void
Simikon8052::ParseListing( void )
{
  emit signalClearListing();

  if ( SS_GET_BOOL( "generate_listing" ) ) {
    QFile fpLST( FFP( activeData->lstFile, activeData->dir ) );

    if ( fpLST.open( IO_ReadOnly ) ) {
      QTextStream tfpLST( &fpLST );
      
      while ( !tfpLST.eof() ) {
        QString lineStr = tfpLST.readLine();
        emit signalAddListingLine( lineStr );
      }
    } else {
      emit signalInvalidLST();
    }
  } else {
    emit signalInvalidLST();
  }
}

void
Simikon8052::AnalyzePerformance( void )
{
  QListIterator<PerformanceData> it( *activePerformance );

  PerformanceData *pdata;

  while ( ( pdata = it.current() ) != 0 ) {

    if ( pdata->counting ) {
      long int end;

      if ( pdata->item->text( 3 ).isEmpty() ) {
        end = -1;
      } else { 
        end = strToUInt( pdata->item->text( 3 ) );
      }

      if ( end == ( long int ) PC ) {

        pdata->cycles += activeCode[ PC ].cycles;
        
        pdata->item->setText( 4, QString::number( pdata->cycles ) );
        pdata->item->setText( 5, uintToTimeStr( pdata->cycles, 
                                 activeData->controller.clock.toUInt() ) );

        pdata->counting = false;
        pdata->cycles   = 0;

      } else if ( end == -1 ) {

        if ( activeCode[ PC ].opCode == 0x22 || 
             activeCode[ PC ].opCode == 0x32 ) {

          if ( depth == pdata->depth ) {

            pdata->cycles += activeCode[ PC ].cycles;

            pdata->item->setText( 2, ( activeCode[ PC ].opCode == 0x22 ) ?
                                  "ret" : "reti" );
            pdata->item->setText( 3, "0x" + uintToHexStr16( PC ) );
            pdata->item->setText( 4, QString::number( pdata->cycles ) );
            pdata->item->setText( 5, uintToTimeStr( pdata->cycles,
                                     activeData->controller.clock.toUInt() ) );

            pdata->counting = false;
            pdata->cycles   = 0;
          }
        }
      }

      if ( pdata->counting ) // ako nije dosao do kraja
        pdata->cycles += activeCode[ PC ].cycles;

     } else if ( !pdata->counting ) {
       uint start = strToUInt( pdata->item->text( 1 ) );

      if ( start == PC ) {
        pdata->counting = true;
        pdata->depth = depth;

        pdata->cycles += activeCode[ PC ].cycles;
      }
     }
    ++it;
  }
}

void
Simikon8052::setActive( Project *prj )
{
  activeData        = &prj->pData;
  activeCode        = prj->program;
  activePerformance = prj->performanceData;

  analyzePerformance = &prj->analyzePerformance;

  signalUpdateProgramMemory = ( ( prj->pData.controller.programMemory ) ?
    &Simikon8052::signalUpdateIROM : &Simikon8052::signalUpdateXROM );
  signalClearProgramMemory = ( ( prj->pData.controller.programMemory ) ?
    &Simikon8052::signalClearIROM : &Simikon8052::signalClearXROM );

  // ocisti programsku memoriju
  emit ( this->*signalClearProgramMemory )();

  programLength = prj->programLength;
  
  uint bytes = 0;
  
  for ( uint i = 0; i < 0xFFFF; i++ ) {
    if ( activeCode[ i ].bytes ) {
    
      emit ( this->*signalUpdateProgramMemory )( i, activeCode[ i ].opCode );

      switch ( activeCode[ i ].bytes ) {
      case 2:
        emit ( this->*signalUpdateProgramMemory )( i+1, activeCode[ i ].b1 );
        break;
      case 3:
        emit ( this->*signalUpdateProgramMemory )( i+1, activeCode[ i ].b1 );
        emit ( this->*signalUpdateProgramMemory )( i+2, activeCode[ i ].b2 );
        break;
      }
      bytes += activeCode[ i ].bytes;

      if ( bytes >= programLength ) {
        break;
      }
    }
  }
}

void
Simikon8052::initialize( void )
{  
  // Init SFR
  emit signalAddSFR( sfr_registers[0].name, "---" ); // PC

  for ( uint i = 1; i < NO_OF_SFR; i++ ) {
    emit signalAddSFR( sfr_registers[i].name, 
                    "0x" + uintToHexStr8( sfr_registers[i].address ) + "h" );
  }

  // Init GPR
  for ( uint i = 0; i < NO_OF_GPR; i++ ) {
    emit signalAddGPR( gpr_registers[i].name, 
                    "0x" + uintToHexStr8( gpr_registers[i].address ) + "h" );
  }
 
  // Init IE
  for ( uint i = 0; i < 8; i++ ) {
    emit signalAddIEbit( i, ie_sfr[i].name, 
                         uintToHexStr8( ie_sfr[i].address ), false );
  }

  // Init IP
  for ( uint i = 0; i < 8; i++ ) {
    emit signalAddIPbit( i, ip_sfr[i].name,
                         uintToHexStr8( ip_sfr[i].address ), false );
  }

  // Init PSW
  for ( uint i = 0; i < 8; i++ ) {
    emit signalAddPSWbit( i, psw_sfr[i].name, 
                          uintToHexStr8( psw_sfr[i].address ), false );
  }
  
  // Init SCON
  for ( uint i = 0; i < 8; i++ ) {
    emit signalAddSCONbit( i, scon_sfr[i].name, 
                           uintToHexStr8( scon_sfr[i].address ), false );
  }

  // Init PCON
  for ( uint i = 0; i < 8; i++ ) {
    emit signalAddPCONbit( i, pcon_sfr[i].name, 
                           uintToHexStr8( pcon_sfr[i].address ), false );
  }

  // Init TCON
  for ( uint i = 0; i < 8; i++ ) {
    emit signalAddTCONbit( i, tcon_sfr[i].name, 
                           uintToHexStr8( tcon_sfr[i].address ), false );
  }

  // Init TMOD
  for ( uint i = 0; i < 8; i++ ) {
    emit signalAddTMODbit( i, tmod_sfr[i].name,
                           uintToHexStr8( tmod_sfr[i].address ), false );
  }
  
  // Init Ports ( P0..P3 )
  for ( uint i = 0; i < 4; i++ ) {
    for ( uint j = 0; j < 8; j++ ) {
      emit signalAddPXbit( i, j, px_sfr[i][j].name, 
                           uintToHexStr8( px_sfr[i][j].address ), true );
    }
  }
}

void
Simikon8052::setBit( uint address, uint bit, bool state )
{
  uint value = iram[ address ];

  value = ( state ) ? ( value | ( 0x0001 << bit ) ) :
                      ( value & ~( 0x0001 << bit ) );

  updateIRAM( address, value );
}

void
Simikon8052::setBit( uint bitAddress, bool state )
{
  uint address, bit;

  if ( bitAddress <= 0x2F ) {  // 0x00 ... 0x7Fh
    address = 0x20 + bitAddress / 8;
    bit     = bitAddress % 8;
  } else {
    bit     = bitAddress % 8;
    address = bitAddress - bit;
  }

  setBit( address, bit, state );
}

bool
Simikon8052::getBit( uint address, uint bit )
{
  return ( iram[ address ] & ( 0x0001 << bit ) );
}

bool
Simikon8052::getBit( uint bitAddress )
{
  uint address, bit;

  if ( bitAddress <= 0x2F ) {  // 0x00 ... 0x7Fh
    address = 0x20 + bitAddress / 8;
    bit     = bitAddress % 8;
  } else {
    bit     = bitAddress % 8;
    address = bitAddress - bit;
  }

  return getBit( address, bit );
}

void
Simikon8052::setIEbit( uint bit, bool state )
{
  setBit( IE, bit, state );
}

void
Simikon8052::setIPbit( uint bit, bool state )
{
  setBit( IP, bit, state );
}


bool
Simikon8052::getIEbit( uint bit )
{
  return getBit( IE, bit );
}

bool
Simikon8052::getIPbit( uint bit )
{
  return getBit( IP, bit );
}

void
Simikon8052::setPSWbit( uint bit, bool state )
{
  setBit( PSW, bit, state );
}

void
Simikon8052::setPXbit( uint port, uint bit, bool state )
{
  uint address = 0;

  switch ( port ) {
    case 0: address = P0; break;
    case 1: address = P1; break;
    case 2: address = P2; break;
    case 3: address = P3; break;
  }

  if ( address )
    setBit( address, bit, state );
}

void
Simikon8052::setRIflag( void )
{
  setBit( RI, true );
}

bool
Simikon8052::getRIflag( void )
{
  return getBit( RI );
}

void
Simikon8052::setPC( uint value )
{
  if ( debuggerMode && !running ) {
    PC = value;
    markNextInstruction();
  }
}

void
Simikon8052::setRX( uint r, uint value )
{
  uint address = Simikon8052::GetRXaddress( r );
  updateIRAM( address, value );
}

void
Simikon8052::reset( void )
{
  if ( running ) {
    running = steppingOut = false;
    timer->stop();
  }

  resetInProgress = true;

  for ( uint i = 0x00; i < 0x7F; i++ )
    updateIRAM( i, 0 );

  for ( uint i = 0; i < NO_OF_SFR; i++ )
    updateIRAM( sfr_registers[ i ].address, sfr_registers[ i ].value );

  for ( uint i = 0; i < NO_OF_GPR; i++ )
    updateIRAM( gpr_registers[ i ].address, gpr_registers[ i ].value );

  PC = 0x0000;

  emit signalUpdateSFR( PC_I, uintToHexStr16( 0 ) );

  depth = 0;

  emit signalResetStack();
  emit signalResetSerial();
  emit signalResetPerformance();
  emit signalResetMemoryView();

  lowPriorityInterruptExecuting  = false;
  highPriorityInterruptExecuting = false;

  if ( debuggerMode )
    markNextInstruction();

  PCoutOfBounds    = false;
  instructionFault = false;
  resetInProgress  = false;
}

void
Simikon8052::refreshActions( void )
{
  if ( activeCode[ PC ].bytes ) { // postoji instrukcija
    SA_ENABLED( "StepInto", true );
    if ( activeCode[ PC + activeCode[ PC ].bytes ].bytes ) {
      SA_ENABLED( "StepOver", true );
    } else {
      SA_ENABLED( "StepOver", false );
    }
  } else {
    PCoutOfBounds = true;
  }

  if ( depth > 0 ) {
    SA_ENABLED( "StepOut", true );
  } else {
    SA_ENABLED( "StepOut", false );
  }
}

void
Simikon8052::stopExecuting( void )
{
  timer->stop();
  running = steppingOut = false;

  markNextInstruction();

  if ( debuggerMode )
    refreshActions();
}

void
Simikon8052::startSimulation( void )
{
  running      = true;
  debuggerMode = false;
  timer->start( 0, false );
}

void
Simikon8052::stopSimulation( void )
{
  running      = false;
  debuggerMode = true;
  timer->stop();
}

void
Simikon8052::updateIRAM( uint address, uint value )
{
  uint mask = 1;

  value &= 0xFF;

  iram[ address ] = value;
  emit signalUpdateIRAM( address, iram[ address ] );

  int r = address - registersBank * 8;

  if ( ( address < 0x08 && registersBank == 0 ) ||
       ( address > 0x07 && address < 0x10 && registersBank == 1 ) ||
       ( address > 0x0F && address < 0x18 && registersBank == 2 ) || 
       ( address > 0x17 && address < 0x20 && registersBank == 3 ) ) {

    emit signalUpdateGPR( r, uintToHexStr8( value ) );
    return;
  }

  switch ( address ) {

    case SP:
      emit signalUpdateSFR( SP_I, uintToHexStr8( iram[ SP ] ) );
    break;

    case DPL:
      emit signalUpdateSFR( DPL_I, uintToHexStr8( iram[ DPL ] ) );
    break;

    case DPH:
      emit signalUpdateSFR( DPH_I, uintToHexStr8( iram[ DPH ] ) );
    break;

    case PCON:
      emit signalUpdateSFR( PCON_I, uintToHexStr8( iram[ PCON ] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdatePCONbit( bit, iram[ PCON ] & mask );
        mask *= 2;
      }
    break;

    case TCON:
      emit signalUpdateSFR( TCON_I, uintToHexStr8( iram[ TCON ] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdateTCONbit( bit, iram[ TCON ] & mask );
        mask *= 2;
      }

      if ( getBit( TCON, 4 ) )  { // TR0
        if ( !getBit( TMOD, 0 ) && !getBit( TMOD, 1 ) )
          Timer0mode = &Simikon8052::Timer0mode0;
        else if ( getBit( TMOD, 0 ) && !getBit( TMOD, 1 ) )
          Timer0mode = &Simikon8052::Timer0mode1;
        else if ( !getBit( TMOD, 0 ) && getBit( TMOD, 1 ) )
          Timer0mode = &Simikon8052::Timer0mode2;
        else if ( getBit( TMOD, 0 ) && getBit( TMOD, 1 ) )
          Timer0mode = &Simikon8052::Timer0mode3;
      } else
        Timer0mode = 0;

      if ( getBit( TCON, 6 ) ) { // TR1
        if ( !getBit( TMOD, 4 ) && !getBit( TMOD, 5 ) )
          Timer1mode = &Simikon8052::Timer1mode0;
        else if ( getBit( TMOD, 4 ) && !getBit( TMOD, 5 ) )
          Timer1mode = &Simikon8052::Timer1mode1;
        else if ( !getBit( TMOD, 4 ) && getBit( TMOD, 5 ) )
          Timer1mode = &Simikon8052::Timer1mode2;
        else if ( getBit( TMOD, 4 ) && getBit( TMOD, 5 ) )
          Timer1mode = &Simikon8052::Timer1mode3;
      } else
        Timer1mode = 0;

    break;

    case TMOD:
      emit signalUpdateSFR( TMOD_I, uintToHexStr8( iram[ TMOD ] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdateTMODbit( bit, iram[ TMOD ] & mask );
        mask *= 2;
      }
    break;

    case TL0:
      emit signalUpdateSFR( TL0_I, uintToHexStr8( iram[ TL0 ] ) );
    break;

    case TL1:
      emit signalUpdateSFR( TL1_I, uintToHexStr8( iram[ TL1 ] ) );
    break;

    case TH0:
      emit signalUpdateSFR( TH0_I, uintToHexStr8( iram[ TH0 ] ) );
    break;

    case TH1:
      emit signalUpdateSFR( TH1_I, uintToHexStr8( iram[ TH1 ] ) );
    break;

    case SCON:
      emit signalUpdateSFR( SCON_I, uintToHexStr8( iram[ SCON ] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdateSCONbit( bit, iram[ SCON ] & mask );
        mask *= 2;
      }

      emit signalEnableSerialReceive( getBit( REN ) );
    break;

    case SBUF:
      emit signalUpdateSFR( SBUF_I, uintToHexStr8( iram[ SBUF ] ) );

      if ( !resetInProgress ) {
        if ( getBit( EA ) && getBit( ES ) ) {

          // RI bit bi bio postavljen u slucaju da je interrupt generiran iz 
          // serial emulatora, ako nije znaci da je transmit interrupt

          if ( !getBit( RI ) ) {
            setBit( TI, true );
            emit signalSerialSend( iram[ SBUF ] );
          }

          interrupt( INT_S );
        }
      }
    break;

    case P0:
      emit signalUpdateSFR( P0_I, uintToHexStr8( iram[P0] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdatePXbit( 0, bit, iram[P0] & mask );
        mask *= 2;
      }
   break;

    case P1:
      emit signalUpdateSFR( P1_I, uintToHexStr8( iram[P1] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdatePXbit( 1, bit, iram[P1] & mask );
        mask *= 2;
      }
    break;

    case P2:
      emit signalUpdateSFR( P2_I, uintToHexStr8( iram[P2] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdatePXbit( 2, bit, iram[P2] & mask );
        mask *= 2;
      }
    break;

    case P3:
      emit signalUpdateSFR( P3_I, uintToHexStr8( iram[P3] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdatePXbit( 3, bit, iram[P3] & mask );
        mask *= 2;
      }
    break;

    case IE:
      emit signalUpdateSFR( IE_I, uintToHexStr8( iram[IE] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdateIEbit( bit, iram[IE] & mask );
        mask *= 2;
      }

      emit signalEnableInterrupts();
    break;

    case IP:
      emit signalUpdateSFR( IP_I, uintToHexStr8( iram[ IP ] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdateIPbit( bit, iram[ IP ] & mask );
        mask *= 2;
      }
    break;

    case ACC:
      emit signalUpdateSFR( ACC_I, uintToHexStr8( iram[ ACC ] ) );
    break;

    case B:
      emit signalUpdateSFR( B_I, uintToHexStr8( iram[ B ] ) );
    break;

    case PSW:
      emit signalUpdateSFR( PSW_I, uintToHexStr8( iram[ PSW ] ) );

      for ( uint bit = 0; bit < 8; bit++ ) {
        emit signalUpdatePSWbit( bit, iram[ PSW ] & mask );
        mask *= 2;
      }

      unsigned int rs_flags = ( iram[ PSW ] & 0x18 ) >> 3;

      if ( registersBank != rs_flags ) {
        registersBank = rs_flags;
        
        for ( uint i = 0; i < 8; i++ ) {
          emit signalUpdateGPR( i, uintToHexStr8( iram[ GetRXaddress( i ) ] ) );
        }
        emit signalUpdateRXaddresses( GetRXaddress( 0 ) );        
      }
    break;
  }

  if ( debuggerMode && !resetInProgress && !running
       && !instructionExecutionInProgress ) {

    markNextInstruction();
  }
}

void
Simikon8052::markNextInstruction( void )
{
  uint nextInstructionLine = 0;
  uint nextPC;

  if ( activeCode[ PC ].PC != 0 ) {
    nextPC = activeCode[ PC ].PC( this );
  } else {
    nextPC = PC + activeCode[ PC ].bytes;
  }

  emit signalSelectInstruction( activeCode[ PC ].lineHEX );
  
  if ( asmFileOK )
    emit signalSelectEditorLine( activeCode[ PC ].lineASM );

  if ( activeCode[ nextPC ].isValid ) {
    nextInstructionLine = activeCode[ nextPC ].lineHEX;
    emit signalMarkInstruction( nextInstructionLine );
  } else {
    //    SC_WARNING( tr( "" ) );
  }
}

// PRIVATE_MEMBER_FUNCTIONS

uint
Simikon8052::GetRXvalue( uint n )
{
  return iram[ n + registersBank * 8 ];
}

uint
Simikon8052::GetRXaddress( uint n )
{
  return n + registersBank * 8;
}

void
Simikon8052::Push( uint value )
{
  updateIRAM( SP, iram[ SP ] + 1 );
  updateIRAM( iram[ SP ], value );

  emit signalPushStack( uintToHexStr16( iram[ SP ] ), uintToHexStr8( value ) );
  stack->push( new uint( value ) );
}

uint
Simikon8052::Pop( void )
{
  if ( !stack->isEmpty() ) {
    updateIRAM( SP, iram[ SP ] - 1 );
    emit signalPopStack();
    return *stack->pop();
  } else {
    SC_ERROR( "Trying to pop data from empty stack!" );
    instructionFault = true;
  }

  return 0;
}

void
Simikon8052::InterruptX0( void )
{
  PC = X0_HANDLER_VECTOR;
}

void
Simikon8052::InterruptX1( void )
{
  PC = X1_HANDLER_VECTOR;
}

void
Simikon8052::InterruptT0( void )
{
  PC = T0_HANDLER_VECTOR;
  setBit( TCON, 4, 0 ); // TF0
}

void
Simikon8052::InterruptT1( void )
{
  PC = T1_HANDLER_VECTOR;
  setBit( TCON, 7, 0 ); // TF1
}

void
Simikon8052::InterruptS( void )
{
  PC = S_HANDLER_VECTOR;
}

void
Simikon8052::Timer0mode0( void )
{
  cout << "Simikon8052::Timer0mode0()" << endl;
}

void
Simikon8052::Timer0mode1( void )
{
  cout << "Simikon8052::Timer0mode1()" << endl;
}

void
Simikon8052::Timer0mode2( void ) // aka. auto-reload mode
{
  iram[ TL0 ] += activeCode[ PC ].cycles;

  if ( iram[ TL0 ] >= 0xFF ) { // check for overflow
    iram[ TL0 ] = iram[ TH0 ]; // reload TL0 with TH0
    setBit( TCON, 5, 1 ); // set timer overflow flag (TF0)
  }

  updateIRAM( TL0, iram[ TL0 ] );
}

void
Simikon8052::Timer0mode3( void )
{
  cout << "Simikon8052::Timer0mode3()" << endl;
}

void
Simikon8052::Timer1mode0( void )
{
  cout << "Simikon8052::Timer1mode0()" << endl; 
}

void
Simikon8052::Timer1mode1( void )
{
  cout << "Simikon8052::Timer1mode1()" << endl;
}

void
Simikon8052::Timer1mode2( void ) // aka. auto reload mode
{
  iram[ TL1 ] += activeCode[ PC ].cycles;

  if ( iram[ TL1 ] >= 0xFF ) { // check for overflow
    iram[ TL1 ] = iram[ TH1 ]; // reload TL1 with TH1
    setBit( TCON, 7, 1 ); // set timer overflow flag (TF1)
  }
  
  updateIRAM( TL1, iram[ TL1 ] );
}

void
Simikon8052::Timer1mode3( void )
{
  cout << "Simikon8052::Timer1mode3()" << endl;
}

// PUBLIC_SLOTS

void
Simikon8052::slotStepInto( void )
{
  if ( !activeCode[ PC ].bytes || PC >= activeData->address ) {
    SC_ERROR( tr( "Trying to execute code from undefined address!" ) );
    if ( !debuggerMode )
      emit signalStopExecuting();

    return;
  }

  if ( activeCode[ PC ].opCode == 0x11 ||  // acall
       activeCode[ PC ].opCode == 0x12 ||  // lcall
       activeCode[ PC ].opCode == 0x31 ||  // acall
       activeCode[ PC ].opCode == 0x51 ||  // acall
       activeCode[ PC ].opCode == 0x71 ||  // acall
       activeCode[ PC ].opCode == 0x91 ||  // acall
       activeCode[ PC ].opCode == 0xB1 ||  // acall
       activeCode[ PC ].opCode == 0xD1 ||  // acall
       activeCode[ PC ].opCode == 0xF1 ) { // acall
    depth++;

    if ( depth == 1 && debuggerMode && !running )
      refreshActions();
  }

  if ( *analyzePerformance )
    AnalyzePerformance();

  if ( activeCode[ PC ].opCode == 0x22 ||  // ret
       activeCode[ PC ].opCode == 0x32 ) { // reti
    depth--;

    if ( depth == 0 )
      if ( debuggerMode && !running )
        refreshActions();
  }

  if ( Timer0mode ) ( this->*Timer0mode )();
  if ( Timer1mode ) ( this->*Timer1mode )();

  instructionExecutionInProgress = true;
  activeCode[ PC ].execute( this );
  instructionExecutionInProgress = false;

  if ( instructionFault ) {
    if ( !debuggerMode )
      emit signalStopExecuting();

    SC_ERROR( tr( "Instruction fault has occurred!" ) );

    return;
  }

  // check for timer overflows and raise interrupts
  if ( getBit( TCON, 5 ) ) // TF0
    interrupt( INT_T0 );

  if ( getBit( TCON, 7 ) ) // TF1
    interrupt( INT_T1 );

  emit signalUpdateSFR( PC_I, uintToHexStr16( PC ) );

  if ( debuggerMode && !running ) {

    refreshActions();
    markNextInstruction();
    
    if ( PCoutOfBounds ) {
      SC_WARNING( tr( "Program Counter pointing to undefined code!" ) );
      reset();
      refreshActions();
    }
  }
}

void
Simikon8052::slotStepOut( void )
{
  if ( !running ) {
    running = steppingOut = true;
    stepOutDepth = depth;
    slotStepInto();
    timer->start( 0, false );
  }
}

void
Simikon8052::slotStepOver( void )
{
  if ( !running ) {
    if ( activeCode[ PC + activeCode[ PC ].bytes ].bytes ) {
      PC += activeCode[ PC ].bytes;
      emit signalUpdateSFR( PC_I, uintToHexStr16( PC ) );
      markNextInstruction();
    } else {
      SA_ENABLED( "StepOver", false );
    }
  }
}

void
Simikon8052::slotGotoBreakpoint( void )
{ 
  if ( !running ) {
    running = true;
    Simikon8052::slotStepInto();
    timer->start( 0, false );
  }
}

void
Simikon8052::slotReset( void )
{
  running = steppingOut = false;
  reset();
  timer->stop();
}

void
Simikon8052::slotInterruptX0( void )
{
  interrupt( INT_X0 );
}

void
Simikon8052::slotInterruptX1( void )
{
  interrupt( INT_X1 );
}

void
Simikon8052::slotInterruptT0( void )
{
  interrupt( INT_T0 );
}

void
Simikon8052::slotInterruptT1( void )
{
  interrupt( INT_T1 );
}

void
Simikon8052::slotSetPSWbit( uint bit, bool state )
{
  setBit( PSW, bit, state );
}

void
Simikon8052::slotSetIEbit( uint bit, bool state )
{
  setBit( IE, bit, state );
}

void
Simikon8052::slotSetIPbit( uint bit, bool state )
{
  setBit( IP, bit, state );
}

void
Simikon8052::slotSetSCONbit( uint bit, bool state )
{
  setBit( SCON, bit, state );
}

void
Simikon8052::slotSetPCONbit( uint bit, bool state )
{
  setBit( PCON, bit, state );
}

void
Simikon8052::slotSetTCONbit( uint bit, bool state )
{
  setBit( TCON, bit, state );
}

void
Simikon8052::slotSetTMODbit( uint bit, bool state )
{
  setBit( TMOD, bit, state );
}

// PRIVATE_SLOTS

void
Simikon8052::SlotSimulatorClock( void )
{
  if ( debuggerMode ) {

    if ( activeCode[ PC ].isBreak ) {
      Simikon8052::stopExecuting();
      return;
    }

    if ( steppingOut && ( stepOutDepth == depth  ) ) {
      if ( activeCode[ PC ].opCode == 0x22 || activeCode[ PC ].opCode == 0x32 ) {
        Simikon8052::stopExecuting();
      }
    }
  }

  slotStepInto();
}

/*===========================================================================*/
/* FUNCTION_DEFINITIONS                                                      */
/*===========================================================================*/

uint
calculateNumberOfStepsForHEX( const QString &hexFile )
{
  uint lines = 0;
  uint bytes = 0;
  QString s;

  QFile fp( hexFile );

  fp.open( IO_ReadOnly );

  QTextStream tfp( &fp );

  while ( !tfp.eof() ) {
    s = tfp.readLine();
    bytes += s.length();
    lines++;
  }

  return ( bytes - lines * 11 ) / 2;
}

uint
calculateNumberOfStepsForASM( const QString &asmFile )
{
  uint lines = 0;
  
  QFile fp( asmFile );

  fp.open( IO_ReadOnly );

  QTextStream tfp( &fp );

  while ( !tfp.eof() ) {
    tfp.readLine();
    lines++;
  }
  
  return lines;
}
