/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef SIMIKON_8052_HH
#define SIMIKON_8052_HH

#include <qobject.h>
#include <qstack.h>
#include <qlist.h>

class QTimer;
class QListViewItem;

class Project;

#include "DataStructures.hh"


// GENERAL PURPOSE REGISTERS (GPR)
enum {
  R0=0x00, R1, R2, R3, R4, R5, R6, R7
};

// SPECIAL FUNCTION REGISTERS (SFR)
enum {
  P0   = 0x80, SP   = 0x81, DPL = 0x82, DPH = 0x83, PCON = 0x87,
  TCON = 0x88, TMOD = 0x89, TL0 = 0x8A, TL1 = 0x8B, TH0  = 0x8C, TH1 = 0x8D,
  P1   = 0x90,
  SCON = 0x98, SBUF = 0x99,
  P2   = 0xA0,
  IE   = 0xA8,
  P3   = 0xB0,
  IP   = 0xB8,
  PSW  = 0xD0,
  ACC  = 0xE0,
  B    = 0xF0
};

enum {
  EX0 = 0xA8,
  ET0 = 0xA9,
  EX1 = 0xAA,
  ET1 = 0xAB,
  ES  = 0xAC,
  EA  = 0xAF
};

enum {
  PX0 = 0xB8,
  PT0 = 0xB9,
  PX1 = 0xBA,
  PT1 = 0xBB,
  PS  = 0xBC
};

enum {
  P   = 0xD0,
  F0  = 0xD1,
  OV  = 0xD2,
  RS0 = 0xD3,
  RS1 = 0xD4,
  F1  = 0xD5,
  AC  = 0xD6,
  C   = 0xD7
};

enum {
  RI  = 0x98,
  TI  = 0x99,
  RB8 = 0x9A,
  TB8 = 0x9B,
  REN = 0x9C,
  SM2 = 0x9D,
  SM1 = 0x9E,
  SM0 = 0x9F
};

typedef enum {
  OK,              // asm & hex are loaded
  FUCKED,          // neither asm or hex are loaded 
  ASM_ERROR,       // hex is loaded
  HEX_ERROR        // asm is loaded
} INSTALL_RETTYPE;

/*===========================================================================*/
/* class: Simikon8052                                                        */
/*===========================================================================*/

class Simikon8052: public QObject {
  Q_OBJECT

public:
  Simikon8052();
  ~Simikon8052();

  INSTALL_RETTYPE install( Project *prj, bool reloadASM = true );
  void setActive( Project *prj );

  void reset( void );
  void initialize( void );
  void refreshActions( void );

  void stopExecuting( void );

  void startSimulation( void );
  void stopSimulation( void );

  void interrupt( Interrupt i );

  void setIEbit( uint bit, bool state );
  void setIPbit( uint bit, bool state );
  bool getIEbit( uint bit );
  bool getIPbit( uint bit );
  void setPSWbit( uint bit, bool state );

  void setPXbit( uint port, uint bit, bool state );

  void setBit( uint address, uint bit, bool state );
  void setBit( uint bitAddress, bool state );
  bool getBit( uint address, uint bit );
  bool getBit( uint bitAddress );

  void setPC( uint value );
  void setRX( uint r, uint value );

  void updateIRAM( uint address, uint value );

  void markNextInstruction( void );
  
  void setRIflag( void );
  bool getRIflag( void );

public slots:  
  void slotStepInto( void );
  void slotStepOut( void );
  void slotStepOver( void );
  void slotGotoBreakpoint( void );
  void slotReset( void );
  // INTERRUPT_STUFF
  void slotInterruptX0( void );
  void slotInterruptX1( void );
  void slotInterruptT0( void );
  void slotInterruptT1( void );

private slots:
  void SlotSimulatorClock( void );

  void slotSetPSWbit( uint, bool );
  void slotSetIEbit( uint, bool );
  void slotSetIPbit( uint, bool );
  void slotSetSCONbit( uint, bool );
  void slotSetPCONbit( uint, bool );
  void slotSetTCONbit( uint, bool );
  void slotSetTMODbit( uint, bool );

private:
  // PRIVATE MEMBER FUNCTIONS
  uint GetRXvalue( uint n );
  uint GetRXaddress( uint n );
  void Push( uint value );
  uint Pop( void );

  INSTALL_RETTYPE Parse( const QString &hexFile, const QString &asmFile, 
                         const QString &dir, bool reloadASM );
  void ParseListing( void );

  void AnalyzePerformance( void );

  PerformanceData *FindPerformanceData( QListViewItem *i );

  // INTERRUPT HANDLERS
  void InterruptX0( void );
  void InterruptX1( void );
  void InterruptT0( void );
  void InterruptT1( void );
  void InterruptS( void );

  // TIMER MODE FUNCTIONS
  void Timer0mode0( void );
  void Timer0mode1( void );
  void Timer0mode2( void );
  void Timer0mode3( void );
  void Timer1mode0( void );
  void Timer1mode1( void );
  void Timer1mode2( void );
  void Timer1mode3( void );

signals:
  // interrupts
  void signalAddIEbit( uint, const QString &, const QString &, bool );
  void signalAddIPbit( uint, const QString &, const QString &, bool );
  void signalUpdateIEbit( uint, bool );
  void signalUpdateIPbit( uint, bool );

  // registers
  void signalAddSFR( const QString &, const QString & );
  void signalAddGPR( const QString &, const QString & );
  void signalUpdateSFR( uint, const QString & );
  void signalUpdateGPR( uint, const QString & );
  void signalUpdateRXaddresses( uint );

  // PSW
  void signalAddPSWbit( uint, const QString &, const QString &, bool );
  void signalUpdatePSWbit( uint, bool );

  // SCON
  void signalAddSCONbit( uint, const QString &, const QString &, bool );
  void signalUpdateSCONbit( uint, bool );

  // PCON
  void signalAddPCONbit( uint, const QString &, const QString &, bool );
  void signalUpdatePCONbit( uint, bool );

  // TCON
  void signalAddTCONbit( uint, const QString &, const QString &, bool );
  void signalUpdateTCONbit( uint, bool );

  // TMOD
  void signalAddTMODbit( uint, const QString &, const QString &, bool );
  void signalUpdateTMODbit( uint, bool );

  // memory
  void signalUpdateIRAM( uint, uint );
  void signalUpdateIROM( uint, uint );
  void signalClearIROM( void );
  void signalUpdateXRAM( uint, uint );
  void signalUpdateXROM( uint, uint );
  void signalClearXROM( void );

  void signalResetMemoryView( void );

  // dissembler
  void signalAddInstruction( const QString &, const QString &, const QString &,
                             const QString & );
  void signalSelectInstruction( uint );
  void signalMarkInstruction( uint );

  // editor
  void signalAddEditorLine( const QString & );
  void signalSelectEditorLine( uint );

  // listing
  void signalAddListingLine( const QString & );
  void signalClearListing( void );

  void signalInvalidASM( void );
  void signalInvalidHEX( void );
  void signalInvalidLST( void );
  void signalValidALL( void );

  // stack
  void signalPushStack( const QString &, const QString & );
  void signalPopStack( void );
  void signalResetStack( void ); 

  // interrupts
  void signalEnableInterrupts( void );

  // I/O Ports
  void signalAddPXbit( uint, uint, const QString &, const QString &, bool );
  void signalUpdatePXbit( uint, uint, bool );

  // serial
  void signalSerialSend( uint characterCode );
  void signalResetSerial( void );
  void signalEnableSerialReceive( bool );

  // performance
  void signalResetPerformance( void );


  void signalStopExecuting( void );

private:
  // DATA
  uint *iram;
  uint PC; // program counter
  
  uint depth;
  uint stepOutDepth;

  uint registersBank;
  
  void ( Simikon8052::*signalUpdateProgramMemory )( uint, uint );
  void ( Simikon8052::*signalClearProgramMemory )( void );

  // timer functions
  void ( Simikon8052::*Timer0mode )( void );
  void ( Simikon8052::*Timer1mode )( void );
  
  InstructionData *activeCode;
  ProjectData     *activeData;
  uint             programLength;
  QList<PerformanceData> *activePerformance;

  static const InstructionData instructionTable[ 256 ];
  QStack<uint> *stack;
  
  bool *analyzePerformance;
  
  bool lowPriorityInterruptExecuting;
  bool highPriorityInterruptExecuting;

  QTimer *timer;

  bool debuggerMode; // if false, simulator
  bool running;
  bool steppingOut;
  bool resetInProgress;
  bool instructionExecutionInProgress;
  bool PCoutOfBounds;
  bool asmFileOK;
  bool hexFileOK;

  bool tmp_carry;
  bool instructionFault;

private:
  static uint ajmp_PC( Simikon8052 * );
  static uint ljmp_PC( Simikon8052 * );
  static uint jbc_PC( Simikon8052 * );
  static uint acall_PC( Simikon8052 * );
  static uint lcall_PC( Simikon8052 * );
  static uint jb_PC( Simikon8052 * );
  static uint jnb_PC( Simikon8052 * );
  static uint jnc_PC( Simikon8052 * );
  static uint jnz_PC( Simikon8052 * );
  static uint jc_PC( Simikon8052 * );
  static uint jz_PC( Simikon8052 * );
  static uint cjne_A_DATA_PC( Simikon8052 * );
  static uint cjne_A_IRAM_PC( Simikon8052 * );
  static uint cjne_A_DIRR_PC( Simikon8052 * ); 
  static uint cjne_DIRR_DATA_PC( Simikon8052 * );
  static uint djnz_DIRR_PC( Simikon8052 * );
  static uint ret_PC( Simikon8052 * );
  static uint sjmp_PC( Simikon8052 * );

  // INSTRUCTIONS
  static void nop( Simikon8052 * );
  static void ajmp( Simikon8052 * );
  static void ljmp( Simikon8052 * );
  static void rr( Simikon8052 * );
  static void inc_A( Simikon8052 * );
  static void inc_IRAM( Simikon8052 * );
  static void inc_INDR( Simikon8052 * );
  static void inc_DIRR( Simikon8052 * );
  static void jbc( Simikon8052 * );
  static void acall( Simikon8052 * );
  static void lcall( Simikon8052 * );
  static void rrc( Simikon8052 * );
  static void dec_A( Simikon8052 * );
  static void dec_IRAM( Simikon8052 * );
  static void dec_INDR( Simikon8052 * );
  static void dec_DIRR( Simikon8052 * );
  static void jb( Simikon8052 * );
  static void ret( Simikon8052 * );
  static void rl( Simikon8052 * );
  static void add_DATA( Simikon8052 * );
  static void add_IRAM( Simikon8052 * );
  static void add_INDR( Simikon8052 * );
  static void add_DIRR( Simikon8052 * );
  static void jnb( Simikon8052 * );
  static void reti( Simikon8052 * );
  static void rlc( Simikon8052 * );
  static void addc_DATA( Simikon8052 * );
  static void addc_IRAM( Simikon8052 * );
  static void addc_INDR( Simikon8052 * );
  static void addc_DIRR( Simikon8052 * );
  static void jc( Simikon8052 * );
  static void orl_IRAM_A( Simikon8052 * );
  static void orl_IRAM_DATA( Simikon8052 * );
  static void orl_DATA( Simikon8052 * );
  static void orl_IRAM( Simikon8052 * );
  static void orl_INDR( Simikon8052 * );
  static void orl_DIRR( Simikon8052 * );
  static void jnc( Simikon8052 * );
  static void anl_IRAM_A( Simikon8052 * );
  static void anl_IRAM_DATA( Simikon8052 * );
  static void anl_DATA( Simikon8052 * );
  static void anl_IRAM( Simikon8052 * );
  static void anl_INDR( Simikon8052 * );
  static void anl_DIRR( Simikon8052 * );
  static void jz( Simikon8052 * );
  static void xrl_IRAM_A( Simikon8052 * );
  static void xrl_IRAM_DATA( Simikon8052 * );
  static void xrl_DATA( Simikon8052 * );
  static void xrl_IRAM( Simikon8052 * );
  static void xrl_INDR( Simikon8052 * );
  static void xrl_DIRR( Simikon8052 * );
  static void jnz( Simikon8052 * );
  static void orl_C_BADDR( Simikon8052 * );
  static void jmp( Simikon8052 * );
  static void mov_A_DATA( Simikon8052 * );
  static void mov_IRAM_DATA( Simikon8052 * );
  static void mov_INDR_DATA( Simikon8052 * );
  static void mov_DIRR_DATA( Simikon8052 * );
  static void sjmp( Simikon8052 * );
  static void anl_C_BADDR( Simikon8052 * );
  static void movc_PC( Simikon8052 * );
  static void div( Simikon8052 * );
  static void mov_IRAM_IRAM( Simikon8052 * );
  static void mov_IRAM_INDR( Simikon8052 * );
  static void mov_IRAM_DIRR( Simikon8052 * );
  static void mov_DPTR_DATA( Simikon8052 * );
  static void mov_BADDR_C( Simikon8052 * );
  static void movc_DPTR( Simikon8052 * );
  static void subb_DATA( Simikon8052 * );
  static void subb_IRAM( Simikon8052 * );
  static void subb_INDR( Simikon8052 * );
  static void subb_DIRR( Simikon8052 * );
  static void orl_C_bADDR( Simikon8052 * );
  static void mov_C_BADDR( Simikon8052 * );
  static void inc_DPTR( Simikon8052 * );
  static void mul_AB( Simikon8052 * );
  static void undefind( Simikon8052 * );
  static void mov_INDR_IRAM( Simikon8052 * );
  static void mov_DIRR_IRAM( Simikon8052 * );
  static void anl_C_bADDR( Simikon8052 * );
  static void cpl_BADDR( Simikon8052 * );
  static void cpl_C( Simikon8052 * );
  static void cjne_A_DATA( Simikon8052 * );
  static void cjne_A_IRAM( Simikon8052 * );
  static void cjne_INDR_DATA( Simikon8052 * );
  static void cjne_DIRR_DATA( Simikon8052 * );
  static void push( Simikon8052 * );
  static void clr_BADDR( Simikon8052 * );
  static void clr_C( Simikon8052 * );
  static void swap( Simikon8052 * );
  static void xch_IRAM( Simikon8052 * );
  static void xch_A_INDR( Simikon8052 * );
  static void xch_A_DIRR( Simikon8052 * );
  static void pop( Simikon8052 * );
  static void setb_BADDR( Simikon8052 * );
  static void setb_C( Simikon8052 * );
  static void da( Simikon8052 * );
  static void djnz_IRAM( Simikon8052 * );
  static void xchd( Simikon8052 * );
  static void djnz_DIRR( Simikon8052 * );
  static void movx_A_INDDPTR( Simikon8052 * );
  static void movx_A_INDR( Simikon8052 * );
  static void clr_A( Simikon8052 * );
  static void mov_A_IRAM( Simikon8052 * );
  static void mov_A_INDR( Simikon8052 * );
  static void mov_A_DIRR( Simikon8052 * );
  static void movx_INDDPTR_A( Simikon8052 * );
  static void movx_INDR_A( Simikon8052 * );
  static void cpl_A( Simikon8052 * );
  static void mov_IRAM_A( Simikon8052 * );
  static void mov_INDR_A( Simikon8052 * );
  static void mov_DIRR_A( Simikon8052 * );

};

#endif // SIMIKON_8052_HH
