/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "config.h"

#ifdef LEAK_TRACKER

// STL includes
#include <iostream>
#include <fstream>
#include <map>      // STL hash map

#include <stdlib.h>

typedef struct {
  long address;
  size_t size;
  char file[64];
  size_t line;
} ALLOC_INFO;

struct lstr { 
  bool operator() ( long n1, long n2 ) const {
    return n1 < n2;
  }
};

typedef map<long, ALLOC_INFO, lstr> ALLOC_MAP;
ALLOC_MAP allocMap;


inline void AddLeakTrack( long addr, size_t size, 
                          const char *fname, size_t line )
{
  ALLOC_INFO allocInfo;

  allocInfo.address = addr;
  allocInfo.size    = size;
  allocInfo.line    = line;
  strncpy( allocInfo.file, fname, 63 );

  allocMap[ addr ] = allocInfo;
}

inline void RemoveLeakTrack( long addr )
{
  allocMap.erase( addr );
}

inline void DumpMemoryLeaks( void )
{
  ALLOC_MAP::iterator it  = allocMap.begin();
  ALLOC_MAP::iterator end = allocMap.end();

  ofstream fs;
  char *file;

  if ( (file = getenv( LEAK_DUMP_ENV ) ) == NULL )
    file = LEAK_DUMP_FILE;

  fs.open( file );

  if ( !fs.is_open() ) {
    cerr << "LeakTracker: Can't open dump file!" << endl;
    return;
  }

  long   totalUnfreed = 0;

  fs << "MEMORY LEAKS" << endl;
  fs << endl;
  fs << "| FILE              | LINE |   ADDRESS |    SIZE |" << endl;
  fs << "--------------------------------------------------" << endl;

  do {
    long i = (*it).first;

    ALLOC_INFO alloc = allocMap[ i ];

    if ( alloc.size == 0 ) break;

    fs << alloc.file;
    fs.fill( ' ' );
    fs.width( 15 );
    fs << alloc.line;
    fs << "  ";
    fs.fill( ' ' );
    fs.width( 10 );
    fs << alloc.address;
    fs.fill( ' ' );
    fs.width( 10 );
    fs << alloc.size << endl;

    totalUnfreed += alloc.size;

  } while ( ++it != end );

  fs << "--------------------------------------------------" << endl;  

  fs << endl << "Total unfreed: " << totalUnfreed << " bytes." << endl;

  fs.close();
}

// overload OPERATOR new and new[]
void *operator new( size_t size, const char *file, size_t line )
{
  void *ptr = ( void * ) malloc( size );
  AddLeakTrack( (long)ptr, size, file, line );
  return ptr;
}

void *operator new[]( size_t size, const char *file, size_t line )
{
  void *ptr = ( void * ) malloc( size );
  AddLeakTrack( (long)ptr, size, file, line );
  return ptr;
}

// overload OPERATOR delete and delete[]
void operator delete( void *ptr )
{
  RemoveLeakTrack( (long)ptr );
  free( ptr );
}

void operator delete[]( void *ptr )
{
  RemoveLeakTrack( (long)ptr );
  free( ptr );
}

static class LeakTracker
{
public:
  LeakTracker() {
    cout << "LeakTracker v1.0, Copyright (c) 2000 Alen Kovac" << endl;
    cout << endl;
  }
  
  ~LeakTracker() {
    cout << "LeakTracker: Dumping memory leaks!!!" << endl;
    DumpMemoryLeaks();
  }
  
} tracker;

#endif // LEAK_TRACKER
