/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "Register.hh"
#include "SimikonUtils.hh"
#include "LeakTracker.hh"

// QT_INCLUDES
#include <qcheckbox.h>

/*===========================================================================*/
/* class: RegisterView                                                       */
/*===========================================================================*/

RegisterView::RegisterView( QWidget *parent, const char *name )
  : QTable( parent, name )
{
  setNumRows( 8 );
  setNumCols( 3 );

  setShowGrid( false );

  QHeader *header;

  // horizontal header stuff
  header = horizontalHeader();

  header->setLabel( 0, tr( "Name" ) );
  header->setLabel( 1, tr( "Addr" ) );
  header->setLabel( 2, "" );
  header->setResizeEnabled( false );
  header->setMovingEnabled( false );
  header->resizeSection( 0, 45 );
  header->resizeSection( 1, 42 );
  header->resizeSection( 2, 12 );

  // vertical header stuff
  header = verticalHeader();

  header->setResizeEnabled( false );
  header->setMovingEnabled( false );

  for ( uint i = 0; i < 8; i++ ) {
    header->setLabel( i, QString::number( i ) );
  }

  for ( uint i = 0; i < 8; i++ ) {
    check[i] = new QCheckBox( this );
    connect( check[i], SIGNAL( toggled( bool ) ),
             this, SLOT( SlotCheckToggled( bool ) ) );
    setCellWidget( i, 2, check[i] );
  }

  setLeftMargin( 15 );
}

RegisterView::~RegisterView()
{

}

bool
RegisterView::bitState( uint bit )
{
  return check[ bit ]->isChecked();
}

void
RegisterView::slotAddBit( uint bit, const QString &name,
                                    const QString &address,
                          bool value )
{
  setText( bit, 0, name );
  setText( bit, 1, "0x" + address + "h" );

  disconnect( check[bit], SIGNAL( toggled( bool ) ),
              this, SLOT( SlotCheckToggled( bool ) ) );

  check[bit]->setChecked( value );

  connect( check[bit], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotCheckToggled( bool ) ) );
}

void
RegisterView::slotUpdateBit( uint bit, bool state )
{
  disconnect( check[bit], SIGNAL( toggled( bool ) ),
              this, SLOT( SlotCheckToggled( bool ) ) );

  check[bit]->setChecked( state );

  connect( check[bit], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotCheckToggled( bool ) ) );
}

void
RegisterView::SlotCheckToggled( bool state )
{
  for ( uint i = 0; i < 8; i++ ) {
    if ( check[i] == sender() ) {
      emit signalBitToggled( i, state );
      break;
    }
  }
}

QWidget *
RegisterView::beginEdit( int, int, bool )
{
  return 0;
}

/*===========================================================================*/
/* class: RegistersView                                                      */
/*===========================================================================*/

RegistersView::RegistersView( QWidget *parent, const char *name )
  : QTable( parent, name )
{
  setNumRows( 1 );
  setNumCols( 3 );

  // hide first (dummy) row
  hideRow( 0 );

  setCurrentCell( 1, 0 );

  QHeader *header = horizontalHeader();

  header->setLabel( 0, tr( "Register" ) );
  header->setLabel( 1, tr( "Addr" ) );
  header->setLabel( 2, tr( "Value" ) );
  header->resizeSection( 0, 58 );
  header->resizeSection( 1, 42 );
  header->resizeSection( 2, 50 );

  header->setResizeEnabled( false );
  header->setMovingEnabled( false );

  setShowGrid( false );
  setColumnMovingEnabled( false );
  setRowMovingEnabled( false );

  setSelectionMode( Single );

  // get rid off the vertical header
  setLeftMargin( 0 );
  verticalHeader()->hide();

  connect( this, SIGNAL( currentChanged( int, int ) ), 
           this, SLOT( SlotRowSelected( int, int) ) );
  connect( this, SIGNAL( valueChanged( int, int ) ),
           this, SLOT( SlotRowEdited( int, int ) ) );

}

RegistersView::~RegistersView()
{

}

void
RegistersView::setAddress( uint reg, const QString &address )
{
  setText( reg+1, 1, address );
}

void
RegistersView::slotUpdate( uint reg, const QString &value )
{
  setText( reg+1, 2, "0x" + value );
}

void
RegistersView::slotAdd( const QString &name, const QString &address )
{
  int row = numRows();

  setNumRows( row + 1 );
  setText( row, 0, name );
  setText( row, 1, address );
}

void
RegistersView::SlotRowSelected( int row, int )
{
  if ( row == 0 ) {
    setCurrentCell( 1, 0 );
    return;
  }

  emit signalSelected( text( row, 0 ) );
}

void
RegistersView::SlotRowEdited( int row, int col )
{
  uint value = strToUInt( text( row, col ) );
  emit signalValueChanged( row-1, value );
}

QWidget *
RegistersView::beginEdit( int row, int col, bool replace )
{
  editRow = row;
  editCol = col;

  return ( ( col <= 1 ) ? 0 : QTable::beginEdit( row, col, replace ) );
}

bool
RegistersView::eventFilter( QObject *obj, QEvent *e )
{
  if ( e->type() == QEvent::KeyPress ) {
    QKeyEvent *ke = ( QKeyEvent * ) e;
    
    if ( ke->key() == Key_Return || ke->key() == Key_Enter ) {
      endEdit( editRow, editCol, TRUE, FALSE );
      return true;
    }
  }

  return QTable::eventFilter( obj, e );
}
