/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "SimikonActions.hh"
#include "LeakTracker.hh"

// QT_INCLUDES
#include <iostream.h>
#include <qpixmap.h>
#include <qpopupmenu.h>
#include <qmenubar.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>


/*===========================================================================*/
/* class: SimikonActions                                                     */
/*===========================================================================*/

SimikonActions::SimikonActions()
{
  actions = new QDict<ActionData>;
  actions->setAutoDelete( true );

  globalId = 0;
}

SimikonActions::~SimikonActions()
{
  // we don't want memory leaks, do we?
  QDictIterator<ActionData> it( *actions );

  while ( ActionData *adata = it.current() ) {
    if ( adata->pix ) 
      delete adata->pix;
    ++it;
  }

  delete actions;
}

void
SimikonActions::addAction( const char *name, const QString &label,
                           const char **pixmap,
                           const QObject *receiver, const char *slot,
                           int accel, const QString &tooltiptext,
                           bool isToggle )
{
  ActionData *adata = new ActionData;

  adata->id          = globalId++;
  adata->accel       = accel;
  adata->pix         = ( !pixmap ) ? 0L : new QPixmap( pixmap );
  adata->label       = label;
  adata->enabled     = true;
  adata->tooltiptext = tooltiptext;
  adata->receiver    = receiver;
  adata->slot        = slot;
  adata->toolbutton  = 0;
  adata->menubar     = 0;
  adata->popupmenu   = 0;
  adata->isToggle    = isToggle;

  actions->replace( name, adata );
}

void
SimikonActions::setActionsEnabled( const char *actions, bool enabled )
{
  QString names = QString( actions ).stripWhiteSpace() += " ";

  while ( names.find( " " ) != -1 ) {
    QString name = names.left( names.find( " " ) );
    names.remove( 0, names.find( " " ) + 1 );

    if ( name != "|" ) {
      SimikonActions::setActionEnabled( name, enabled );
    }
  }
}

void
SimikonActions::setActionEnabled( const char *action, bool enabled )
{
  ActionData *adata = SimikonActions::FindActionDataByName( action );

  if ( !adata ) return;

  adata->enabled = enabled;

  if ( adata->popupmenu ) {
    if ( adata->popupmenu->isItemChecked( adata->id ) ) {
      adata->popupmenu->setItemChecked( adata->id, false );
    }

    adata->popupmenu->setItemEnabled( adata->id, enabled );
  }
  
  if ( adata->toolbutton ) {
    if ( enabled == false ) {
      adata->toolbutton->setOn( false );
    }

    adata->toolbutton->setEnabled( enabled );
  }
}

bool
SimikonActions::isActionEnabled( const char *action )
{
  ActionData *adata = SimikonActions::FindActionDataByName( action );

  if ( !adata ) return false;

  return adata->enabled;
}

void
SimikonActions::setActionsChecked( const char *actions, bool checked )
{
  QString names = QString( actions ).stripWhiteSpace() += " ";

  while ( names.find( " " ) != -1 ) {
    QString name = names.left( names.find( " " ) );
    names.remove( 0, names.find( " " ) + 1 );

    if ( name != "|" ) {
      setActionChecked( name, checked );
    }
  }
}

void
SimikonActions::setActionChecked( const char *action, bool checked )
{
  ActionData *adata = SimikonActions::FindActionDataByName( action );

  if ( !adata ) return;

  if ( adata->popupmenu ) {
    adata->popupmenu->setItemChecked( adata->id, checked );
  }

  if ( adata->toolbutton ) {
    adata->toolbutton->setOn( checked );
  }
}

bool
SimikonActions::isActionChecked( const char *action )
{
  ActionData *adata = FindActionDataByName( action );

  if ( !adata ) return false;

  if ( adata->popupmenu ) {
    return adata->popupmenu->isItemChecked( adata->id );
  }

  if ( adata->toolbutton ) {
    return adata->toolbutton->isDown();
  }

  return false;
}

void
SimikonActions::addMenuItems( QMenuBar *menubar, const char *data )
{
  QString names = QString( data ).stripWhiteSpace() += " ";
  
  while ( names.find( " " ) != -1 ) {
    QString name = names.left( names.find( " " ) );
    names.remove( 0, names.find( " " ) + 1 );

    if ( name == "|" ) {
      menubar->insertSeparator();
    } else {
      ActionData *adata = SimikonActions::FindActionDataByName( name );
      
      if ( adata ) {
        if ( adata->receiver ) {
          menubar->insertItem( name, adata->receiver, adata->slot, 
                               adata->accel, adata->id );
        } else {
          menubar->insertItem( name, adata->id );
        }
        adata->menubar = menubar;
      }
    }
  }  
}

void
SimikonActions::addPopupMenu( QMenuBar *menubar, const char *name,
                              const char *data )
{
  menubar->insertItem( name, CreateQPopupMenu( menubar, name, data ) );
}

void
SimikonActions::addToolButtons( QToolBar *toolbar, const char *data )
{
  QString names = QString( data ).stripWhiteSpace() += " ";

  while ( names.find( " " ) != -1 ) {
    QString name = names.left( names.find( " " ) );
    names.remove( 0, names.find( " " ) + 1 );

    if ( name == "|" ) {
      toolbar->addSeparator();
    } else {
      ActionData *adata = FindActionDataByName( name );

      if ( adata ) {
        if ( adata->pix ) {
          if ( adata->receiver ) {
            adata->toolbutton =
              new QToolButton( *adata->pix, adata->tooltiptext, QString::null,
                               ( QObject * ) adata->receiver, adata->slot, 
                               toolbar );
          } else {
            adata->toolbutton = new QToolButton( toolbar );
            
            adata->toolbutton->setPixmap( *adata->pix );
            adata->toolbutton->setTextLabel( adata->tooltiptext, true );
          }

          adata->toolbutton->setToggleButton( adata->isToggle );
          
          if ( adata->isToggle ) {
            connect( adata->toolbutton, SIGNAL( toggled( bool ) ),
                     this, SLOT( SlotToolItemToggled( bool ) ) );
          }
        } else {
          cout << "WARNING: SimikonActions::addPopupMenu(): "
               << "Action doesn't have pixmap! \n";
        }
      }
    }
  }
 
  // Ugly Hack
  //  QWidget *dummy = new QWidget( toolbar, "DummyWidget" );
  //  toolbar->setStretchableWidget( dummy );
}

void
SimikonActions::addSubPopupMenu( const char *name, const char *data )
{
  ActionData *adata = SimikonActions::FindActionDataByName( name );

  ActionData *adataItem = FindActionDataById( adata->id - 1 );
  int index = adataItem->popupmenu->indexOf( adataItem->id ) + 1;

  adata->popupmenu = adataItem->popupmenu;
  adata->menubar   = adataItem->menubar;

  QPopupMenu *popup = CreateQPopupMenu( adata->popupmenu, name, data );

  adataItem->popupmenu->insertItem( *adata->pix, name, popup, adata->id,
                                    index );
}

QPopupMenu *
SimikonActions::CreateQPopupMenu( QWidget    *parent, 
                                  const char *caption, const char *data )
{
  QPopupMenu *menu = new QPopupMenu( parent );

  connect( menu, SIGNAL( activated( int ) ),
           this, SLOT( SlotMenuItemActivated( int ) ) );

  QString names = QString( data ).stripWhiteSpace() += " ";

  while ( names.find( " " ) != -1 ) {
    QString name = names.left( names.find( " " ) );
    names.remove( 0, names.find( " " ) + 1 );

    if ( name == "|" ) {
      menu->insertSeparator();
    } else if ( name == "~" ) {
      menu->insertTearOffHandle();

      QString realCaption( caption );

      if ( realCaption[ 0 ] == '&' ) {
        realCaption.remove( 0, 1 ); // remove '&' character
      }

      menu->setCaption( realCaption );

    } else {
      ActionData *adata = FindActionDataByName( name );

      if ( adata ) {
        if ( !adata->label.isEmpty() && !adata->label.isNull() ) {
          if ( adata->pix ) {
            menu->insertItem( *adata->pix, adata->label,
                              adata->receiver, adata->slot, 
                              adata->accel, adata->id );
          } else {
            menu->insertItem( adata->label,
                              adata->receiver, adata->slot, 
                              adata->accel, adata->id );
          }
          menu->setItemEnabled( adata->id, adata->enabled );
        }
        adata->popupmenu = menu;
      }
    }
  }

	return menu;
}

void
SimikonActions::setActionReceiver( const char *action, const QObject *receiver )
{
  ActionData *adata = SimikonActions::FindActionDataByName( action );

  if ( !adata ) return;
  if ( adata->receiver == receiver ) return;

  removeActionReceiver( action );

  // connect all
  adata->receiver = receiver;

  if ( adata->menubar ) {
    adata->menubar->connectItem( adata->id, adata->receiver, adata->slot );
  } else if ( adata->popupmenu ) {
    adata->popupmenu->connectItem( adata->id, adata->receiver, adata->slot );
  }

  if ( adata->toolbutton ) {
    connect( adata->toolbutton, SIGNAL( clicked() ), 
             adata->receiver, adata->slot );
  }
}

void
SimikonActions::removeActionReceiver( const char *action )
{
  ActionData *adata = FindActionDataByName( action );

  if ( !adata ) return;
  if ( !adata->receiver ) return;
  
  if ( adata->menubar ) {
    adata->menubar->disconnectItem( adata->id, adata->receiver, adata->slot );
  } else if ( adata->popupmenu ) {
    adata->popupmenu->disconnectItem( adata->id, adata->receiver, adata->slot );
  }

  if ( adata->toolbutton ) {
    disconnect( adata->toolbutton, 0, adata->receiver, 0 );
  }

  adata->receiver = 0;
}

void
SimikonActions::changeActionText( const char *action, const char *text )
{
  ActionData *adata = SimikonActions::FindActionDataByName( action );

  if ( !adata ) return;

  if ( adata->label.isEmpty() || adata->label.isNull() ) {
    if ( adata->pix ) {
      adata->popupmenu->insertItem( *adata->pix, text,
                        adata->receiver, adata->slot, 
                        adata->accel, adata->id );
    } else {
      adata->popupmenu->insertItem( text,
                        adata->receiver, adata->slot, 
                        adata->accel, adata->id );
    }
  } else {
    adata->popupmenu->changeItem( adata->id, text );
  }
  
  adata->label = text;
}

ActionData *
SimikonActions::FindActionDataByName( QString name )
{
  ActionData *adata = actions->find( name );

  if ( !adata ) {
    cout << "WARNING: SimikonActions:: Cannot find action by name: "
         << name.ascii() << endl;
  }

  return adata;
}

ActionData *
SimikonActions::FindActionDataById( int id )
{
  QDictIterator<ActionData> it( *actions );

  if ( id >= 0 ) {
    while ( it.current() ) {
      ActionData *adata = it.current();
      
      if ( adata->id == id ) {
        return adata;
      }

      ++it;
    }
    cout << "WARNING: SimikonActions:: Cannot find action by id: " 
         << id << endl;
  }

  return 0;
}

ActionData *
SimikonActions::FindActionDataByToolButton( QToolButton *toolbutton )
{
  QDictIterator<ActionData> it( *actions );

  while ( it.current() ) {
    ActionData *adata = it.current();

    if ( adata->toolbutton == toolbutton ) {
      return adata;
    }

    ++it;
  }

  cout << "WARNING: SimikonActions:: can't find action by tool button! \n";
  
  return 0;
}

void
SimikonActions::SlotMenuItemActivated( int id )
{
  const QObject *obj = sender();
  ActionData *adata = FindActionDataById( id );
  
  if ( adata ) {
    if ( adata->isToggle ) {
      bool state = ( ( QMenuBar * ) obj )->isItemChecked( id );
      ( ( QMenuBar * ) obj )->setItemChecked( id, !state );

      if ( adata && adata->toolbutton ) {
        adata->toolbutton->setOn( !state );
      }
    }
  } else {
    //    cout << "Tear of handle!!!\n";
  }
}

void
SimikonActions::SlotToolItemToggled( bool state )
{
  const QObject *obj = sender();

  ActionData *adata = FindActionDataByToolButton( ( QToolButton * ) obj );

  if ( adata && adata->popupmenu ) {
    adata->popupmenu->setItemChecked( adata->id, state );
  }
}
