/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "config.h"
#include "SimikonApp.hh"
#include "SimikonActions.hh"
#include "SimikonProject.hh"
#include "SimikonProperties.hh"
#include "SimikonSplash.hh"
#include "Simikon8052.hh"
#include "LeakTracker.hh"

// QT_INCLUDES
#include <qapplication.h>
#include <qmenubar.h>
#include <qtoolbar.h>
#include <qstatusbar.h>
#include <qsplitter.h>
#include <qpushbutton.h>
#include <qprogressbar.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qvaluelist.h>

// MISC_INCLUDES
#include <iostream.h>

// PIXMAPS
#include "project.xpm"
#include "project_settings.xpm"
#include "open.xpm"
#include "save.xpm"
#include "properties.xpm"
#include "performance.xpm"
//#include "dissembler.xpm"
#include "registers.xpm"
#include "sfr.xpm"
#include "memory.xpm"
#include "stack.xpm"
#include "info.xpm"
#include "serial.xpm"
#include "interrupts.xpm"
#include "io.xpm"
#include "stop.xpm"
#include "start.xpm"
#include "assemble.xpm"

// GLOBAL_OBJECTS
extern SimikonActions  *simikonActions;
extern SimikonSplash   *simikonSplash;
extern SimikonSettings *simikonSettings;
extern QProgressBar    *progress;
extern QLabel          *cursorLabel;
extern QLineEdit       *searchEntry;

/*===========================================================================*/
/* class: SimikonApp                                                         */
/*===========================================================================*/

SimikonApp::SimikonApp()
  : QMainWindow( 0, PROGRAM_NAME )
{
  setCaption( PROGRAM_NAME );

  InitSimikonProperties(); /* Init default settings if there isn't 
                            * .simikonrc file.
                            */
  InitSimikonProject();
  InitSimikonActions();
  InitSimikonMenuBar();
  InitSimikonToolBar();
  InitSimikonStatusBar();
  InitSimikonConsole();
  InitSimikonProjects();

  if ( SS_GET_BOOL( "remember_latest_projects" ) )
    InitLatestProjectsMenu();

  SA_ENABLED( "Settings Save Close "
              "Registers SFR Memory Stack Performance IOPorts Serial "
              "Interrupts Assemble ", false );

  SA_ENABLED( "Stop Reset StepInto StepOut StepOver "
              "GotoBrkpt AssembleEditor", false );
  
  if ( SS_GET_BOOL( "open_last_project" ) ) {

    QString prjFile = SS_GET( "last_project" );

    if ( !prjFile.isEmpty() )
      simikonProject->openProject( prjFile );
  }  

  // resize window
  int width  = SS_GET_INT( "window_width" );
  int height = SS_GET_INT( "window_height" );

  resize( width, height );

  // resize splitter frames
  int debugger_frame = SS_GET_INT( "debugger_frame" );
  int project_frame  = SS_GET_INT( "project_frame" );

  if ( debugger_frame && project_frame ) {
    QValueList<int> sizes;
    sizes.append( debugger_frame );
    sizes.append( project_frame );
    simikonProject->setSizes( sizes );
  }
}

SimikonApp::~SimikonApp()
{
  simikonProject->closeAll();
}

// PUBLIC_MEMBER_FUNCTIONS

// PRIVATE_MEMBER_FUNCTIONS

void
SimikonApp::InitSimikonProperties( void )
{
  SPLASH_MSG( tr( "Initializing SimikonProperties..." ) );

  simikonProperties = new SimikonProperties( this, "SimikonProperties" );
  CHECK_PTR( simikonProperties );
  simikonProperties->hide();
}

void
SimikonApp::InitSimikonProject( void )
{
  SPLASH_MSG( tr( "Initializing SimikonProject..." ) );

  simikonProject = new SimikonProject( this, "SimikonProject" );
  CHECK_PTR( simikonProject );
  setCentralWidget( simikonProject );

  connect( simikonProject, SIGNAL( signalStatusMessage( const char * ) ),
           this, SLOT( SlotStatusMessage( const char * ) ) );
}

void
SimikonApp::InitSimikonActions( void )
{
  SPLASH_MSG( tr( "Initializing SimikonActions..." ) );

  // FILE_MENU ACTIONS
  SA_ADD_ACTION( "Open", tr( "&Open" ), open_xpm,
                 simikonProject, SLOT( slotOpenProject() ), ALT+Key_O,
                 "Open project", false );

  SA_ADD_ACTION( "Reopen", tr( "&Reopen" ), open_xpm,
                 0, 0, ALT+Key_R,
                 "Open Latest projects", false );

  SA_ADD_ACTION( "New", tr( "&New" ), project_xpm,
                 simikonProject, SLOT( slotNewProject() ), ALT+Key_N,
                 "New project", false );

  SA_ADD_ACTION( "Save", tr( "&Save" ), save_xpm,
                 simikonProject, SLOT( slotSaveProject() ), ALT+Key_S,
                 "Save active project", false );

  SA_ADD_ACTION( "Close", tr( "&Close" ), 0,
                 simikonProject, SLOT( slotCloseProject() ), ALT+Key_C,
                 "Close active project", false );

  SA_ADD_ACTION( "Settings", tr( "Se&ttings" ), project_settings_xpm,
                 simikonProject, SLOT( slotProjectSettings() ), ALT+Key_T,
                 "Change active project settings.", false );

  SA_ADD_ACTION( "Properties", tr( "&Properties" ), properties_xpm,
                 this, SLOT( SlotProperties() ), ALT+Key_P,
                 "Simikon Properties", false );

  SA_ADD_ACTION( "Quit", tr( "&Quit" ), 0,
                 this, SLOT( slotQuit() ), ALT+Key_Q,
                 "Exit from simikon", false );

  // VIEW_MENU ACTIONS
  SA_ADD_ACTION( "Toolbar", tr( "Toolbar" ), 0,
                 this, SLOT( SlotToggleToolBar() ), 0,
                 "Show/Hide Toolbar", true );

  SA_ADD_ACTION( "Statusbar", tr( "Statusbar" ), 0,
                 this, SLOT( SlotToggleStatusBar() ), 0,
                 "Show/Hide Statusbar", true );
  
  SA_ADD_ACTION( "Console", tr( "Console" ), 0,
                 simikonProject, SLOT( slotToggleConsole() ), 0,
                 "Show/Hide SimikonConsole", true );

  SA_ADD_ACTION( "Projects", tr( "Projects" ), 0,
                 simikonProject, SLOT( slotToggleProjects() ), 0,
                 "Show/Hide Projects Manager", true );

  // MODULES_MENU ACTIONS
  SA_ADD_ACTION( "Registers", tr( "Registers" ), registers_xpm,
                 simikonProject, SLOT( slotToggleRegisters() ), CTRL+Key_R,
                 "Show/Hide Registers window", true );
                     
  SA_ADD_ACTION( "SFR", tr( "SFR" ), sfr_xpm,
                 simikonProject, SLOT( slotToggleSFR() ), CTRL+Key_F,
                 "Show/Hide SFR window", true );

  SA_ADD_ACTION( "Memory", tr( "Memory" ), memory_xpm,
                 simikonProject, SLOT( slotToggleMemory() ), CTRL+Key_M,
                 "Show/Hide Memory window", true );

  SA_ADD_ACTION( "Stack", tr( "Stack" ), stack_xpm,
                 simikonProject, SLOT( slotToggleStack() ), CTRL+Key_S,
                 "Show/Hide Stack window", true );
  
  SA_ADD_ACTION( "Performance", tr( "Performance" ), performance_xpm,
                 simikonProject, SLOT( slotTogglePerformance() ), CTRL+Key_P,
                 "Show/Hide Performance window", true );

  SA_ADD_ACTION( "IOPorts", tr( "IO Ports" ), io_xpm,
                 simikonProject, SLOT( slotToggleIOPorts() ), CTRL+Key_O,
                 "Show/Hide IO Ports window", true );

  SA_ADD_ACTION( "Serial", tr( "Serial" ), serial_xpm,
                 simikonProject, SLOT( slotToggleSerial() ), CTRL+Key_E,
                 "Show/Hide Serial window", true );

  SA_ADD_ACTION( "Interrupts", tr( "Interrupts" ), interrupts_xpm,
                 simikonProject, SLOT( slotToggleInterrupts() ), CTRL+Key_I,
                 "Show/Hide Interrupts window", true );

  // PROJECT_MENU ACTIONS

  SA_ADD_ACTION( "StartSim", tr( "Start Simulation" ), start_xpm,
                 simikonProject, SLOT( slotStartSimulator() ), Key_F1,
                 "Simulate", false );

  SA_ADD_ACTION( "Stop", tr( "Stop" ), stop_xpm,
                 simikonProject, SLOT( slotStop() ), Key_F3,
                 "Stop", false );

  SA_ADD_ACTION( "Assemble", tr( "Assemble" ), assemble_xpm,
                 simikonProject, SLOT( slotAssembleProject() ), Key_F9,
                 "Assemble project", false );

  // HELP_MENU ACTIONS
  SA_ADD_ACTION( "About", tr( "About SiMiKON" ), info_xpm,
                 simikonProject, SLOT( slotAboutSimikon() ), ALT+Key_F1,
                 "About Simikon dialog", false );

  SA_ADD_ACTION( "InstructionSetHelp", tr( "Instruction Set Manual" ), 0,
                 simikonProject, SLOT( slotInstructionSetHelp() ), ALT+Key_F2,
                 "8051 Instruction Set Manual", true );
}

void
SimikonApp::InitSimikonMenuBar( void )
{
  SPLASH_MSG( tr( "Initializing Menubar..." ) );

  menubar = new QMenuBar( this, "SimikonMenuBar" );

  SA_ADD_MENU( menubar, tr( "&File" ), "~ Open New | Save Close | Properties | Quit |" );
  SA_ADD_MENU( menubar, tr( "&View" ), "~ Toolbar Statusbar Console Projects |" );
  SA_ADD_MENU( menubar, tr( "&Project" ), "~ StartSim StartDebug Stop |"
               " StepInto StepOut StepOver GotoBrkpt Reset | Assemble |"
               " Settings |" );
  SA_ADD_MENU( menubar, tr( "&Modules" ), "~ Registers SFR | Stack Memory |"
               " IOPorts Serial Interrupts | Performance |" );
  SA_ADD_MENU( menubar, tr( "&Help" ), "~ About | InstructionSetHelp |" );
}

void
SimikonApp::InitSimikonToolBar( void )
{
  SPLASH_MSG( tr( "Initializing Toolbar..." ) );

  // MAIN TOOLBAR
  toolbar = new QToolBar( this, "SimikonToolBar" );
  SA_ADD_TOOLBUTTONS( toolbar, "Open New Settings" );

  // MODULES TOOLBAR
  toolbarModules = new QToolBar( this, "SimikonModulesToolBar" );
  SA_ADD_TOOLBUTTONS( toolbarModules, "Registers SFR Stack Memory IOPorts "
                      "Serial Interrupts Performance" );

  // DEBUGGER TOOLBAR
  toolbarDebugger = new QToolBar( this, "SimikonDebuggerToolBar" );
  SA_ADD_TOOLBUTTONS( toolbarDebugger, "StartSim StartDebug Stop | " 
                      "StepInto StepOut StepOver GotoBrkpt Reset | "
                      "AssembleEditor" );

  if ( SS_GET_BOOL( "show_toolbar" ) ) {
    SA_CHECKED( "Toolbar", true );
  } else {
    SlotToggleToolBar();
  }
}

void
SimikonApp::InitSimikonConsole( void )
{
  if ( SS_GET_BOOL( "show_console" ) ) {
    SA_CHECKED( "Console", true );
  } else {
    simikonProject->slotToggleConsole();
  }
}

void
SimikonApp::InitSimikonProjects( void )
{
  if ( SS_GET_BOOL( "show_project" ) ) {
    SA_CHECKED( "Projects", true );
  } else {
    simikonProject->slotToggleProjects();
  }
}

void
SimikonApp::InitSimikonStatusBar( void )
{
  statusbar = new QStatusBar( this, "SimikonStatusbar" );

  /*
  // SEARCH_ENTRY
  searchEntry = new QLineEdit( statusbar );

  QLabel *label = new QLabel( searchEntry, " Search: ", statusbar );
  statusbar->addWidget( label, 0, true );
  statusbar->addWidget( searchEntry, 0, true );

  connect( searchEntry, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotSearch( const QString & ) ) );
  connect( searchEntry, SIGNAL( returnPressed() ),
           this, SLOT( SlotSearchNext() ) );

  statusbar->addWidget( searchEntry );
  */

  // PROGRESSBAR
  progress = new QProgressBar( statusbar );
  statusbar->addWidget( progress, 40 );

  // CURSOR_LABEL
  cursorLabel = new QLabel( "L0:C0  ", statusbar );
  cursorLabel->setAlignment( AlignVCenter | AlignRight );
  statusbar->addWidget( cursorLabel, 5, true );

  if ( SS_GET_BOOL( "show_statusbar" ) ) {
    SA_CHECKED( "Statusbar", true );
  } else {
    SlotToggleStatusBar();
  }
}

void
SimikonApp::InitLatestProjectsMenu( void )
{  
  SA_ADD_ACTION( "Project0", SS_GET( "Project0" ), 0, simikonProject,
                 SLOT( slotOpenProject0() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project1", SS_GET( "Project1" ), 0, simikonProject,
                 SLOT( slotOpenProject1() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project2", SS_GET( "Project2" ), 0, simikonProject,
                 SLOT( slotOpenProject2() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project3", SS_GET( "Project3" ), 0, simikonProject,
                 SLOT( slotOpenProject3() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project4", SS_GET( "Project4" ), 0, simikonProject,
                 SLOT( slotOpenProject4() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project5", SS_GET( "Project5" ), 0, simikonProject,
                 SLOT( slotOpenProject5() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project6", SS_GET( "Project6" ), 0, simikonProject,
                 SLOT( slotOpenProject6() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project7", SS_GET( "Project7" ), 0, simikonProject,
                 SLOT( slotOpenProject7() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project8", SS_GET( "Project8" ), 0, simikonProject,
                 SLOT( slotOpenProject8() ), 0, QString( "" ), false  );
  SA_ADD_ACTION( "Project9", SS_GET( "Project9" ), 0, simikonProject,
                 SLOT( slotOpenProject9() ), 0, QString( "" ), false  );


  SA_ADD_SUB_MENU( "Reopen", "Project0 Project1 Project2 Project3 "
                   "Project4 Project5 Project6 Project7 Project8 Project9" );
}

// PROTECTED
bool
SimikonApp::eventFilter( QObject *o, QEvent *e )
{
  if ( e->type() == QEvent::Close ) {
    qApp->quit();
    return true;
  }

  return QMainWindow::eventFilter( o, e );
}

// PUBLIC_SLOTS

// PRIVATE_SLOTS

void
SimikonApp::SlotProperties( void )
{
  simikonProperties->okButton->setEnabled( false );

  simikonProperties->exec();
}

void
SimikonApp::SlotToggleToolBar( void )
{
  bool state = SA_IS_CHECKED( "Toolbar" );

  if ( state ) {
    toolbar->show();
    toolbarDebugger->show();
    toolbarModules->show();
  } else {
    toolbar->hide();
    toolbarDebugger->hide();
    toolbarModules->hide();
  }

  SS_PUT( "show_toolbar", state );
}

void
SimikonApp::SlotToggleStatusBar( void )
{
  bool state = SA_IS_CHECKED( "Statusbar" );

  if ( state ) {
    statusbar->show();
  } else {
    statusbar->hide();
  }

  SS_PUT( "show_statusbar", state );
}

void
SimikonApp::slotQuit( void )
{
  // remeber main window size
  SS_PUT( "window_width",  width() );
  SS_PUT( "window_height", height() );

  QValueList<int> sizes = simikonProject->sizes();

  SS_PUT( "debugger_frame", sizes[0] );
  SS_PUT( "project_frame", sizes[1] );

  qApp->quit();
}

void
SimikonApp::SlotStatusMessage( const char *msg )
{
  statusbar->message( msg, 4000 );
}
