/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "config.h"
#include "SimikonDebugger.hh"
#include "SimikonActions.hh"
#include "Simikon8052.hh"
#include "SimikonUtils.hh"
#include "LeakTracker.hh"

// QT_INCLUDES
#include <qobjcoll.h>
#include <qlistview.h>
#include <qmessagebox.h>
#include <qevent.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qprogressbar.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qheader.h>

// WIDGET_INCLUDES
#include "stabctl.h"

// MISC_INCLUDES
#include <iostream.h>

// PIXMAPS
#include "start.xpm"
#include "restart.xpm"
#include "stepinto.xpm"
#include "stepout.xpm"
#include "stepover.xpm"
#include "gotobrkpt.xpm"
#include "dot.xpm"
#include "brkpt.xpm"
#include "assemble.xpm"

// DEFINES
//#define HELP_KEY (4144) // F1
#define HELP_KEY (4155) // F12
#define ESCAPE_KEY (4096)


// GLOBAL_OBJECTS
extern SimikonActions *simikonActions;
extern QProgressBar   *progress;
extern QLabel         *cursorLabel;
extern QLineEdit      *searchEntry;

/*===========================================================================*/
/* class: Disassembler                                                       */
/*===========================================================================*/

Disassembler::Disassembler( QWidget *parent, const char *name )
  : QListView( parent, name )
{
  addColumn( "BRK" );
  addColumn( "Address" );
  addColumn( "Symbol" );
  addColumn( "OpCode" );
  addColumn( "Cycles" );

  //  header()->setClickEnabled( false );
  header()->setMovingEnabled( false );
  header()->setResizeEnabled( false );

  setColumnWidth( 1, 100 );
  setColumnWidth( 2, 150 );
  setColumnWidth( 3, 100 );
  setColumnWidth( 4, 50 );

  setSorting( 1, true );
  setAllColumnsShowFocus( true );

  connect( this, SIGNAL( doubleClicked( QListViewItem * ) ),
           this, SLOT( SlotSetBreakPoint( QListViewItem * ) ) );

  activeLine    = 0;
  activeItem    = 0;
  nextItem      = 0;
  breakPointsNo = 0;

  valid = true;
}

Disassembler::~Disassembler()
{
}

void
Disassembler::reset( void )
{
  clear();

  activeLine = 0;
  activeItem = 0;
  nextItem   = 0;

  valid = true;
}

void
Disassembler::setBreakPoint( QListViewItem *i )
{
  SlotSetBreakPoint( i );
}

DisassemblerItem *
Disassembler::GetItem( uint line )
{
  QListViewItem *i = firstChild();
  
  while ( line > 0 ) {
    if ( i ) {
      i = i->nextSibling();
    } else {
      return 0;
    }
    
    line--;
  }
  
  return dynamic_cast<DisassemblerItem *>( i );
}

bool
Disassembler::eventFilter( QObject *o, QEvent *e )
{
  if ( !o || !e ) return false;

  // we intercept (ignore) header events.
  if ( o == header() && e->type() == QEvent::MouseButtonPress )
    return true;
  /*
  else if ( e->type() == QEvent::KeyPress )
    return true; */
  else 
    return QListView::eventFilter( o, e );
}

void
Disassembler::keyPressEvent( QKeyEvent * )
{
}

void
Disassembler::contentsMousePressEvent( QMouseEvent * )
{
}

void
Disassembler::slotAdd( const QString &address, const QString &mnemonic, 
                       const QString &opCode, const QString &cycles )
{
  new DisassemblerItem( this, address, mnemonic, opCode, cycles );
}

void
Disassembler::slotSelect( uint line )
{
  int diff = line - activeLine;

  if ( diff > 0 ) {
    QListViewItem *tmp = activeItem;
    
    while ( diff > 0 ) {
      if ( tmp ) {
        tmp = tmp->itemBelow();
      } else {
        // poruka o gresci
      }
      diff--;
    }
    
    if ( activeItem ) setSelected( activeItem, false );
    activeItem = tmp;

  } else if ( diff < 0 ) {
    QListViewItem *tmp = activeItem;
    
    while ( diff < 0 ) {
      if ( tmp ) {
        tmp = tmp->itemAbove();
      } else {
        // poruka o gresci
      }
      diff++;
    }

    if ( activeItem ) {
      setSelected( activeItem, false );
    }
    activeItem = tmp;
    
  } else if ( line == 0 ) {
    if ( activeItem ) {
      setSelected( activeItem, false );
    }
    activeItem = firstChild();
  }
  
  if ( activeItem ) {
    setSelected( activeItem, true );
    setCurrentItem( activeItem );
  }
  
  int y = itemPos( activeItem );
  center( 0, y );
 
  activeLine = line;
}

void
Disassembler::slotMark( uint line )
{
  DisassemblerItem *di = GetItem( line );

  if ( !di ) return;

  if ( di != nextItem ) {
    
    if ( nextItem ) {
      nextItem->isNext = false;
      nextItem->repaint();
    }

    di->isNext = true;
    nextItem = di;

    di->repaint();
  }
}

void
Disassembler::SlotSetBreakPoint( QListViewItem *i )
{  
  DisassemblerItem *di = dynamic_cast<DisassemblerItem *>( i );

  if ( di->isBreak ) {
    di->setPixmap( 0, QPixmap( dot_xpm ) );
    di->isBreak = false;
    breakPointsNo--;
  } else {
    di->setPixmap( 0, QPixmap( brkpt_xpm ) );
    di->isBreak = true;
    breakPointsNo++;
  }

  emit signalSetBreakPoint( strToUInt( di->text( 1 ) ), di->isBreak );

  if ( breakPointsNo > 0 ) {
    SA_ENABLED( "GotoBrkpt", true );
  } else {
    SA_ENABLED( "GotoBrkpt", false );
    emit signalNoBreakPoints();
  }

  di->repaint();
}

/*===========================================================================*/
/* class: DisassemblerItem                                                   */
/*===========================================================================*/

DisassemblerItem::DisassemblerItem( QListView *parent, 
                                    QString address, 
                                    QString mnemonic, 
                                    QString opCode,
                                    QString cycles )
  : QListViewItem( parent, "", address, mnemonic, opCode, cycles )
{
  isNext     = false;
  isBreak    = false;

  setPixmap( 0, QPixmap( dot_xpm ) );
}

void
DisassemblerItem::paintCell( QPainter *p, const QColorGroup &cg, 
                           int column, int width, int align )
{
  if ( !listView()->isEnabled() || ( !isNext && !isBreak ) ) {
    QListViewItem::paintCell( p, cg, column, width, align );
  } else if ( isNext ) {
    QColorGroup newCg;

    newCg = QColorGroup( cg.foreground(),
                         cg.background(),
                         cg.light(),
                         cg.dark(),
                         cg.mid(),
                         blue,
                         cg.base() );

    QListViewItem::paintCell( p, newCg, column, width, align );

  } else if ( isBreak ) {

    QColorGroup newCg;
    newCg = QColorGroup( cg.foreground(),
                         red,
                         cg.light(),
                         cg.dark(),
                         cg.mid(),
                         black,
                         cg.base() );

    QListViewItem::paintCell( p, newCg, column, width, align );
  }
}

/*===========================================================================*/
/* class: Editor                                                             */
/*===========================================================================*/

Editor::Editor( QWidget *parent, const char *name )
  : QMultiLineEdit( parent, name )
{
  connect( this, SIGNAL( textChanged() ), this, SLOT( SlotChanged() ) );

  changed = false;
  valid   = true;
  view    = false;

  setFont( QFont( "courier", 13 ) );
}

Editor::~Editor()
{
}

void
Editor::reset( bool clear )
{
  disconnect( this, SIGNAL( textChanged() ), this, SLOT( SlotChanged() ) );

  if ( clear ) Editor::clear();

  changed = false;
  valid   = true;

  connect( this, SIGNAL( textChanged() ), this, SLOT( SlotChanged() ) );
}

void
Editor::refresh( void )
{
  slotSelectLine( currentLine );
}

int
Editor::numLines( void )
{
  return numRows();
}

void
Editor::slotAddLine( const QString &line )
{
  disconnect( this, SIGNAL( textChanged() ), this, SLOT( SlotChanged() ) );
  insertLine( line );
  connect( this, SIGNAL( textChanged() ), this, SLOT( SlotChanged() ) );
}

void
Editor::slotSelectLine( uint line )
{
  if ( isReadOnly() ) {

    setCursorPosition( line, 0 );
    setCursorPosition( line, lineLength( line ), true );

    int cursor, col;
    getCursorPosition( &cursor, &col );

    int first, last;
    first = topCell();
    last  = lastRowVisible();

    int center = ( first + last ) / 2;

    int diff = ( cursor - center );

    verticalScrollBar()->setValue( verticalScrollBar()->value() + 
                                   diff * verticalScrollBar()->lineStep() );

    emit signalCursorPosition( line, 0 );
  }

  currentLine = line;
}

void
Editor::keyPressEvent( QKeyEvent *e )
{
  QMultiLineEdit::keyPressEvent( e );

  int x, y;
  getCursorPosition( &y, &x );
  
  if ( e->key() == HELP_KEY ) {

    // get string under cursor
    QString s( stringShown( y ) );

    s.truncate( s.find( ' ', x ) );

    int i = s.findRev( QRegExp( "[ :]" ), x );
    if ( i != -1 )
      s.remove( 0, i + 1 );

    emit signalHelp( s );

  } else {
    emit signalCursorPosition( y, x );
  }
}

void
Editor::mousePressEvent( QMouseEvent *e )
{
  if ( !isReadOnly() ) {
    QMultiLineEdit::mousePressEvent( e );

    int x, y;
  
    getCursorPosition( &y, &x );
    emit signalCursorPosition( y, x );
  }
}

void
Editor::mouseDoubleClickEvent( QMouseEvent * )
{

}

void
Editor::SlotChanged( void )
{
  emit signalChanged();
}

/*===========================================================================*/
/* class: Listing                                                            */
/*===========================================================================*/

Listing::Listing( QWidget *parent, const char *name )
  : QMultiLineEdit( parent, name )
{
  valid = true;

  setReadOnly( true );
}

Listing::~Listing()
{
}

void
Listing::reset( void )
{
  valid = true;
}

void
Listing::slotAddLine( const QString &line )
{
  insertLine( line );
}

/*===========================================================================*/
/* class: SimikonDebugger                                                    */
/*===========================================================================*/

SimikonDebugger::SimikonDebugger( QWidget *parent, Simikon8052 *c )
  : STabCtl( parent, "SimikonDebugger" )
{
  controller = c;

  connect( controller, SIGNAL( signalInvalidASM() ),
           this, SLOT( SlotRemoveEditor() ) );
  connect( controller, SIGNAL( signalInvalidHEX() ),
           this, SLOT( SlotRemoveDisassembler() ) );
  connect( controller, SIGNAL( signalInvalidLST() ),
           this, SLOT( SlotRemoveListing() ) );
  connect( controller, SIGNAL( signalValidALL() ),
           this, SLOT( SlotShowAll() ) );

  InitDebuggerActions();

  connect( this, SIGNAL( pageSelected( QWidget * ) ),
           this, SLOT( SlotPageSelected( QWidget * ) ) );
  
  hideDebugger = true;
  debug        = false;
  searchAgain  = false;

  activeDisassembler = 0;
  activeEditor       = 0;
  activeListing      = 0;
}

SimikonDebugger::~SimikonDebugger()
{
  QObjectList *list;
  QObject     *obj;

  // DISASSEMBLER
  list = queryList( 0, "Disassembler" );
  QObjectListIt itDisassembler( *list );

  while ( (obj = itDisassembler.current()) != 0 ) {
    removeChild( obj );
    ++itDisassembler;
  }
  delete list;
  
  // EDITOR
  list = queryList( 0, "Editor" );
  QObjectListIt itEditor( *list );

  while ( (obj = itEditor.current()) != 0 ) {
    removeChild( obj );
    ++itEditor;
	}
  delete list;

  // LISTING
  list = queryList( 0, "Listing" );
  QObjectListIt itListing( *list );

  while ( (obj = itListing.current()) != 0 ) {
    removeChild( obj );
    ++itListing;
  }
  delete list;
}

Disassembler *
SimikonDebugger::createDisassembler( void )
{
  Disassembler *disassembler = new Disassembler( this, "Disassembler" );

  connect( disassembler, SIGNAL( signalNoBreakPoints() ),
           this, SLOT( SlotNoBreakPoints() ) );

  return disassembler;
}

Editor *
SimikonDebugger::createEditor( void )
{
  Editor *editor = new Editor( this, "Editor" );

  connect( editor, SIGNAL( signalChanged() ), 
           this, SLOT( SlotASMChanged() ) );
  connect( editor, SIGNAL( signalCursorPosition( uint, uint ) ), 
           this, SLOT( SlotSetCursorPosition( uint, uint ) ) );

  return editor;
}

Listing *
SimikonDebugger::createListing( void )
{
  Listing *listing = new Listing( this, "Listing" );
  
  return listing;
}

void
SimikonDebugger::saveASM( const char *asmFile )
{
  QFile fp( asmFile );

  if ( !fp.open( IO_WriteOnly ) ) {
    cout << "WARNING: SimikonDebugger::saveASM(): "
         << "Can't save file! \n";
    return;
  }

  QTextStream tfp( &fp );

  int n = activeEditor->numLines();

  progress->setTotalSteps( activeEditor->numLines() );

  for ( int i = 0; i < n; i++ ) {
    tfp << activeEditor->textLine( i );
    tfp << "\n";
    progress->setProgress( i + 1 );
  }

  progress->reset();

  fp.close();

  setTabTextColor( activeEditor, black );
}

void
SimikonDebugger::InitDebuggerActions( void )
{
  SA_ADD_ACTION( "AssembleEditor", "&Assemble", assemble_xpm,
                 this, SLOT( SlotAssemble() ),
                 Key_F9, "Assemble", false );

  SA_ADD_ACTION( "StartDebug", "&Start Debugger", start_xpm,
                 this, SLOT( SlotStart() ),
                 Key_F2, "Start Debugger", false );

  SA_ADD_ACTION( "Reset", "Reset", restart_xpm,
                 this, SLOT( SlotReset() ),
                 Key_F10, "Reset", false );

  SA_ADD_ACTION( "StepInto", "Step Into", stepinto_xpm,
                 controller, SLOT( slotStepInto() ),
                 Key_F5, "Execute next instruction", false );

  SA_ADD_ACTION( "StepOut", "Step Out", stepout_xpm,
                 controller, SLOT( slotStepOut() ),
                 Key_F6, "Step out of procedure", false );
  
  SA_ADD_ACTION( "StepOver", "Step Over", stepover_xpm,
                 controller, SLOT( slotStepOver() ),
                 Key_F7, "Skip over next instruction", false );
  
  SA_ADD_ACTION( "GotoBrkpt", "Goto Breakpoint", gotobrkpt_xpm,
                 controller, SLOT( slotGotoBreakpoint() ),
                 Key_F8, "Goto Breakpoint", false );
}

void
SimikonDebugger::setActive( Disassembler *d, const QString &labelD,
                            Editor       *e, const QString &labelE,
                            Listing      *l, const QString &labelL )
{
  if ( ( activeDisassembler != d && activeEditor != e ) || refresh ) {

    disconnect( this, SIGNAL( pageSelected( QWidget * ) ),
                this, SLOT( SlotPageSelected( QWidget * ) ) );

    removeProject( activeDisassembler, activeEditor, activeListing );

    int disassemblerId = id( d );
    int editorId       = id( e );
    int listingId      = id( l );

    activeDisassembler = d;
    activeEditor       = e;
    activeListing      = l;

    if ( disassemblerId == -1 ) {

      if ( !labelD.isEmpty() && d->isValid() ) 
        insertPage( d, labelD );

      if ( !labelE.isEmpty() && e->isValid() ) {
        insertPage( e, labelE );
        if ( e->isChanged() ) {
          setTabTextColor( e, red );
        }
      }

      if ( !labelL.isEmpty() && l->isValid() ) {
        insertPage( l, labelL );
      }

    } else {

      if ( !labelD.isEmpty() && d->isValid() ) 
        setPageCaption( disassemblerId, labelD );

      if ( !labelE.isEmpty() && e->isValid() ) {
        setPageCaption( editorId, labelE );
        if ( e->isChanged() ) {
          setTabTextColor( e, red );
        }
      }
      
      if ( !labelL.isEmpty() && l->isValid() ) {
        setPageCaption( listingId, labelL );
      }
    }

    if ( id( activeEditor ) != -1 && activeEditor->isView() )
      setVisiblePage( activeEditor );
    
    connect( this, SIGNAL( pageSelected( QWidget * ) ),
             this, SLOT( SlotPageSelected( QWidget * ) ) );

    activeDisassembler->setEnabled( false );

    refresh = false;
  }
}

void
SimikonDebugger::removeProject( Disassembler *d, Editor *e, Listing *l )
{
  removePage( d );
  removePage( e );
  removePage( l );
}

// PUBLIC_SLOTS

// PRIVATE_SLOTS
void
SimikonDebugger::SlotRemoveEditor( void )
{
  activeEditor->reset();
  activeEditor->setValid( false );
  removePage( activeEditor );
}

void
SimikonDebugger::SlotRemoveDisassembler( void )
{
  activeDisassembler->reset();
  activeDisassembler->setValid( false );
  removePage( activeDisassembler );
}

void
SimikonDebugger::SlotRemoveListing( void )
{
  activeListing->reset();
  activeListing->setValid( false );
  removePage( activeListing );
}

void
SimikonDebugger::SlotShowAll( void )
{
  if ( id( activeDisassembler ) == -1 ||
       id( activeEditor ) == -1 ||
       id( activeListing ) == -1 ) {
    refresh = true;
  }
}

void
SimikonDebugger::SlotPageSelected( QWidget *t )
{
  if ( t == activeEditor ) {
    if ( debug ) {
      activeEditor->refresh();
      activeEditor->update();
      //      searchEntry->setEnabled( false );
    } else {
      activeEditor->setFocus();
      //      searchEntry->setEnabled( true );
    }

    activeEditor->setView( true );

  } else {
    activeEditor->setView( false );
    //    searchEntry->setEnabled( false );
  }
}

void
SimikonDebugger::SlotStart( void )
{
  emit signalDebugStart();

  controller->refreshActions();

  debug        = true;
  hideDebugger = true;

  SA_ENABLED( "StartDebug Open New", false );
  SA_ENABLED( "Stop Reset", true );

  //  searchEntry->setEnabled( false );
  activeDisassembler->setEnabled( true );
  
  if ( visiblePage() == activeEditor ) {
    activeEditor->refresh();
    activeEditor->update();
  }
}

void
SimikonDebugger::stop( void )
{
  QListViewItem *i = activeDisassembler->firstChild();

  while ( i ) {
    if ( dynamic_cast<DisassemblerItem *>( i )->isBreak )
      activeDisassembler->setBreakPoint( i );

    i = i->nextSibling();
  }

  debug = false;

  uint n = pageCount();
  
  for ( ; n > 0; n-- )
    setPageEnabled( n - 1, true );

  SA_ENABLED( "StartDebug Open New", true );
  SA_ENABLED( "Reset StepInto StepOut StepOver GotoBrkpt", false );

  /*  
  if ( visiblePage() == activeEditor ) {
    searchEntry->setEnabled( true );
  }
  */

  activeDisassembler->setEnabled( false );

  activeEditor->deselect();
}

void
SimikonDebugger::SlotReset( void )
{
  SA_ENABLED( "StepOut", false );
  controller->reset();
  controller->refreshActions();
}

void
SimikonDebugger::SlotASMChanged( void )
{
  if ( !activeEditor->isChanged() ) {
    setTabTextColor( activeEditor, red );
    activeEditor->setChanged( true );

    SA_ENABLED( "StartDebug StartSim", false );
    SA_ENABLED( "Assemble AssembleEditor", true );
  }
}

void
SimikonDebugger::SlotNoBreakPoints( void )
{
  controller->stopExecuting();
}

void
SimikonDebugger::SlotSetCursorPosition( uint y, uint x )
{
  cursorLabel->setText( QString( "L%1:C%2  " ).arg( y ).arg( x ) );
}

void
SimikonDebugger::SlotAssemble( void )
{
  emit signalAssemble();
}

void
SimikonDebugger::SlotSearch( const QString &text )
{
  int x = 0, y = 0;
  int n = activeEditor->numLines();
 
  activeEditor->getCursorPosition( &y, &x );

  if ( searchAgain ) {
    y++;
    searchAgain = false;
  }

  for ( ; y < n; y++ ) {

    x = activeEditor->textLine( y ).find( text );

    if ( x != -1 ) {
      activeEditor->setCursorPosition( y, x );
      activeEditor->setCursorPosition( y, x+text.length(), true );

      break;
    }
  }

  if ( y >= n )
    activeEditor->setCursorPosition( 0, 0 );
}

void
SimikonDebugger::SlotSearchNext( void )
{
  QString text = searchEntry->text();
  searchAgain = true;
  SlotSearch( text );
}
