/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef SIMIKON_DEBUGGER_HH
#define SIMIKON_DEBUGGER_HH

#include "stabctl.h"

#include <qlistview.h>
#include <qmultilineedit.h>

class Simikon8052;
class QListView;
class QListViewItem;

/*===========================================================================*/
/* class: DisassemblerItem                                                   */
/*===========================================================================*/

class DisassemblerItem: public QListViewItem {

  friend class Disassembler;
  friend class SimikonDebugger;

public:
  DisassemblerItem( QListView *parent, QString, QString, QString, QString );

protected:
  void paintCell( QPainter *, const QColorGroup &, int, int, int );
  
private:
  bool isNext;
  bool isBreak;
};

/*===========================================================================*/
/* class: Disassembler                                                       */
/*===========================================================================*/

class Disassembler: public QListView {
  Q_OBJECT
  
public:
  Disassembler( QWidget *parent, const char *name );
  ~Disassembler();

  void reset( void );
  void setBreakPoint( QListViewItem *i );

  bool isValid( void ) { return valid; }
  void setValid( bool v ) { valid = v; }

private:
  DisassemblerItem *GetItem( uint line );

protected:
  virtual bool eventFilter( QObject *, QEvent * );
  virtual void keyPressEvent( QKeyEvent * );
  virtual void contentsMousePressEvent( QMouseEvent * );

public slots:
  void slotAdd( const QString &, const QString &, const QString &, 
                const QString & );
  void slotSelect( uint );
  void slotMark( uint );

private slots:
  void SlotSetBreakPoint( QListViewItem *i );

signals:
  void signalSetBreakPoint( uint address, bool state );
  void signalNoBreakPoints( void );

private:
  uint activeLine;
  int  breakPointsNo;
  
  QListViewItem    *activeItem;
  DisassemblerItem *nextItem;
  DisassemblerItem *criticalItem;

  bool valid;
};

/*===========================================================================*/
/* class: Editor                                                             */
/*===========================================================================*/

class Editor: public QMultiLineEdit {
  Q_OBJECT

public:
  Editor( QWidget *parent, const char *name );
  ~Editor();

  void refresh( void );
  void reset( bool clear = true );

  int  numLines( void );

  bool isValid( void ) { return valid; }
  bool isChanged( void ) { return changed; }
  bool isView( void ) { return view; }

  void setValid( bool v ) { valid = v; }
  void setChanged( bool c ) { changed = c; }
  void setView( bool v ) { view = v; }

protected:
  virtual void keyPressEvent( QKeyEvent * );
  virtual void mousePressEvent( QMouseEvent * );
  virtual void mouseDoubleClickEvent( QMouseEvent * );

public slots:
  void slotAddLine( const QString & );
  void slotSelectLine( uint );

private slots:
  void SlotChanged( void );

signals:
  void signalChanged( void );
  void signalCursorPosition( uint y, uint x );
  void signalHelp( const char * );

private:
  bool changed;
  bool valid;
  bool view;

  uint currentLine;
};

/*===========================================================================*/
/* class: Listing                                                            */
/*===========================================================================*/

class Listing: public QMultiLineEdit {
  Q_OBJECT

public:
  Listing( QWidget *parent, const char *name );
  ~Listing();

  void reset( void );

  bool isValid( void ) { return valid; }
  void setValid( bool v ) { valid = v; }

public slots:
  void slotAddLine( const QString & );

private:
  bool valid;
};

/*===========================================================================*/
/* class: SimikonDebugger                                                    */
/*===========================================================================*/

class SimikonDebugger: public STabCtl {
  Q_OBJECT

public:
  SimikonDebugger( QWidget *parent, Simikon8052 *controller );
  ~SimikonDebugger();

  Disassembler *createDisassembler( void );
  Editor       *createEditor( void );
  Listing      *createListing( void );

  void setActive( Disassembler *d, const QString &labelD,
                  Editor       *e, const QString &labelE,
                  Listing      *l, const QString &labelL );
  void removeProject( Disassembler *d, Editor *e, Listing *l );
  void saveASM( const char *asmFile );
  void stop( void );

private:
  void InitDebuggerActions( void );

private slots:
  void SlotStart( void );
  //  void SlotStop( void );
  void SlotReset( void );
  void SlotASMChanged( void );
  void SlotNoBreakPoints( void );
  void SlotAssemble( void );
  void SlotSetCursorPosition( uint y, uint x );
  void SlotSearch( const QString &text );
  void SlotSearchNext( void );
  void SlotPageSelected( QWidget *t );
  void SlotRemoveEditor( void );
  void SlotRemoveDisassembler( void );
  void SlotRemoveListing( void );
  void SlotShowAll( void );

signals:
  void signalDebugStart( void );
  void signalDebugStop( void );
  void signalAssemble( void );
  
private:
  QListView    *listview;

  Disassembler *activeDisassembler;
  Editor       *activeEditor;
  Listing      *activeListing;

  Simikon8052  *controller;

  bool debug, hideDebugger;
  bool searchAgain;
  bool refresh;
};

#endif // SIMIKON_DEBUGGER_HH
