/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "config.h"
#include "SimikonHelp.hh"
#include "SimikonActions.hh"
#include "SimikonSettings.hh"
#include "LeakTracker.hh"

// QT_INCLUDES
#include <qtoolbar.h>
#include <qtextbrowser.h>
#include <qcombobox.h>

// PIXMAPS
#include "home.xpm"
#include "forward.xpm"
#include "backward.xpm"

// DEFINES
#define ESCAPE_KEY (4096)

#include <iostream.h>

// GLOBAL_OBJECTS
extern SimikonActions *simikonActions;
extern SimikonSettings *simikonSettings;

/*===========================================================================*/
/* class: SimikonHelp                                                        */
/*===========================================================================*/

SimikonHelp::SimikonHelp( QWidget *parent, const char *name )
  : QMainWindow( parent, name )
{
  setCaption( "Instruction Reference" );

  actionName = new QString( name );

  browser = new QTextBrowser( this );
  browser->mimeSourceFactory()->setFilePath( "." );
 
  QString helpHome( SS_GET( "help_dir" ) );

  if ( !helpHome.isEmpty() ) {
    helpHome += INSTRUCTION_HELP;
  } else {
    helpHome = DEFAULT_HELP_DIR;
    helpHome += "/";
    helpHome += INSTRUCTION_HELP;
  }

  browser->setSource( helpHome );

  setCentralWidget( browser );

  SimikonHelp::InitActions();
  SimikonHelp::InitToolBar();

  connect( browser, SIGNAL( textChanged() ),
           this, SLOT( SlotTextChanged() ) );
  connect( browser, SIGNAL( backwardAvailable( bool ) ),
           this, SLOT( SlotBackwardAvailable( bool ) ) );
  connect( browser, SIGNAL( forwardAvailable( bool ) ),
           this, SLOT( SlotForwardAvailable( bool ) ) );

  toolbar->addSeparator();

  // PATH_COMBO
  path = new QComboBox( true, toolbar );
  path->insertItem( helpHome );

  connect( path, SIGNAL( activated( const QString & ) ),
           this, SLOT( SlotPathSelected( const QString & ) ) );

  toolbar->setStretchableWidget( path );
  setRightJustification( true );
  setDockEnabled( Left, false );
  setDockEnabled( Right, false );

  browser->setFocus();

  resize( 550, 300 );
}

SimikonHelp::~SimikonHelp()
{
  delete actionName;
}

// PROTECTED
bool
SimikonHelp::eventFilter( QObject *o, QEvent *e )
{
  if ( e->type() == QEvent::Hide ) {
    SA_CHECKED( actionName->ascii(), false );
    return true;
  } else if ( e->type() == QEvent::KeyPress ) {
    if ( dynamic_cast<QKeyEvent *>( e )->key() == ESCAPE_KEY ) {
      close();
      return true;
    }
  }

  return QMainWindow::eventFilter( o, e );
}

// PRIVATE
void
SimikonHelp::InitActions( void )
{
  SA_ADD_ACTION( "Home", tr( "&Home" ), home_xpm,
                 browser, SLOT( backward() ), ALT+Key_H,
                 "Home", false );
  SA_ADD_ACTION( "Backward", tr( "&Backward" ), backward_xpm,
                 browser, SLOT( backward() ), ALT+Key_B,
                 "Backward", false );
  SA_ADD_ACTION( "Forward", tr( "&Forward" ), forward_xpm,
                 browser, SLOT( forward() ), ALT+Key_F,
                 "Forward", false );
}

void
SimikonHelp::InitToolBar( void )
{
  toolbar = new QToolBar( this, "HelpBrowserToolBar" );

  SA_ADD_TOOLBUTTONS( toolbar, "Home Backward Forward" );

  SA_ENABLED( "Backward", false );
  SA_ENABLED( "Forward",  false );
}

// PUBLIC_SLOTS
void
SimikonHelp::slotHelp( const char *str )
{
  QString helpFile( SS_GET( "help_dir" ) );

  helpFile += "/";
  helpFile += str;
  helpFile += ".html";

  browser->setSource( helpFile );
}


// PRIVATE_SLOTS
void
SimikonHelp::SlotForwardAvailable( bool available )
{
  SA_ENABLED( "Forward", available );
}

void
SimikonHelp::SlotBackwardAvailable( bool available )
{
  SA_ENABLED( "Backward", available );
}

void
SimikonHelp::SlotTextChanged( void )
{
  QString url = browser->context();

  bool exists = false;
  int i;

  for ( i = 0; i < path->count(); i++ ) {
    if ( path->text( i ) == url ) {
      exists = true;
      break;
    }
  }

  if ( !exists ) {
    path->insertItem( url, 0 );
    path->setCurrentItem( 0 );
  } else {
    path->setCurrentItem( i );
  }
}

void
SimikonHelp::SlotPathSelected( const QString &new_path )
{
  browser->setSource( new_path );
}
