/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "config.h"
#include "SimikonMemory.hh"
#include "Simikon8052.hh"
#include "SimikonUtils.hh"
#include "LeakTracker.hh"

// QT_INCLUDES
#include <qobjcoll.h>
#include <qlineedit.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qframe.h>

// WIDGETS_INCLUDES
#include "stabctl.h"

// PIXMAPS
#include "konj.xpm"

/*===========================================================================*/
/* class: MemoryItem                                                         */
/*===========================================================================*/

MemoryItem::MemoryItem( QTable *table, EditType et, const QString &t,
                        const QPixmap &p )
  : QTableItem( table, et, t, p )
{
}

MemoryItem::~MemoryItem()
{
}

int
MemoryItem::alignment() const
{
  return ( AlignRight | AlignVCenter );
}

/*===========================================================================*/
/* class: Memory                                                             */
/*===========================================================================*/

Memory::Memory( QWidget *parent, const char *name )
  : QTable( parent, name )
{
  setNumRows( 8 );
  setNumCols( 8 );

  setShowGrid( false );
  setColumnMovingEnabled( false );
  setRowMovingEnabled( false );
  setSelectionMode( Single );

  QHeader *header;

  // horizontal header stuff
  header = horizontalHeader();
  setTopMargin( 0 );
  header->hide();

  // vertical header stuff
  header = verticalHeader();
  setLeftMargin( 55 );
  header->setResizeEnabled( false );
  header->setMovingEnabled( false );
  
  for ( uint i = 0; i < 8; i++ ) {
    setColumnWidth( i, 20 );
    header->setLabel( i, "0x" + uintToHexStr16( i ) );
  }

  writable = false;
  editable = false;
  size = 0;

  memory = new int [ 64 * 1024 ];

  startAddress = 0;

  for ( uint i = 0; i < size; i++ ) {
    memory[i] = 0;
  }
  
  for ( uint y = 0; y < 8; y++ )
    for ( uint x = 0; x < 8; x++ )
      setText( y, x, "00" );

}

Memory::~Memory()
{
  delete [] memory;

  /*
   * HACK: Memory Leak in Memory::setText() (new MemoryItem) ????
   */

  for ( uint i = 0; i < 8; i++ )
    for ( uint j = 0; j < 8; j++ )
      delete item( i, j );
      
}

void
Memory::setText( int row, int col, const QString &text )
{
  QTableItem *i = item( row, col );

  if ( i ) {
    i->setText( text );
    updateCell( row, col );
  } else {
    i = new MemoryItem( this, QTableItem::OnActivate, text, QPixmap() );
    //    i = new QTableItem( this, QTableItem::OnActivate, text );
    setItem( row, col, i );
  }
}

uint
Memory::getAddress( uint x, uint y )
{
  return ( startAddress + x * 8 + y );
}

void
Memory::setMemory( uint _size, bool _writable )
{
  size     = _size;
  writable = _writable;  
}

void
Memory::setEditable( bool _editable )
{
  editable = _editable;
}

void
Memory::refreshAll( uint address )
{
  startAddress = address;

  StartAddressChanged();

  for ( uint y = 0; y < 8; y++ ) {
    for ( uint x = 0; x < 8; x++ ) {
        
      QString cellValue;
      
      cellValue.setNum( memory[ address ], 16 );
      cellValue = cellValue.upper();
        
      if ( memory[ address ] <= 15 ) 
        cellValue.prepend( '0' );
        
      address++;
      
      setText( y, x, cellValue );
    }
  }
}

void
Memory::Refresh( uint address, uint value )
{
  if ( ( address >= startAddress ) && ( address < ( startAddress + 64 ) ) ) {

    QString cellValue;
    
    cellValue.setNum( value, 16 );
    cellValue = cellValue.upper();

    if ( value <= 15 ) cellValue.prepend( '0' );

    // proracunaj koordinate odgovarajuce memorijske lokacije
    uint x, y;
    
    x = ( address - startAddress ) % 8;
    y = ( address - startAddress ) / 8;

    setText( y, x, cellValue );
  }
}

void
Memory::StartAddressChanged( void )
{
  QHeader *header = verticalHeader();

  for ( uint i = 0; i < 8; i++ ) {
    header->setLabel( i, "0x" + uintToHexStr16( startAddress + i * 8 ) );
  }
}

void
Memory::slotUpdate( uint address, uint value )
{
  memory[ address ] = value;
  Refresh( address, value );
}

void
Memory::slotClear( void )
{
  for ( uint i = 0; i < 0xFFFF; i++ ) {
    memory[ i ] = 0;
    Refresh( i, 0 );
  }
}

QWidget *
Memory::beginEdit( int row, int col, bool replace )
{
  return ( ( !editable ) ? 0 : QTable::beginEdit( row, col, replace ) );
}

/*===========================================================================*/
/* class: SimikonMemory                                                      */
/*===========================================================================*/

SimikonMemory::SimikonMemory( QWidget *parent, const char *name, 
                              const char *action, Simikon8052 *c )
  : SimikonModule( parent, name, action )
{
  controller = c;

  connect( controller, SIGNAL( signalResetMemoryView( void ) ),
           this, SLOT( SlotResetMemoryView( void ) ) );
  connect( stabctl, SIGNAL( pageSelected( QWidget * ) ),
           this, SLOT( SlotActiveMemoryChanged( QWidget * ) ) );

  QFrame *frame = new QFrame( this );
  frame->setFrameStyle( QFrame::WinPanel | QFrame::Raised );
  frame->setLineWidth( 2 );
  topLayout->addWidget( frame );
  
  QBoxLayout *hboxLayout;

  // ADDRESS_ENTRY
  hboxLayout = new QHBoxLayout( frame, 2 );

  addressEntry = new QLineEdit( frame );
  addressEntry->setMaxLength( 4 ); // FFFF
  connect( addressEntry, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotAddressChanged( const QString & ) ) );
  hboxLayout->addWidget( addressEntry );

  cb1 = new QCheckBox( "Auto", frame );
  cb1->setChecked( true );
  connect( cb1, SIGNAL( toggled( bool ) ),
           this, SLOT( SlotAutoToggled( bool ) ) );
  hboxLayout->addWidget( cb1 );

  addressButton = new QPushButton( frame );
  addressButton->setEnabled( false );
  connect( addressButton, SIGNAL( clicked() ),
           this, SLOT( SlotAddressButtonClicked() ) );
  addressButton->setPixmap( QPixmap( konj_xpm ) );
  hboxLayout->addWidget( addressButton );
  
  cb2 = new QCheckBox( "Editable", frame );
  cb2->setChecked( true );
  connect( cb2, SIGNAL( toggled( bool ) ),
           this, SLOT( SlotEditableToggled( bool ) ) );
  hboxLayout->addWidget( cb2 );

  InitIRAM();
  InitIROM();
  InitXRAM();
  InitXROM();

  resize( 235, 219 );
  setFixedSize( 235, 219 );
}

SimikonMemory::~SimikonMemory()
{

}

void
SimikonMemory::InitIRAM( void )
{
  iram = new Memory( stabctl, "IRAM" );

  iram->setEditable( true );

  connect( controller, SIGNAL( signalUpdateIRAM( uint, uint ) ),
           iram, SLOT( slotUpdate( uint, uint ) ) );
  connect( iram, SIGNAL( valueChanged( int, int ) ),
           this, SLOT( SlotUpdateIRAM( int, int ) ) );

  stabctl->insertPage( iram, "IRAM" );
}

void
SimikonMemory::InitIROM( void )
{
  irom = new Memory( stabctl, "IROM" );

  connect( controller, SIGNAL( signalUpdateIROM( uint, uint ) ),
           irom, SLOT( slotUpdate( uint, uint ) ) );
  connect( controller, SIGNAL( signalClearIROM( void ) ),
           irom, SLOT( slotClear( void ) ) );

  stabctl->insertPage( irom, "IROM" );
}

void
SimikonMemory::InitXRAM( void )
{
  xram = new Memory( stabctl, "XRAM" );

  xram->setEditable( true );

  connect( controller, SIGNAL( signalUpdateXRAM( uint, uint ) ),
           xram, SLOT( slotUpdate( uint, uint ) ) );

  stabctl->insertPage( xram, "XRAM" );
}

void
SimikonMemory::InitXROM( void )
{
  xrom = new Memory( stabctl, "XROM" );

  connect( controller, SIGNAL( signalUpdateXROM( uint, uint ) ),
           xrom, SLOT( slotUpdate( uint, uint ) ) );
  connect( controller, SIGNAL( signalClearXROM( void ) ),
           xrom, SLOT( slotClear( void ) ) );

  stabctl->insertPage( xrom, "XROM" );
}

bool
SimikonMemory::eventFilter( QObject *o, QEvent *e )
{
  if ( e->type() == QEvent::WindowActivate ) {
    addressEntry->setFocus();
    addressEntry->selectAll();
    return true;
  }

  return SimikonModule::eventFilter( o, e );
}

void
SimikonMemory::setActiveMemory( const MemoryData &memory )
{
  uint n = stabctl->pageCount();

  for ( ; n > 0; n-- ) {
    stabctl->removePage( n-1 );
  }

  activeMemory = 0;

  if ( memory.iram ) {
    activeMemory = iram;
    iram->setMemory( strToUInt( memory.iramEnd ), true );
    stabctl->insertPage( iram, memory.iramName );
  }

  if ( memory.irom ) {
    if ( !activeMemory ) activeMemory = irom;
    irom->setMemory( strToUInt( memory.iromEnd ), false );
    stabctl->insertPage( irom, memory.iromName );
  }

  if ( memory.xram ) {
    if ( !activeMemory ) activeMemory = xram;
    xram->setMemory( strToUInt( memory.xramEnd ), true );
    stabctl->insertPage( xram, memory.xramName );
  }

  if ( memory.xrom ) {
    if ( !activeMemory ) activeMemory = xrom;
    xrom->setMemory( strToUInt( memory.xromEnd ), false );
    stabctl->insertPage( xrom, memory.xromName );
  }

}

void
SimikonMemory::SlotResetMemoryView( void )
{
  addressEntry->setText( "0000" );
  addressEntry->selectAll();
  activeMemory->refreshAll( 0 );
}

void
SimikonMemory::SlotActiveMemoryChanged( QWidget *current )
{
  activeMemory = ( Memory * ) current;

  if ( activeMemory->isWritable() ) {
    cb2->setEnabled( true );
  } else {
    cb2->setEnabled( false );
  }

  addressEntry->selectAll();
}

void
SimikonMemory::SlotAutoToggled( bool state )
{
  if ( !addressEntry->text().isEmpty() ) {
    addressButton->setEnabled( !state );
  }
}

void
SimikonMemory::SlotEditableToggled( bool state )
{
  activeMemory->setEditable( state );
}

void
SimikonMemory::SlotAddressChanged( const QString &address )
{
  if ( !cb1->isChecked() ) {
    if ( !address.isEmpty() ) {
      addressButton->setEnabled( true );
      addressEntry->setText( address.upper() );
    } else {
      addressButton->setEnabled( false );
    }
  } else {
    bool ok;

    uint startAddress;

    if ( !address.isEmpty() ) {
      startAddress = address.toUInt( &ok, 16 );
      addressEntry->setText( address.upper() );
    } else {
      startAddress = 0;
      addressEntry->setText( "0" );
      addressEntry->selectAll();
      ok = true; 
    }

    if ( ok ) {
      activeMemory->refreshAll( startAddress );
    }
  }
}

void
SimikonMemory::SlotAddressButtonClicked( void )
{
  bool state;
  uint startAddress = addressEntry->text().toUInt( &state, 16 );

  if ( state ) {
    activeMemory->refreshAll( startAddress );
  }
}

void
SimikonMemory::SlotUpdateIRAM( int y, int x )
{
  QString value   = activeMemory->text( y, x );
  uint    address = activeMemory->getAddress( x, y );

  controller->updateIRAM( address, strToUInt( "0x" + value ) );
}

