/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "config.h"
#include "SimikonPerformance.hh"
#include "SimikonActions.hh"
#include "Simikon8052.hh"
#include "SimikonUtils.hh"
#include "LeakTracker.hh"

// QT_INCLUDES
#include <qobjcoll.h>
#include <qevent.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qlistview.h>
#include <qframe.h>

// WIDGET_INCLUDES
#include "stabctl.h"

// PIXMAPS
#include "delete.xpm"
#include "new.xpm"
#include "add_auto.xpm"

// GLOBAL_OBJECTS
extern SimikonActions *simikonActions;

/*===========================================================================*/
/* class: Performance                                                        */
/*===========================================================================*/

Performance::Performance( QWidget *parent, const char *name )
  : QVBox( parent, name )
{
  listview = new QListView( this );

  listview->setAllColumnsShowFocus( true );
  
  listview->addColumn( "Symbol" );
  listview->addColumn( "Start" );
  listview->addColumn( "Symbol" );
  listview->addColumn( "End" );
  listview->addColumn( "Cycles" );
  listview->addColumn( "Time" );
  
  listview->setColumnWidth( 0, 75 );
  listview->setColumnWidth( 1, 50 );
  listview->setColumnWidth( 2, 75 );
  listview->setColumnWidth( 3, 50 );
  listview->setColumnWidth( 4, 70 );
  listview->setColumnWidth( 5, 70 );

  listview->setSorting( -1 );
}

Performance::~Performance()
{

}

QListViewItem *
Performance::add( uint address, const QString &symbol )
{  
  QListViewItem *i = new QListViewItem( listview );

  i->setText( 0, symbol );
  i->setText( 1, "0x" + uintToHexStr16( address ) );

  return i;
}

/*===========================================================================*/
/* class: SimikonPerformance                                                 */
/*===========================================================================*/

SimikonPerformance::SimikonPerformance( QWidget *parent, const char *name,
                                        const char *action,
                                        Simikon8052 *c )
  : SimikonModule( parent, name, action )
{
  controller = c;

  connect( controller, SIGNAL( signalResetPerformance( void ) ),
           this, SLOT( SlotReset( void ) ) );
  
  QFrame *frame = new QFrame( this );
  frame->setFrameStyle( QFrame::WinPanel | QFrame::Raised );
  frame->setLineWidth( 2 );
  topLayout->addWidget( frame );

  commands = CreateCommands( frame );

  resize( 415, 150 );
}

SimikonPerformance::~SimikonPerformance()
{
  QObjectList *performanceList = stabctl->queryList( 0, "Performance" );
  QObjectListIt it( *performanceList );
  QObject *obj;
  
  while ( ( obj = it.current() ) != 0 ) {
    stabctl->removeChild( obj );
    ++it;
  }
  
  delete performanceList;
}

Performance *
SimikonPerformance::createPerformance( void )
{
  return new Performance( stabctl, "Performance" );
}

void
SimikonPerformance::setActivePerformance( Performance *performance, 
                                          const char *name )
{
  removeAll();

  int id = stabctl->id( performance );

  activePerformance = performance;

  if ( id == -1 ) {
    stabctl->insertPage( performance, name );
  } else {
    stabctl->setVisiblePage( id );
  }
}

void
SimikonPerformance::removeAll( void )
{
  int n = stabctl->pageCount();

  if ( n > 0 )
    for ( ; n > 0; n-- )
      stabctl->removePage( n-1 );
}

void
SimikonPerformance::removePerformance( Performance *p )
{
  stabctl->removePage( p );
}

QWidget *
SimikonPerformance::CreateCommands( QWidget *parent )
{
  QHBoxLayout *hbox = new QHBoxLayout( parent, 2 );

  QLabel *label;

  label = new QLabel( "  Start: ", parent );
  hbox->addWidget( label );
  start = new QLineEdit( parent );
  start->setMaxLength( 4 );
  hbox->addWidget( start );

  label = new QLabel( "  End: ", parent );
  hbox->addWidget( label );
  end = new QLineEdit( parent );
  end->setMaxLength( 4 );
  hbox->addWidget( end );

  addButton = new QPushButton( parent );
  connect( addButton, SIGNAL( clicked() ),
           this, SLOT( SlotAdd() ) );
  addButton->setPixmap( QPixmap( new_xpm ) );

  removeButton = new QPushButton( parent );
  connect( removeButton, SIGNAL( clicked() ),
           this, SLOT( SlotRemove() ) );
  removeButton->setPixmap( QPixmap( delete_xpm ) );
  removeButton->setEnabled( false );

  addAutoButton = new QPushButton( parent );
  connect( addAutoButton, SIGNAL( clicked() ),
           this, SLOT( SlotAddAuto( void ) ) );
  addAutoButton->setPixmap( QPixmap( add_auto_xpm ) );

  removeAllButton = new QPushButton( parent );
  connect( removeAllButton, SIGNAL( clicked() ),
           this, SLOT( SlotRemoveAll() ) );
  removeAllButton->setPixmap( QPixmap( delete_xpm ) );
  removeAllButton->setEnabled( false );

  hbox->addWidget( addButton );
  hbox->addWidget( removeButton );
  hbox->addWidget( addAutoButton );
  hbox->addWidget( removeAllButton );  

  return parent;
}

void
SimikonPerformance::SlotAdd( void )
{
  if ( !start->text().isEmpty() && !end->text().isEmpty() ) {
    QListViewItem *item = new QListViewItem( activePerformance->listview );

    item->setText( 1, "0x" + start->text() );
    item->setText( 3, "0x" + end->text() );

    if ( activePerformance->listview->childCount() == 1 ) {
      removeButton->setEnabled( true );
      removeAllButton->setEnabled( true );
    }

    start->clear();
    end->clear();

    emit signalAdd( item, false );
  }
}

void
SimikonPerformance::SlotAddAuto( void )
{
  emit signalAddAuto();

  if ( activePerformance->listview->childCount() > 0 ) {
    removeButton->setEnabled( true );
    removeAllButton->setEnabled( true );
    addAutoButton->setEnabled( false );
  }

}

void
SimikonPerformance::SlotRemove( void )
{
  QListViewItem *i = activePerformance->listview->selectedItem();

  if ( i ) {
    emit signalRemove( i );
    activePerformance->listview->removeItem( i );
  }
  
  if ( activePerformance->listview->childCount() == 0 ) {
    removeButton->setEnabled( false );
    removeAllButton->setEnabled( false );
    addAutoButton->setEnabled( true );
  }
}

void
SimikonPerformance::SlotRemoveAll( void )
{
  QListViewItem *i = activePerformance->listview->firstChild();

  while ( i ) {
    emit signalRemove( i );
    QListViewItem *tmp = i;
    i = i->nextSibling();
    activePerformance->listview->removeItem( tmp );
  }

  removeButton->setEnabled( false );
  removeAllButton->setEnabled( false );
  addAutoButton->setEnabled( true );
}

void
SimikonPerformance::SlotReset( void )
{
  emit signalReset();
}
