/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "config.h"
#include "SimikonProject.hh"
#include "SimikonActions.hh"
#include "SimikonSettings.hh"
#include "SimikonSplash.hh"
#include "SimikonRegisters.hh"
#include "SimikonSFR.hh"
#include "SimikonMemory.hh"
#include "SimikonStack.hh"
#include "SimikonDebugger.hh"
#include "SimikonPerformance.hh"
#include "SimikonSerial.hh"
#include "SimikonInterrupts.hh"
#include "SimikonIOPorts.hh"
#include "SimikonConsole.hh"
#include "Simikon8052.hh"
#include "SimikonAS31.hh"
#include "SimikonUtils.hh"
#include "SimikonAbout.hh"
#include "SimikonHelp.hh"
#include "LeakTracker.hh"

// MISC_INCLUDES
#include <iostream.h>

// WIDGETS
#include "stabctl.h"

// QT_INCLUDES
#include <qlayout.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qfiledialog.h>
#include <qpushbutton.h>
#include <qdatetime.h>
#include <qgroupbox.h>
#include <qradiobutton.h>
#include <qbuttongroup.h>
#include <qmessagebox.h>
#include <qarray.h>
#include <qtable.h>
#include <qgroupbox.h>

// PIXMAPS
#include "file.xpm"
#include "folder.xpm"
#include "project.xpm"
#include "project_active.xpm"

// GLOBAL_OBJECTS
extern SimikonActions  *simikonActions;
extern SimikonSettings *simikonSettings;
extern SimikonSplash   *simikonSplash;
extern SimikonConsole  *simikonConsole;

extern QArray<ErrorData> errors;

static const char *ACTIONS_TABLE[ NO_OF_ACTIONS ] = {
  "Save", "Assemble", "AssembleEditor", "StartDebug", "Debugger"
};

/*===========================================================================*/
/* class: SimikonProject                                                     */
/*===========================================================================*/

SimikonProject::SimikonProject( QWidget *parent, const char *name )
  : QSplitter( parent, name )
{
  setOrientation( Vertical );

  // SIMIKON_8052
  SPLASH_MSG( tr( "Initializing Simikon8052 engine..." ) );

  stabctl = new STabCtl( this );

  controller = new Simikon8052();
  CHECK_PTR( controller );
  connect( controller, SIGNAL( signalStopExecuting() ),
           this, SLOT( slotStop() ));

  // ASSEMBLER
  SPLASH_MSG( tr( "Initializing 8052 assembler..." ) );

  assembler = new SimikonAS31();
  CHECK_PTR( assembler );
  connect( assembler, SIGNAL( signalAssembled() ), 
           this, SLOT( SlotAssemblerProjectAssembled() ) );
  connect( assembler, SIGNAL( signalSyntaxError() ),
           this, SLOT( SlotAssemblerSyntaxError() ) );

  // PROJECT_SETTINGS
  SPLASH_MSG( tr( "Initializing project settings..." ) );

  projectSettings = new ProjectSettings( stabctl, "ProjectSettings" );
  CHECK_PTR( projectSettings );
  projectSettings->hide();

  // SIMIKON_ABOUT
  SPLASH_MSG( tr( "Initializing SimikonAbout..." ) );

  simikonAbout = new SimikonAbout( stabctl, "SimikonAbout" );
  CHECK_PTR( simikonAbout );
  simikonAbout->hide();

  // SIMIKON_HELP
  SPLASH_MSG( tr( "Initialize Simikon Instruction Set Help..." ) );

  simikonHelp = new SimikonHelp( stabctl, "InstructionSetHelp" );
  CHECK_PTR( simikonHelp );
  simikonHelp->hide();


  // SIMIKON_DEBUGGER
  SPLASH_MSG( tr( "Initializing SimikonDebugger..." ) );

  simikonDebugger = new SimikonDebugger( this, controller );
  CHECK_PTR( simikonDebugger );

  connect( simikonDebugger, SIGNAL( signalDebugStart() ),
           this, SLOT( SlotStartDebug() ) );
  connect( simikonDebugger, SIGNAL( signalAssemble() ),
           this, SLOT( slotAssembleProject() ) );

  moveToFirst( simikonDebugger );

  simikonDebugger->hide();

  // SIMIKON_PERFORMANCE
  SPLASH_MSG( tr( "Initializing SimikonPerformance..." ) );

  simikonPerformance = new SimikonPerformance( stabctl, "SimikonPerformance", 
                                               "Performance", controller );
  CHECK_PTR( simikonPerformance );
  simikonPerformance->hide();
  module[ MODULE_PERFORMANCE_E ].name = "Performance";
  module[ MODULE_PERFORMANCE_E ].module = simikonPerformance;

  // SIMIKON_SERIAL
  SPLASH_MSG( tr( "Initializing SimikonSerial..." ) );

  simikonSerial = new SimikonSerial( stabctl, "SimikonSerial", 
                                     "Serial", controller );
  CHECK_PTR( simikonSerial );
  simikonSerial->hide();
  module[ MODULE_SERIAL_E ].name = "Serial";
  module[ MODULE_SERIAL_E ].module = simikonSerial;

  // SIMIKON_REGISTERS
  SPLASH_MSG( tr( "Initializing SimikonRegisters..." ) );

  simikonRegisters = new SimikonRegisters( stabctl, "SimikonRegisters", 
                                           "Registers", controller );
  CHECK_PTR( simikonRegisters );
  simikonRegisters->hide();
  module[ MODULE_REGISTERS_E ].name = "Registers";
  module[ MODULE_REGISTERS_E ].module = simikonRegisters;

  // SIMIKON_SFR
  SPLASH_MSG( tr( "Initializing SimikonSFR..." ) );

  simikonSFR = new SimikonSFR( stabctl, "SimikonSFR", "SFR", controller );
  CHECK_PTR( simikonSFR );
  simikonSFR->hide();
  module[ MODULE_SFR_E ].name = "SFR";
  module[ MODULE_SFR_E ].module = simikonSFR;

  connect( simikonRegisters, SIGNAL( signalSFRClicked( const QString & ) ),
           simikonSFR, SLOT( slotSFRClicked( const QString & ) ) );

  // SIMIKON_INTERRUPTS
  SPLASH_MSG( tr( "Initializing SimikonInterrupts..." ) );

  simikonInterrupts = new SimikonInterrupts( stabctl, "SimikonInterrupts",
                                             "Interrupts", controller );
  CHECK_PTR( simikonInterrupts );
  module[ MODULE_INTERRUPTS_E ].name = "Interrupts";
  module[ MODULE_INTERRUPTS_E ].module = simikonInterrupts;

  // SIMIKON_MEMORY
  SPLASH_MSG( tr( "Initializing SimikonMemory..." ) );

  simikonMemory = new SimikonMemory( stabctl, "SimikonMemory", 
                                     "Memory", controller );
  CHECK_PTR( simikonMemory );
  simikonMemory->hide();
  module[ MODULE_MEMORY_E ].name = "Memory";
  module[ MODULE_MEMORY_E ].module = simikonMemory;

  // SIMIKON_STACK
  SPLASH_MSG( tr( "Initializing SimikonStack..." ) );

  simikonStack = new SimikonStack( stabctl, "SimikonStack", "Stack", controller );
  CHECK_PTR( simikonStack );
  simikonStack->hide();
  module[ MODULE_STACK_E ].name = "Stack";
  module[ MODULE_STACK_E ].module = simikonStack;

  // SIMIKON_IO_PORTS
  SPLASH_MSG( tr( "Initializing Simikon IO Ports..." ) );

  simikonIOPorts = new SimikonIOPorts( stabctl, "SimikonIOPorts", 
                                       "IOPorts", controller );
  CHECK_PTR( simikonIOPorts );
  simikonIOPorts->hide();
  module[ MODULE_IOPORTS_E ].name = "IOPorts";
  module[ MODULE_IOPORTS_E ].module = simikonIOPorts;

  // INITIALIZE
  controller->initialize();

  InitProjectTab();
  InitConsoleTab();

  connect( stabctl, SIGNAL( pageSelected( QWidget * ) ),
           this, SLOT( SlotPageSelected( QWidget * ) ) );

  activeProject = 0;
  assembling    = false;
  mode          = false;

  SC_MESSAGE( tr( "SiMiKON environment initialized!!!" ) );
}

SimikonProject::~SimikonProject()
{
  delete simikonDebugger;
  delete controller;
  delete assembler;
}

void
SimikonProject::closeAll( void )
{
  if ( activeProject ) {
    QString lastActiveProject( activeProject->pData.dir + "/" + 
                               activeProject->pData.name + ".sim" );

    QListViewItem *i = project->firstChild();

    while ( i ) {
      Project *prj = dynamic_cast<Project *>( i );
      i = i->nextSibling();

      UninstallProject( prj );
    }

    SS_PUT( "last_project", lastActiveProject );
  }
}

void
SimikonProject::openProject( const QString &prjFile )
{
  QFile fp( prjFile );
  QDataStream dfp( &fp );

  if ( fp.open( IO_ReadOnly ) ) {

    ProjectData pdata;

    dfp >> pdata.author;
    dfp >> pdata.name;
    dfp >> pdata.dir;

    QString realProjectDir = stripFileName( prjFile );

    if ( pdata.dir != realProjectDir ) {
      pdata.dir = realProjectDir;
      SC_WARNING( tr( "Project's root directory has changed!" ) );
    }

    dfp >> pdata.hexFile;
    dfp >> pdata.asmFile;
    dfp >> pdata.lstFile;
    dfp >> pdata.date;

    dfp >> pdata.controller.clock;
    dfp >> pdata.controller.programMemory;

    dfp >> pdata.memory.iram;
    dfp >> pdata.memory.irom;
    dfp >> pdata.memory.xram;
    dfp >> pdata.memory.xrom;
    dfp >> pdata.memory.iramName;
    dfp >> pdata.memory.iramStart;
    dfp >> pdata.memory.iramEnd;
    dfp >> pdata.memory.iromName;
    dfp >> pdata.memory.iromStart;
    dfp >> pdata.memory.iromEnd;
    dfp >> pdata.memory.xramName;
    dfp >> pdata.memory.xramStart;
    dfp >> pdata.memory.xramEnd;
    dfp >> pdata.memory.xromName;
    dfp >> pdata.memory.xromStart;
    dfp >> pdata.memory.xromEnd;

    for ( uint i = 0; i < 8; i++ )
      dfp >> pdata.moduleState[i];
    
    pdata.isNew = false;

    fp.close();

    CreateProject( pdata );
  }
}

void
SimikonProject::InitProjectTab( void )
{
  SPLASH_MSG( tr( "Initialize Simikon Project Manager..." ) );

  project = new QListView( stabctl, "ProjectView" );
  project->setEnabled( false );

  project->setSorting( -1 );
  project->setAllColumnsShowFocus( true );

  project->addColumn( "Project" );
  project->addColumn( "Author" );
  project->addColumn( "Date" );

  project->setColumnWidth( 0, 135 );
  project->setColumnWidth( 1, 100 );
  project->setColumnWidth( 2, 95 );

  connect( project, SIGNAL( doubleClicked( QListViewItem * ) ),
           this, SLOT( slotProjectSettings() ) );

  stabctl->insertPage( project, tr( "Projects" ) );
}

void
SimikonProject::InitConsoleTab( void )
{
  SPLASH_MSG( tr( "Initialize Simikon Console..." ) );

  simikonConsole = new SimikonConsole( stabctl, "SimikonConsole" );
  stabctl->insertPage( simikonConsole, tr( "Console" ) );
}

bool
SimikonProject::CreateProject( const ProjectData &pdata )
{
  ProjectData *data = FindProject( pdata.name );

  if ( data ) {
    SC_WARNING( pdata.name + tr( " project already loaded!" ) );
    return false;
  }

  if ( project->childCount() == 0 ) {
    project->setEnabled( true );
    simikonDebugger->show();
    SA_ENABLED( "Settings Close", true );
  }

  Project *prj = new Project( project,
                              simikonDebugger->createDisassembler(),
                              simikonDebugger->createEditor(),
                              simikonDebugger->createListing(),
                              simikonPerformance->createPerformance(),
                              pdata );

  connect( prj, SIGNAL( signalProjectSelected( Project * ) ),
           this, SLOT( SlotChangeActiveProject( Project * ) ) );

  connect( prj->editor, SIGNAL( signalHelp( const char * ) ),
           simikonHelp, SLOT( slotHelp( const char * ) ) );

  if ( pdata.isNew ) {
    prj->actionState[ ACTION_SAVE_E ] = true;
  }

  SetActiveProject( prj );
  InstallActiveProject();

  if ( SS_GET_BOOL( "remember_latest_projects" ) )
    RecreateLatestProjectsList( pdata );

  return true;
}

void
SimikonProject::InstallActiveProject( void )
{
  QFile asmFile( FFP(activeProject->pData.asmFile, activeProject->pData.dir));
  QFile hexFile( FFP(activeProject->pData.hexFile, activeProject->pData.dir));

  if ( !asmFile.exists() && !hexFile.exists() ) {

    switch ( QMessageBox::warning( this, "Simikon",
                                   "Project files are missing!!!\n"
                                   "Do you want to create them?\n\n",
                                   "Yes", "No", 0, 1 ) ) {

    case 0:
      asmFile.open( IO_WriteOnly );
      hexFile.open( IO_WriteOnly );

      asmFile.writeBlock( "\n", 1 );
      hexFile.writeBlock( ":00000001FF", 11 ); // End of Record

      asmFile.close();
      hexFile.close();
      break;

    case 1:
      break;
    }
  }

  INSTALL_RETTYPE retval = controller->install( activeProject, !assembling );

  switch ( retval ) {

  case OK:
    activeProject->hexLoaded = true;
    activeProject->asmLoaded = true;
   
    activeProject->disassembler->setValid( true );
    activeProject->editor->setValid( true );
    activeProject->listing->setValid( true );

    activeProject->actionState[ ACTION_ASSEMBLE_E ]        = false;
    activeProject->actionState[ ACTION_ASSEMBLE_EDITOR_E ] = false;
    activeProject->actionState[ ACTION_DEBUGGER_E ]        = true;
    activeProject->actionState[ ACTION_START_E ]           = true;

    SC_MESSAGE( activeProject->pData.name + 
                tr( " project successfully loaded!" ) );
    break;

  case HEX_ERROR:
    activeProject->hexLoaded = false;
    activeProject->asmLoaded = true;

    activeProject->actionState[ ACTION_ASSEMBLE_E ]        = true;
    activeProject->actionState[ ACTION_ASSEMBLE_EDITOR_E ] = true;
    activeProject->actionState[ ACTION_DEBUGGER_E ]        = true;
    activeProject->actionState[ ACTION_START_E ]           = false;

    SC_ERROR( tr( "Error occured while loading HEX file!" ) );

    break;

  case ASM_ERROR:
    activeProject->hexLoaded = true;
    activeProject->asmLoaded = false;

    activeProject->actionState[ ACTION_ASSEMBLE_E ] = false;
    activeProject->actionState[ ACTION_ASSEMBLE_EDITOR_E ] = false;
    activeProject->actionState[ ACTION_START_E ] = true;
    activeProject->actionState[ ACTION_DEBUGGER_E ] = true;

    SC_ERROR( tr( "Error occured while loading " ) + 
              activeProject->pData.asmFile + " !" );
    break;

  case FUCKED:
    activeProject->hexLoaded = false;
    activeProject->asmLoaded = false;

    activeProject->actionState[ ACTION_DEBUGGER_E ] = false;
    activeProject->actionState[ ACTION_START_E ] = false;
    activeProject->actionState[ ACTION_ASSEMBLE_E ] = false;
    activeProject->actionState[ ACTION_ASSEMBLE_EDITOR_E ] = false;

    SC_ERROR( tr( "Project's " ) + activeProject->pData.asmFile + 
              tr( " and " ) + activeProject->pData.hexFile + 
              tr( " files are missing!" ) );
    break;
  };

  SimikonProject::RefreshActions();
}

void
SimikonProject::UninstallActiveProject( void )
{
  UninstallProject( activeProject );
}

void
SimikonProject::UninstallProject( Project *p )
{
  simikonDebugger->removeProject( p->disassembler, p->editor, p->listing );
  simikonPerformance->removePerformance( p->performance );

  if ( project->childCount() == 1 ) {

    QString projectFile( activeProject->pData.dir + "/" +
                         activeProject->pData.name + ".sim" );

    SS_PUT( "last_project", projectFile );
    activeProject = 0;

  } else {

    QListViewItem *item = activeProject->itemAbove();

    if ( !item )
      item = activeProject->itemBelow();

    SetActiveProject( dynamic_cast<Project *>( item ) );
  }

  SaveProject( p->pData );

  delete p;
}

void
SimikonProject::SetActiveProject( Project *prj )
{
  if ( project->childCount() > 1 ) {
    activeProject->setPixmap( 0, QPixmap( project_xpm ) );
    activeProject->setSelected( false );
  }

  prj->setSelected( true );
  prj->setPixmap( 0, QPixmap( project_active_xpm ) );
 
  project->setCurrentItem( prj );

  SetActiveDisassembler( prj );
  SetActiveEditor( prj );
  SetActiveListing( prj );
  SetActivePerformance( prj );

  simikonDebugger->setActive( prj->disassembler, prj->pData.hexFile,
                              prj->editor, prj->pData.asmFile,
                              prj->listing, prj->pData.lstFile );

  simikonMemory->setActiveMemory( prj->pData.memory );

  controller->setActive( prj );

  activeProject = prj;
}

void
SimikonProject::SetActiveDisassembler( Project *prj )
{
  // disconnect old disassembler
  disconnect( controller, SIGNAL( signalAddInstruction( const QString &, 
                                                        const QString &,
                                                        const QString &,
                                                        const QString & ) ), 
              0, 0 );
  disconnect( controller, SIGNAL( signalSelectInstruction( uint ) ), 
              0, 0 );
  disconnect( controller, SIGNAL( signalMarkInstruction( uint ) ), 
              0, 0 );

  // connect new disassembler
  connect( controller, 
           SIGNAL( signalAddInstruction( const QString &, 
                                         const QString &, 
                                         const QString &,
                                         const QString & ) ),
           prj->disassembler, 
           SLOT( slotAdd( const QString &, const QString &, const QString &,
                          const QString & ) ) );
  connect( controller, SIGNAL( signalSelectInstruction( uint ) ),
           prj->disassembler, SLOT( slotSelect( uint ) ) );
  connect( controller, SIGNAL( signalMarkInstruction( uint ) ),
           prj->disassembler, SLOT( slotMark( uint ) ) );
}

void
SimikonProject::SetActiveEditor( Project *prj )
{
  // disconnect old editor
  disconnect( controller, SIGNAL( signalAddEditorLine( const QString & ) ), 
              0, 0 );
  disconnect( controller, SIGNAL( signalSelectEditorLine( uint ) ), 
              0, 0 );

  // connect new editor
  connect( controller, SIGNAL( signalAddEditorLine( const QString & ) ),
           prj->editor, SLOT( slotAddLine( const QString & ) ) );
  connect( controller, SIGNAL( signalSelectEditorLine( uint ) ),
           prj->editor, SLOT( slotSelectLine( uint ) ) );
}

void
SimikonProject::SetActiveListing( Project *prj )
{
  // disconnect old listing
  disconnect( controller, SIGNAL( signalAddListingLine( const QString & ) ),
              0, 0 );
  disconnect( controller, SIGNAL( signalClearListing( void ) ),
              0, 0 );

  // connect new listing
  connect( controller, SIGNAL( signalAddListingLine( const QString & ) ),
           prj->listing, SLOT( slotAddLine( const QString & ) ) );
  connect( controller, SIGNAL( signalClearListing( void ) ),
           prj->listing, SLOT( clear( void ) ) );
}

void
SimikonProject::SetActivePerformance( Project *prj )
{
  simikonPerformance->setActivePerformance( prj->performance, "Performance" );

  // disconnect old performance
  disconnect( simikonPerformance, SIGNAL( signalAdd( QListViewItem *, bool  ) ), 
              0, 0 );
  disconnect( simikonPerformance, SIGNAL( signalAddAuto( void ) ), 
              0, 0 );
  disconnect( simikonPerformance, SIGNAL( signalRemove( QListViewItem * ) ), 
              0, 0 );
  disconnect( simikonPerformance, SIGNAL( signalReset( void ) ), 
              0, 0 );
  
  // connect new performance
  connect( simikonPerformance, SIGNAL( signalAdd( QListViewItem *, bool ) ),
           prj, SLOT( slotAddPerformance( QListViewItem *, bool ) ) );
  connect( simikonPerformance, SIGNAL( signalAddAuto( void ) ),
           prj, SLOT( slotAddAutoPerformance( void ) ) );
  connect( simikonPerformance, SIGNAL( signalRemove( QListViewItem * ) ),
           prj, SLOT( slotRemovePerformance( QListViewItem * ) ) );
  connect( simikonPerformance, SIGNAL( signalReset( void ) ),
           prj, SLOT( slotResetPerformance( void ) ) );
}

bool
SimikonProject::SaveProject( const ProjectData &pdata )
{
  QDir dir;
  int index = 0;

  do {
    index = pdata.dir.find( '/', index + 1 );


    QString tmpDirPath( pdata.dir );
#ifdef SIMIKON_UNIX
    tmpDirPath.truncate( index );
#endif

#ifdef SIMIKON_WIN32
    tmpDirPath.truncate( index + 1 );
#endif

    SC_MESSAGE( tmpDirPath );

    if ( !dir.exists( tmpDirPath ) ) {
      SC_MESSAGE( "***" );
      SC_MESSAGE( tmpDirPath );
      SC_MESSAGE( "***" );

      if ( !dir.mkdir( tmpDirPath ) ) {
        SC_ERROR( tr( "Couldn't create project dir!" ) );
        return false;
      }
    }
  } while ( index != -1 );

  QFile fp( pdata.dir + "/" + pdata.name + ".sim" );
  QDataStream dfp( &fp );
  
  if ( fp.open( IO_WriteOnly ) ) {

    dfp << pdata.author;
    dfp << pdata.name;
    dfp << pdata.dir;
    dfp << pdata.hexFile;
    dfp << pdata.asmFile;
    dfp << pdata.lstFile;
    dfp << pdata.date;

    dfp << pdata.controller.clock;
    dfp << pdata.controller.programMemory;

    dfp << pdata.memory.iram;
    dfp << pdata.memory.irom;
    dfp << pdata.memory.xram;
    dfp << pdata.memory.xrom;
    dfp << pdata.memory.iramName;
    dfp << pdata.memory.iramStart;
    dfp << pdata.memory.iramEnd;
    dfp << pdata.memory.iromName;
    dfp << pdata.memory.iromStart;
    dfp << pdata.memory.iromEnd;
    dfp << pdata.memory.xramName;
    dfp << pdata.memory.xramStart;
    dfp << pdata.memory.xramEnd;
    dfp << pdata.memory.xromName;
    dfp << pdata.memory.xromStart;
    dfp << pdata.memory.xromEnd;

    for ( uint i = 0; i < 8; i++ ) {
      dfp << pdata.moduleState[ i ];
    }

    fp.close();
  } else {
    SC_ERROR( tr( "Error occured while trying to save " ) + pdata.name + 
              tr( " project!!!" ) );
    return false;
  }

  return true;
}

void
SimikonProject::ConnectAllModules( void )
{
  for ( uint i = 0; i < NO_OF_MODULES; i++ ) {
    connect( module[ i ].module, SIGNAL( signalClosed() ), 
             this, SLOT( SlotModuleClosed() ) );
  }
}

void
SimikonProject::DisconnectAllModules( void )
{
  for ( uint i = 0; i < NO_OF_MODULES; i++ ) {
    disconnect( module[ i ].module, SIGNAL( signalClosed() ),
                this, SLOT( SlotModuleClosed() ) );
  }
}

ProjectData *
SimikonProject::FindProject( const QString &name )
{
  QListViewItem *i = project->firstChild();

  while ( i ) {
    if ( i->text( 0 ) == name )
      return &( dynamic_cast<Project *>( i ) )->pData;

    i = i->nextSibling();
  }

  return 0;  
}

void
SimikonProject::Refresh( void )
{
  activeProject->setText( 0, activeProject->pData.name );
  activeProject->setText( 1, activeProject->pData.author );
  activeProject->setText( 2, activeProject->pData.date );

  SetActiveProject( activeProject );
}

void
SimikonProject::RefreshActions( void )
{
  for ( uint i = 0; i < NO_OF_ACTIONS; i++ )
    if ( i != ACTION_DEBUGGER_E )
      SA_ENABLED( ACTIONS_TABLE[ i ], activeProject->actionState[ i ] );
 
  if ( activeProject->actionState[ ACTION_DEBUGGER_E ] ) {
    SA_ENABLED( "StartSim", true );
  } else {
    SA_ENABLED( "StartSim", false );
  }
}

void
SimikonProject::RecreateLatestProjectsList( const ProjectData &pdata )
{
   QString stringToAdd( pdata.dir + "/" + pdata.name + ".sim" );

   uint n = SS_GET_INT( "remember_latest_projects_no" );
   uint index = n - 1;

   for ( uint i = 0; i < n; i++ ) {
     if ( SS_GET( QString( "Project%1" ).arg( i ) ) == stringToAdd ) {
       index = i;
       break;
     }
   }

   if ( stringToAdd != SS_GET( "Project0" ) ) {
    
     for ( uint i = index; i > 0; i-- ) {
       QString stringToMove( SS_GET( QString( "Project%1" ).arg( i - 1 ) ) );
  
       if ( !stringToMove.isEmpty() && !stringToMove.isNull() ) {
         QString stringToChange( QString( "Project%1" ).arg( i ) );

         SS_PUT( stringToChange, stringToMove );
         SA_CHANGE_TEXT( stringToChange, stringToMove );
       }
     }

     SS_PUT( "Project0", stringToAdd );
     SA_CHANGE_TEXT( "Project0", stringToAdd );
   }
}

void
SimikonProject::ReloadActiveProject( void )
{
  QString prjFile( activeProject->pData.dir + "/" + 
                   activeProject->pData.name + ".sim" );

  ProjectData pdata = activeProject->pData;
  SaveProject( pdata );
  
  pdata.isNew = false;

  activeProject->reset();
  InstallActiveProject();
}

void
SimikonProject::SlotChangeActiveProject( Project *prj )
{
  if ( ( prj != activeProject ) && ( project->childCount() != 1 ) ) {
    SetActiveProject( prj );
    RefreshActions();
  }
}

void
SimikonProject::SlotStartDebug( void )
{
  mode = true;

  if ( SA_IS_ENABLED( "Save" ) ) {
    if ( SaveProject( activeProject->pData ) ) {
      activeProject->actionState[ ACTION_SAVE_E ] = false;
      SA_ENABLED( "Save", false );
    } else {
      // poruka o gresci
      return;
    }
  }

  activeProject->editor->setReadOnly( true );

  SA_ENABLED( "Registers SFR Memory Stack IOPorts Serial Interrupts Performance", true );
  SA_ENABLED( "Close Open Reopen New Settings Assemble AssembleEditor StartSim", false );

  project->setEnabled( false );

  ConnectAllModules();

  if ( SS_GET_BOOL( "remember_module_state" ) ) {
    for ( uint i = 0; i < NO_OF_MODULES; i++ ) {
      if ( activeProject->pData.moduleState[ i ] ) {
        module[ i ].module->show();
        SA_CHECKED( module[ i ].name, true );
      }
    }
  }
}

void
SimikonProject::SlotModuleClosed( void )
{
  for ( uint i = 0; i < NO_OF_MODULES; i++ ) {
    if ( sender() == module[ i ].module ) {
      activeProject->pData.moduleState[ i ] = false;
      break;
    }
  }
}

void
SimikonProject::SlotAssemblerProjectAssembled( void )
{
  activeProject->reset( false );
  SimikonProject::InstallActiveProject();
  SimikonProject::Refresh();

  assembling = false;
  project->setEnabled( true );
}

void
SimikonProject::SlotAssemblerSyntaxError( void )
{
  cout << "DEBUG: SimikonProject::SlotAssemblerSyntaxError()" << endl;

  cout << "errors.count() = " << errors.count() << endl;

  for ( uint i = 0; i < errors.count(); i++ ) {
    cout << "==> " << errors[i].type << endl;
    cout << "==> " << errors[i].line << endl;
    cout << "==> " << errors[i].message << endl;
  }

  activeProject->reset( false );
  SimikonProject::InstallActiveProject();
  SimikonProject::Refresh();

  assembling = false;
  project->setEnabled( true );
}

void
SimikonProject::SlotPageSelected( QWidget *w )
{
  if ( w == simikonConsole )
    stabctl->setTabTextColor( simikonConsole, black );
}

void
SimikonProject::slotNewProject( void )
{
  ProjectSettings newProject( stabctl, "SimikonNewProject" );
  ProjectData pdata;

  if ( newProject.exec() ) {
    pdata = newProject.getProjectData();
    pdata.isNew = true;

    SaveProject( pdata );
    CreateProject( pdata );
  }
}

void
SimikonProject::slotOpenProject( void )
{
  QString projectFile = QFileDialog::getOpenFileName( QString::null, "*.*" );

  if ( projectFile.isEmpty() )
    return;

  if ( projectFile.find( ".hex" ) != -1 ) {
    ProjectSettings newProject( stabctl, "SimikonNewProject" );

    newProject.entry[ PROJECT_HEX_E ]->setText( projectFile );

    if ( newProject.exec() ) {
      ProjectData pdata;

      pdata = newProject.getProjectData();
      pdata.isNew = true;

      CreateProject( pdata );
    } else {
      SC_WARNING( tr( "New project canceled!" ) );
      return;
    }
  } else if ( projectFile.find( ".sim" ) != -1 ) {
    openProject( projectFile );
    return;
  } else {
    SC_ERROR( tr( "Unknown extension file type!" ) );
    return;
  }
}

void
SimikonProject::slotSaveProject( void )
{
  ProjectData &pdata = activeProject->pData;

  if ( SimikonProject::SaveProject( pdata ) ) {
    activeProject->actionState[ ACTION_SAVE_E ] = false;
    SA_ENABLED( "Save", false );
  } 
}

void
SimikonProject::slotCloseProject( void )
{
  if ( project->childCount() == 1 ) {
    simikonDebugger->close();

    project->setEnabled( false );
 
    SA_ENABLED( "Settings Save Close Assemble AssembleEditor", false );
    SA_ENABLED( "StartSim StartDebug Stop", false ); 
  }

  SC_MESSAGE( activeProject->pData.name + tr( " project closed!" ) );

  UninstallActiveProject();
}

void
SimikonProject::slotAssembleProject( void )
{
  const char *asmFile = activeProject->pData.asmFile.ascii();
  const char *prjDir  = activeProject->pData.dir.ascii();

  assembling = true;
  project->setEnabled( false );

  simikonDebugger->saveASM( FFP( asmFile, prjDir ) );

  assembler->assemble( &activeProject->pData );

  activeProject->actionState[ ACTION_ASSEMBLE_E ] = false;
  activeProject->actionState[ ACTION_ASSEMBLE_EDITOR_E ] = false;
  SA_ENABLED( "Assemble AssembleEditor", false );
}

void
SimikonProject::slotProjectSettings( void )
{
  projectSettings->setProjectData( activeProject->pData );
  projectSettings->okButton->setEnabled( false );

  if ( projectSettings->exec() ) {
    ProjectData pdata = projectSettings->getProjectData();

    for ( uint i = 0; i < NO_OF_MODULES; i++ )
      pdata.moduleState[i] = activeProject->pData.moduleState[i];

    activeProject->pData = pdata;

    // this sucks
    activeProject->actionState[ ACTION_SAVE_E ] = true;
    SA_ENABLED( "Save", true );

    if ( projectSettings->reload ) {
      SimikonProject::ReloadActiveProject();
    } else {
      SimikonProject::Refresh();
    }

    SC_MESSAGE( tr( "Project settings changed!" ) );
  }
}

void
SimikonProject::slotToggleRegisters( void )
{
  bool state = SA_IS_CHECKED( "Registers" );

  if ( state ) {
    simikonRegisters->show();
  } else {
    simikonRegisters->close();
  }

  activeProject->pData.moduleState[ MODULE_REGISTERS_E ] = state;
}

void
SimikonProject::slotToggleSFR( void )
{
  bool state = SA_IS_CHECKED( "SFR" );

  if ( state ) {
    simikonSFR->show();
  } else {
    simikonSFR->close();
  }
  
  activeProject->pData.moduleState[ MODULE_SFR_E ] = state;
}

void
SimikonProject::slotToggleMemory( void )
{
  bool state = SA_IS_CHECKED( "Memory" );

  if ( state ) {
    simikonMemory->show();
  } else {
    simikonMemory->close();
  }

  activeProject->pData.moduleState[ MODULE_MEMORY_E ] = state;
}

void
SimikonProject::slotToggleStack( void )
{
  bool state = SA_IS_CHECKED( "Stack" );

  if ( state ) {
    simikonStack->show();
  } else {
    simikonStack->close();
  }

  activeProject->pData.moduleState[ MODULE_STACK_E ] = state;
}

void
SimikonProject::slotTogglePerformance( void )
{
  bool state = SA_IS_CHECKED( "Performance" );

  if ( state ) {
    simikonPerformance->show();
  } else {
    simikonPerformance->close();
  }

  activeProject->pData.moduleState[ MODULE_PERFORMANCE_E ] = state;
}

void
SimikonProject::slotToggleIOPorts( void )
{
  bool state = SA_IS_CHECKED( "IOPorts" );

  if ( state ) {
    simikonIOPorts->show();
  } else {
    simikonIOPorts->close();
  }

  activeProject->pData.moduleState[ MODULE_IOPORTS_E ] = state;
}

void
SimikonProject::slotToggleSerial( void )
{
  bool state = SA_IS_CHECKED( "Serial" );

  if ( state ) {
    simikonSerial->show();
  } else {
    simikonSerial->close();
  }

  activeProject->pData.moduleState[ MODULE_SERIAL_E ] = state;
}

void
SimikonProject::slotToggleInterrupts( void )
{
  bool state = SA_IS_CHECKED( "Interrupts" );

  if ( state ) {
    simikonInterrupts->show();
  } else {
    simikonInterrupts->close();
  }

  activeProject->pData.moduleState[ MODULE_INTERRUPTS_E ] = state;
}

void
SimikonProject::slotToggleDebugger( void )
{
  bool state = SA_IS_CHECKED( "Debugger" );

  if ( state ) {
    simikonDebugger->show();
  } else {
    simikonDebugger->close();
  }
}

void
SimikonProject::slotToggleConsole( void )
{
  bool state = SA_IS_CHECKED( "Console" );

  if ( state ) {
    stabctl->insertPage( simikonConsole, tr( "Console" ) );
    if ( !stabctl->isVisible() )
      stabctl->show();
  } else {
    stabctl->removePage( simikonConsole );

    if ( !SA_IS_CHECKED( "Projects" ) )
      stabctl->hide();
  }

  SS_PUT( "show_console", state );
}

void
SimikonProject::slotToggleProjects( void )
{
  bool state = SA_IS_CHECKED( "Projects" );

  if ( state ) {
    stabctl->insertPage( project, tr( "Projects" ) );

    // project is always on first page
    if ( SA_IS_CHECKED( "Console" ) ) {
      stabctl->removePage( simikonConsole );
      stabctl->insertPage( simikonConsole, tr( "Console" ) );
    }

    if ( !stabctl->isVisible() )
      stabctl->show();
  } else {
    stabctl->removePage( project );

    if ( !SA_IS_CHECKED( "Console" ) )
      stabctl->hide();
  }

  SS_PUT( "show_project", state );
}

void
SimikonProject::slotAboutSimikon( void )
{
  simikonAbout->exec();
}

void
SimikonProject::slotInstructionSetHelp( void )
{
  bool state = SA_IS_CHECKED( "InstructionSetHelp" );

  if ( state ) {
    simikonHelp->show();
  } else {
    simikonHelp->close();
  }
}

void
SimikonProject::slotStartSimulator( void )
{
  SA_ENABLED( "Stop", true );
  SA_ENABLED( "StartDebug", false );

  controller->startSimulation();
  SlotStartDebug();

  mode = false; // ugly hack
}

void
SimikonProject::slotStop( void )
{
  if ( mode ) { // debugger mode
    mode = false;  
    simikonDebugger->stop();
  } else { // simulator
    controller->stopSimulation();
    activeProject->editor->deselect();
    SA_ENABLED( "StartSim Open New", true );
  }

  controller->reset();
  
  SA_ENABLED( "Registers SFR Memory Stack IOPorts Serial Interrupts "
              "Performance", false );
  SA_CHECKED( "Registers SFR Memory Stack IOPorts Serial Interrupts "
              "Performance", false );
  SA_ENABLED( "Close Open Reopen New Settings", true );
  SA_ENABLED( "Stop", false );

  project->setEnabled( true );
  
  DisconnectAllModules();
  
  simikonRegisters->close();
  simikonSFR->close();
  simikonMemory->close();
  simikonStack->close();
  simikonIOPorts->close();
  simikonSerial->close();
  simikonInterrupts->close();
  simikonPerformance->close();
  
  activeProject->editor->setReadOnly( false );
  
  RefreshActions();  
}

void
SimikonProject::slotOpenProject0( void ) { 
  SimikonProject::openProject( SS_GET( "Project0" ) );
}

void
SimikonProject::slotOpenProject1( void ) {
  SimikonProject::openProject( SS_GET( "Project1" ) ); 
}

void
SimikonProject::slotOpenProject2( void ) {
  SimikonProject::openProject( SS_GET( "Project2" ) );
}

void
SimikonProject::slotOpenProject3( void ) {
  SimikonProject::openProject( SS_GET( "Project3" ) );
}

void
SimikonProject::slotOpenProject4( void ) {
  SimikonProject::openProject( SS_GET( "Project4" ) );
}

void
SimikonProject::slotOpenProject5( void ) {
  SimikonProject::openProject( SS_GET( "Project5" ) );
}

void
SimikonProject::slotOpenProject6( void ) {
  SimikonProject::openProject( SS_GET( "Project6" ) );
}

void
SimikonProject::slotOpenProject7( void ) {
  SimikonProject::openProject( SS_GET( "Project7" ) );
}

void
SimikonProject::slotOpenProject8( void ) {
  SimikonProject::openProject( SS_GET( "Project8" ) );
}

void
SimikonProject::slotOpenProject9( void ) {
  SimikonProject::openProject( SS_GET( "Project9" ) );
}


/*===========================================================================*/
/* class: Project                                                            */
/*===========================================================================*/

Project::Project( QListView *parent,
                  Disassembler *d,
                  Editor *e,
                  Listing *l,
                  Performance *p,
                  const ProjectData &pdata )
  : QListViewItem( parent )
{
  setText( 0, pdata.name );
  setText( 1, pdata.author );
  setText( 2, pdata.date );

  hexLoaded          = false;
  asmLoaded          = false;
  analyzePerformance = false;

  pData              = pdata;

  // DISASSEMBLER
  disassembler = d;
  connect( disassembler, SIGNAL( signalSetBreakPoint( uint, bool ) ),
           this, SLOT( SlotSetBreakPoint( uint, bool ) ) );

  // EDITOR
  editor = e;
  connect( editor, SIGNAL( signalChanged() ),
           this, SLOT( SlotASMChanged() ) );

  // LISTING
  listing = l;
  
  // PERFORMANCE
  performance = p;

  performanceData = new QList<PerformanceData>;
  performanceData->setAutoDelete( true );

  program = new InstructionData[ 64 * 1024 ]; // 64k 
  programLength = 0;

  for ( uint i = 0; i < NO_OF_ACTIONS; i++ ) {
    actionState[ i ] = false;
  }
}

Project::~Project()
{
  delete performanceData;

  if ( disassembler ) { delete disassembler; disassembler = 0; }
  if ( editor ) { delete editor; editor = 0; }
  if ( listing ) { delete listing; listing = 0; }
  if ( performance ) { delete performance; performance = 0; }
  if ( program ) { delete [] program; program = 0; }
}

void
Project::reset( bool resetASM )
{
  hexLoaded          = false;
  asmLoaded          = false;
  analyzePerformance = false;

  performanceData->clear();

  if ( program ) { delete program; program = 0; }

  program = new InstructionData[ 64 * 1024 ];
  CHECK_PTR( program );

  programLength = 0;

  slotResetPerformance();
//  performance->reset();
  disassembler->reset();
  editor->reset( resetASM );

  for ( uint i = 0; i < NO_OF_ACTIONS; i++ ) {
    actionState[ i ] = false;
  }
}

void
Project::activate( void )
{
  emit signalProjectSelected( this );
}

void
Project::slotAddPerformance( QListViewItem *i, bool dynamic )
{
  PerformanceData *pdata = new PerformanceData;

  pdata->item     = i;
  pdata->cycles   = 0;
  pdata->depth    = 0;
  pdata->counting = false;
  pdata->dynamic  = dynamic;

  performanceData->append( pdata );

  if ( dynamic == false ) { // find symbol for start and end address
    QString asmFilePath = FFP( pData.asmFile, pData.dir );

    uint start = strToUInt( pdata->item->text( 1 ) );
    uint end   = strToUInt( pdata->item->text( 3 ) );

    QString startSymbol = getSymbol( asmFilePath, program[ start ].lineASM );
    QString endSymbol   = getSymbol( asmFilePath, program[ end ].lineASM );

    pdata->item->setText( 0, startSymbol );
    pdata->item->setText( 2, endSymbol );
  }

  analyzePerformance = true;
}

// Pronalazi sve pocetne adrese svih procedura

void
Project::slotAddAutoPerformance( void )
{
  uint address = 0;

  while ( address < programLength ) {

    // acall
    if ( program[ address ].opCode == 0x11 ||
         program[ address ].opCode == 0x31 ||
         program[ address ].opCode == 0x51 ||
         program[ address ].opCode == 0x71 ||
         program[ address ].opCode == 0x91 ||
         program[ address ].opCode == 0xB1 ||
         program[ address ].opCode == 0xD1 ||
         program[ address ].opCode == 0xF1 ) { 

      QString symbol = getSymbol( FFP( pData.asmFile, pData.dir ), 
                                  program[ address ].lineASM );

      QListViewItem *i = performance->add( address, symbol );

      slotAddPerformance( i, true );
    }
    
    if ( program[ address ].execute != 0 ) {
      address += program[ address ].bytes;
    } else {
      address++;
    }
  }

}

void
Project::slotRemovePerformance( QListViewItem *i )
{
  PerformanceData *pdata = FindPerformanceData( i );

  if ( pdata ) {
    performanceData->remove( pdata );
  }

  if ( performanceData->isEmpty() ) {
    analyzePerformance = false;
  }
}

void
Project::slotResetPerformance( void )
{
  QListIterator<PerformanceData> it( *performanceData );
  PerformanceData *pdata;

  while ( ( pdata = it.current() ) != 0 ) {
    pdata->cycles   = 0;
    pdata->depth    = 0;
    pdata->counting = false;

    pdata->item->setText( 4, "" );
    pdata->item->setText( 5, "" );

    if ( pdata->dynamic ) {
      pdata->item->setText( 2, "" );
      pdata->item->setText( 3, "" );
    }

    ++it;
  }
}

void
Project::SlotSetBreakPoint( uint address, bool state )
{
  program[ address ].isBreak = state;
}

void
Project::SlotASMChanged( void )
{
  actionState[ ACTION_ASSEMBLE_E ]        = true;
  actionState[ ACTION_ASSEMBLE_EDITOR_E ] = true;
  actionState[ ACTION_START_E ]           = false;
}

PerformanceData *
Project::FindPerformanceData( QListViewItem *i )
{
  QListIterator<PerformanceData> it( *performanceData );
  PerformanceData *pdata;
  
  while ( ( pdata = it.current() ) != 0 ) {
    if ( pdata->item == i ) {
      return pdata;
    }
    ++it;
  }

  return 0;
}

/*===========================================================================*/
/* class: ProjectSettings                                                    */
/*===========================================================================*/

ProjectSettings::ProjectSettings( QWidget *parent, const char *name )
  : QDialog( parent, name, true )
{
  setCaption( tr( "Project Settings" ) );

  isNew  = true;
  reload = false;

  QBoxLayout *topLayout = new QVBoxLayout( this );
  STabCtl *stabctl = new STabCtl( this, "STabCtl" );
  topLayout->addWidget( stabctl, 10 );

  InitGeneralTab( stabctl );
  InitControllerTab( stabctl );

  QFrame *frame = new QFrame( this );
  frame->setFrameStyle( QFrame::WinPanel | QFrame::Raised );
  frame->setLineWidth( 2 );
  topLayout->addWidget( frame );

  QWidget *buttons = new QWidget( frame, "Buttons" );
  QBoxLayout *hboxLayout = new QHBoxLayout( frame, 2 );
  hboxLayout->addWidget( buttons );

  InitButtons( buttons );

  setFixedSize( 310, 248 );
}

ProjectSettings::~ProjectSettings()
{

}

ProjectData
ProjectSettings::getProjectData( void )
{
  ProjectData pdata;

  pdata.author  = entry[ PROJECT_AUTHOR_E ]->text();
  pdata.name    = entry[ PROJECT_NAME_E ]->text();
  pdata.dir     = entry[ PROJECT_DIR_E ]->text();
  pdata.hexFile = entry[ PROJECT_HEX_E ]->text();
  pdata.asmFile = entry[ PROJECT_ASM_E ]->text();
  pdata.lstFile = entry[ PROJECT_LST_E ]->text();

  pdata.controller.clock         = clock->text();
  pdata.controller.programMemory = rb1->isChecked();
  pdata.date                     = QDate::currentDate().toString();

  pdata.memory.iram = cb_mem[0]->isChecked();
  pdata.memory.irom = cb_mem[1]->isChecked();
  pdata.memory.xram = cb_mem[2]->isChecked();
  pdata.memory.xrom = cb_mem[3]->isChecked();

  pdata.memory.iramName  = table->text( 0, 0 );
  pdata.memory.iramStart = table->text( 0, 1 );
  pdata.memory.iramEnd   = table->text( 0, 2 );

  pdata.memory.iromName  = table->text( 1, 0 );
  pdata.memory.iromStart = table->text( 1, 1 );
  pdata.memory.iromEnd   = table->text( 1, 2 );

  pdata.memory.xramName  = table->text( 2, 0 );
  pdata.memory.xramStart = table->text( 2, 1 );
  pdata.memory.xramEnd   = table->text( 2, 2 );

  pdata.memory.xromName  = table->text( 3, 0 );
  pdata.memory.xromStart = table->text( 3, 1 );
  pdata.memory.xromEnd   = table->text( 3, 2 );

  for ( uint i = 0; i < 10; i++ ) {
    pdata.moduleState[ i ] = 0;
  }

  return pdata;
}

void
ProjectSettings::setProjectData( const ProjectData &pdata )
{
  entry[ PROJECT_AUTHOR_E ]->setText( pdata.author );
  entry[ PROJECT_NAME_E ]->setText( pdata.name );
  entry[ PROJECT_DIR_E ]->setText( pdata.dir );
  entry[ PROJECT_HEX_E ]->setText( pdata.hexFile );
  entry[ PROJECT_ASM_E ]->setText( pdata.asmFile );
  entry[ PROJECT_LST_E ]->setText( pdata.lstFile );

  clock->setValue( pdata.controller.clock.toInt() );

  ( pdata.controller.programMemory ) ?
    rb1->setChecked( true ) :
      rb2->setChecked( true );

  cb_mem[0]->setChecked( pdata.memory.iram );
  cb_mem[1]->setChecked( pdata.memory.irom );
  cb_mem[2]->setChecked( pdata.memory.xram );
  cb_mem[3]->setChecked( pdata.memory.xrom );

  table->setText( 0, 0, pdata.memory.iramName );
  table->setText( 0, 1, pdata.memory.iramStart );
  table->setText( 0, 2, pdata.memory.iramEnd );

  table->setText( 1, 0, pdata.memory.iromName );
  table->setText( 1, 1, pdata.memory.iromStart );
  table->setText( 1, 2, pdata.memory.iromEnd );

  table->setText( 2, 0, pdata.memory.xramName );
  table->setText( 2, 1, pdata.memory.xramStart );
  table->setText( 2, 2, pdata.memory.xramEnd );

  table->setText( 3, 0, pdata.memory.xromName );
  table->setText( 3, 1, pdata.memory.xromStart );
  table->setText( 3, 2, pdata.memory.xromEnd );

  isNew  = false;
  reload = false; // drugo rjesenje bi bilo da disconnectam slot: 
                  // SlotSettingsChanged sa svih entry[...]-a.
                  // pa ih ponovo reconectam kada popuni entry-e
}

void
ProjectSettings::InitGeneralTab( STabCtl *stabctl )
{
  QWidget *generalTab = new QWidget( stabctl, "ProjectSettingsGeneralTab" );
  QGridLayout *gridLayout = new QGridLayout( generalTab, 7, 3, 3 );

  QLabel *label;
  QPushButton *button;

  // Project Name
  entry[ PROJECT_NAME_E ] = new QLineEdit( generalTab );
  connect( entry[ PROJECT_NAME_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotSettingsChanged( void ) ) );
  connect( entry[ PROJECT_NAME_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotProjectNameChanged( const QString & ) ) );
  label = new QLabel( entry[ PROJECT_NAME_E ], tr( "&Project: " ), generalTab );

  gridLayout->addWidget( label, 0, 0 );
  gridLayout->addWidget( entry[PROJECT_NAME_E], 0, 1 );

  // Author Name
  entry[ PROJECT_AUTHOR_E ] = new QLineEdit( generalTab );
  connect( entry[ PROJECT_AUTHOR_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotSettingsChanged( void ) ) );
  label = new QLabel( entry[PROJECT_AUTHOR_E], tr( "&Author: " ), generalTab );

  gridLayout->addWidget( label, 1, 0 );
  gridLayout->addWidget( entry[PROJECT_AUTHOR_E], 1, 1 );

  // Project Dir
  entry[ PROJECT_DIR_E ] = new QLineEdit( generalTab );
  entry[ PROJECT_DIR_E ]->setText( SS_GET( "projects_dir" ) );
  connect( entry[ PROJECT_DIR_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotSettingsChanged( void ) ) );
  label = new QLabel( entry[ PROJECT_DIR_E ], tr( "&Directory: " ), generalTab );

  button = new QPushButton( generalTab );
  button->setPixmap( folder_xpm );
  connect( button, SIGNAL( clicked() ),
           this, SLOT( SlotDirectoryButtonClicked() ) );

  gridLayout->addWidget( label, 2, 0 );
  gridLayout->addWidget( entry[PROJECT_DIR_E], 2, 1 );
  gridLayout->addWidget( button, 2, 2 );

  // Hex
  entry[ PROJECT_HEX_E ] = new QLineEdit( generalTab );
  connect( entry[ PROJECT_HEX_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotSettingsChanged( void ) ) );
  label = new QLabel( entry[ PROJECT_HEX_E ], tr( "&Hex: " ), generalTab );

  button = new QPushButton( generalTab );
  button->setPixmap( QPixmap( file_xpm ) );
  connect( button, SIGNAL( clicked() ),
           this, SLOT( SlotHexButtonClicked() ) );

  gridLayout->addWidget( label, 3, 0 );
  gridLayout->addWidget( entry[ PROJECT_HEX_E ], 3, 1 );
  gridLayout->addWidget( button, 3, 2 );

  // Asm
  entry[ PROJECT_ASM_E ] = new QLineEdit( generalTab );
  connect( entry[ PROJECT_ASM_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotSettingsChanged( void ) ) );
  label = new QLabel( entry[ PROJECT_ASM_E ], tr( "&Asm: " ), generalTab );

  button = new QPushButton( generalTab );
  button->setPixmap( QPixmap( file_xpm ) );
  connect( button, SIGNAL( clicked() ),
           this, SLOT( SlotAsmButtonClicked() ) );

  gridLayout->addWidget( label, 4, 0 );
  gridLayout->addWidget( entry[ PROJECT_ASM_E ], 4, 1 );
  gridLayout->addWidget( button, 4, 2 );

  // lst
  entry[ PROJECT_LST_E ] = new QLineEdit( generalTab );
  connect( entry[ PROJECT_LST_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotSettingsChanged( void ) ) );
  label = new QLabel( entry[ PROJECT_LST_E ], tr( "&Listing: " ), generalTab );

  button = new QPushButton( generalTab );
  button->setPixmap( QPixmap( file_xpm ) );
  connect( button, SIGNAL( clicked() ),
           this, SLOT( SlotLstButtonClicked() ) );

  gridLayout->addWidget( label, 5, 0 );
  gridLayout->addWidget( entry[ PROJECT_LST_E ], 5, 1 );
  gridLayout->addWidget( button, 5, 2 );

  gridLayout->setRowStretch( 6, 10 );

  stabctl->insertPage( generalTab, tr( "General" ) );
}

void
ProjectSettings::InitControllerTab( STabCtl *stabctl )
{
  QWidget *controllerTab = new QWidget( stabctl, "ProjectSettingsControllerTab" );
  QVBoxLayout *vbox = new QVBoxLayout( controllerTab );

  QGridLayout *gridLayout = new QGridLayout( vbox, 4, 3, 3 );

  gridLayout->setColStretch( 1, 10 );

  QLabel *label;
  QHBoxLayout *hbox;

  // Clock
  clock = new QSpinBox( 1, 16000000, 1000, controllerTab );
  clock->setValue( 12000000 );
  connect( clock, SIGNAL( valueChanged( int ) ),
           this, SLOT( SlotSettingsChanged( void ) ) );
  label = new QLabel( clock, tr( " &Clock (Hz): " ), controllerTab );

  gridLayout->addWidget( label, 1, 0 );
  gridLayout->addWidget( clock, 1, 1 );

  QGroupBox *memoryGroupBox = new QGroupBox( 1, Qt::Vertical, tr( "Memory" ), 
                                             controllerTab );

  table = new QTable( memoryGroupBox );

  table->setNumRows( 4 );
  table->setNumCols( 4 );
  table->setShowGrid( false );

  QHeader *header;

  // horizontal header stuff
  header = table->horizontalHeader();
  header->setLabel( 0, tr( "Type" ) );
  header->setLabel( 1, tr( "Start" ) );
  header->setLabel( 2, tr( "End" ) );
  header->setLabel( 3, "" );
  header->setResizeEnabled( false );
  header->setMovingEnabled( false );

  // vertical header stuff
  header = table->verticalHeader();
  table->setLeftMargin( 0 );
  header->hide();

  table->setColumnWidth( 0, 108 );
  table->setColumnWidth( 1, 80 );
  table->setColumnWidth( 2, 80 );
  table->setColumnWidth( 3, 12 );

  for ( uint i = 0; i < 5; i++ ) {
    cb_mem[i] = new QCheckBox( table );
    connect( cb_mem[i], SIGNAL( clicked() ), this, SLOT( SlotSettingsChanged() ) );
    table->setCellWidget( i, 3, cb_mem[i] );
  }

  // DEFAULT MEMORY SETTINGS

  // Internal RAM
  table->setText( 0, 0, "IRAM" );
  table->setText( 0, 1, "0x00h" );
  table->setText( 0, 2, "0xFFh" );
  cb_mem[0]->setChecked( true );

  // Internal ROM
  table->setText( 1, 0, "IROM" );
  table->setText( 1, 1, "0x00h" );
  table->setText( 1, 2, "0xFFh" );
  cb_mem[1]->setChecked( false );
  
  // External RAM
  table->setText( 2, 0, "XRAM" );
  table->setText( 2, 1, "0x0000h" );
  table->setText( 2, 1, "0xFFFFh" );
  cb_mem[2]->setChecked( false );

  // External ROM
  table->setText( 3, 0, "XROM" );
  table->setText( 3, 1, "0x0000h" );
  table->setText( 3, 2, "0xFFFFh" );
  cb_mem[3]->setChecked( true );

  vbox->addWidget( memoryGroupBox );

  // Program Memory
  QButtonGroup *bg = new QButtonGroup( 2, Qt::Horizontal, 
                                       tr( "Program Memory" ), controllerTab );
  bg->hide();
  connect( bg, SIGNAL( clicked( int ) ), 
           this, SLOT( SlotProgramMemoryChanged( int ) ) );
  connect( bg, SIGNAL( clicked( int ) ),
           this, SLOT( SlotSettingsChanged( void ) ) );

  gridLayout->addMultiCellWidget( bg, 2, 3, 0, 1 );

  hbox = new QHBoxLayout( vbox );

  label = new QLabel( bg, tr( " &Program Memory: " ), controllerTab );
  hbox->addWidget( label );

  rb1 = new QRadioButton( "IROM  ", controllerTab );
  hbox->addWidget( rb1 );

  rb2 = new QRadioButton( "XROM", controllerTab );
  rb2->setChecked( true );
  hbox->addWidget( rb2, 2, 1 );

  bg->insert( rb1, 0 );
  bg->insert( rb2, 1 );

  vbox->addStretch( 100 );

  stabctl->insertPage( controllerTab, tr( "Controller" ) );
}

void
ProjectSettings::InitButtons( QWidget *parent )
{
  QBoxLayout *topLayout = new QHBoxLayout( parent, 2 );
  topLayout->addStretch( 10 );

  okButton = new QPushButton( "   OK   ", parent );
  connect( okButton, SIGNAL( clicked() ), 
           this, SLOT( SlotOkButtonClicked() ) );
  okButton->setEnabled( false );
  topLayout->addWidget( okButton );
 
  QPushButton *cancelButton = new QPushButton( "Cancel", parent );
  connect( cancelButton, SIGNAL( clicked() ), 
           this, SLOT( SlotCancelButtonClicked() ) );
  topLayout->addWidget( cancelButton );
}

void
ProjectSettings::SlotOkButtonClicked( void )
{
  accept();
}

void
ProjectSettings::SlotCancelButtonClicked( void )
{
  reject();
}

void
ProjectSettings::SlotDirectoryButtonClicked( void )
{
  QString dirName = QFileDialog::getExistingDirectory();

  if ( !dirName.isEmpty() ) {
    entry[ PROJECT_DIR_E ]->setText( dirName );
  }
}

void
ProjectSettings::SlotHexButtonClicked( void )
{
  QString fileName = QFileDialog::getOpenFileName( QString::null, "*.hex" );

  if ( !fileName.isEmpty() ) {
    entry[ PROJECT_HEX_E ]->setText( fileName );
  }
}

void
ProjectSettings::SlotAsmButtonClicked( void )
{
  QString fileName = QFileDialog::getOpenFileName( QString::null, "*.asm" );

  if ( !fileName.isEmpty() ) {
    entry[ PROJECT_ASM_E ]->setText( fileName );
  }
}

void
ProjectSettings::SlotLstButtonClicked( void )
{
  QString fileName = QFileDialog::getOpenFileName( QString::null, "*.lst" );

  if ( !fileName.isEmpty() ) {
    entry[ PROJECT_LST_E ]->setText( fileName );
  }
}

void
ProjectSettings::SlotSettingsChanged( void )
{
  if ( sender() == entry[ PROJECT_HEX_E ] ||
       sender() == entry[ PROJECT_ASM_E ] ) {

    reload = true;
  }

  if ( entry[ PROJECT_HEX_E ]->text().isEmpty() ||
       entry[ PROJECT_NAME_E ]->text().isEmpty() ||
       entry[ PROJECT_AUTHOR_E ]->text().isEmpty() ||
       entry[ PROJECT_DIR_E ]->text().isEmpty() ) {

    okButton->setEnabled( false );

  } else {
    okButton->setEnabled( true );
  }
}

void
ProjectSettings::SlotProjectNameChanged( const QString &projectName )
{
  if ( isNew ) {
    if ( projectName.isEmpty() ) {
      entry[ PROJECT_DIR_E ]->setText( SS_GET( "projects_dir" ) );
      entry[ PROJECT_HEX_E ]->clear();
      entry[ PROJECT_ASM_E ]->clear();
      entry[ PROJECT_LST_E ]->clear();
    } else {
      QString projectDir( SS_GET( "projects_dir" ) );

      projectDir += "/";
      projectDir += projectName;

      entry[ PROJECT_DIR_E ]->setText( projectDir );
      entry[ PROJECT_HEX_E ]->setText( projectName.lower() + ".hex" );
      entry[ PROJECT_ASM_E ]->setText( projectName.lower() + ".asm" );
      entry[ PROJECT_LST_E ]->setText( projectName.lower() + ".lst" );
    }
  }
}

void
ProjectSettings::SlotProgramMemoryChanged( int id )
{
  if ( id == 0 ) {         // IROM
    cb_mem[1]->setChecked( true );
    cb_mem[3]->setChecked( false );
  } else if ( id == 1 ) {  // XROM
    cb_mem[1]->setChecked( false );
    cb_mem[3]->setChecked( true );
  }
}

