/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef SIMIKON_PROJECT_HH
#define SIMIKON_PROJECT_HH

// QT_INCLUDES
#include <qlist.h>
#include <qarray.h>
#include <qlistview.h>
#include <qdialog.h>
#include <qsplitter.h>

// SIMIKON_INCLUDES
#include "DataStructures.hh"
#include "ErrorData.hh"

class Simikon8052;
class SimikonAS31;
class SimikonRegisters;
class SimikonSFR;
class SimikonMemory;
class SimikonStack;
class SimikonDebugger;
class SimikonPerformance;
class SimikonSerial;
class SimikonInterrupts;
class SimikonIOPorts;
class SimikonModule;
class SimikonConsole;
class SimikonProject;
class SimikonAbout;
class SimikonHelp;

class Registers;
class Memory;
class Stack;
class Disassembler;
class Editor;
class Listing;
class Performance;

class QLineEdit;
class QSpinBox;
class QCheckBox;
class QComboBox;
class QPushButton;
class QRadioButton;
class QTable;
class QLabel;

class STabCtl;

// ENUMERATIONS
enum {
  PROJECT_AUTHOR_E, PROJECT_NAME_E, PROJECT_DIR_E, 
  PROJECT_HEX_E, PROJECT_ASM_E, PROJECT_LST_E,
  NO_OF_ENTRYS
};

enum {
  ACTION_SAVE_E, ACTION_ASSEMBLE_E, ACTION_ASSEMBLE_EDITOR_E, ACTION_START_E,
  ACTION_DEBUGGER_E,
  NO_OF_ACTIONS
};

enum {
  MODULE_REGISTERS_E, MODULE_SFR_E, MODULE_STACK_E, MODULE_MEMORY_E, 
  MODULE_SERIAL_E, MODULE_IOPORTS_E, MODULE_PERFORMANCE_E,
  MODULE_INTERRUPTS_E,
  NO_OF_MODULES
};

/*===========================================================================*/
/* class: Project                                                            */
/*===========================================================================*/

class Project: public QObject, public QListViewItem {
  Q_OBJECT

  friend class SimikonProject;
  friend class Simikon8052;

public:
  Project( QListView *parent, 
           Disassembler *d, Editor *e, Listing *l, Performance *p, 
           const ProjectData &pdata );
  ~Project();

  void reset( bool resetASM = true );

private:
  virtual void activate( void );
  PerformanceData *FindPerformanceData( QListViewItem *i );

public slots:
  void slotAddPerformance( QListViewItem *i, bool dynamic );
  void slotAddAutoPerformance( void );
  void slotRemovePerformance( QListViewItem *i );
  void slotResetPerformance( void );

private slots:
  void SlotSetBreakPoint( uint address, bool state );
  void SlotASMChanged( void );

signals:
  void signalProjectSelected( Project *prj );

private:
  bool actionState[ NO_OF_ACTIONS ];

  ProjectData pData;

  Disassembler           *disassembler;
  Editor                 *editor;
  Listing                *listing;
  Performance            *performance;

  QList<PerformanceData> *performanceData;

  InstructionData *program;
  uint programLength;

  bool hexLoaded;
  bool asmLoaded;

  bool analyzePerformance;
};

/*===========================================================================*/
/* class: ProjectSettings                                                    */
/*===========================================================================*/

class ProjectSettings: public QDialog {
  Q_OBJECT

  friend class SimikonProject;

public:
  ProjectSettings( QWidget *parent, const char *name );
  ~ProjectSettings();

  ProjectData getProjectData( void );
  void setProjectData( const ProjectData &pdata );

private:
  void InitGeneralTab( STabCtl *stabctl );
  void InitControllerTab( STabCtl *stabctl );
  void InitButtons( QWidget *parent );

private slots:
  void SlotOkButtonClicked( void );
  void SlotCancelButtonClicked( void );
  void SlotDirectoryButtonClicked( void );

  void SlotHexButtonClicked( void );
  void SlotAsmButtonClicked( void );
  void SlotLstButtonClicked( void );

  void SlotSettingsChanged( void );
  void SlotProjectNameChanged( const QString &projectName );
  void SlotProgramMemoryChanged( int );

private:
  QLineEdit   *entry[NO_OF_ENTRYS];
  QSpinBox    *clock;
  QTable      *table;

  QPushButton  *okButton;
  QCheckBox    *cb_mem[4];
  QRadioButton *rb1, *rb2;
  QCheckBox    *cbP0, *cbP1, *cbP2, *cbP3;
  QComboBox    *library, *model;

  bool isNew;
  bool reload;
};

/*===========================================================================*/
/* class: SimikonProject                                                     */
/*===========================================================================*/

class SimikonProject: public QSplitter {
  Q_OBJECT

  friend class SimikonApp;

public:
  SimikonProject( QWidget *parent, const char *name );
  ~SimikonProject();

  void closeAll( void );
  void openProject( const QString &prjFile );

private:
  void InitProjectTab( void );
  void InitConsoleTab( void );

  bool CreateProject( const ProjectData &pdata );
  void UninstallProject( Project *p );

  void InstallActiveProject( void );
  void UninstallActiveProject( void );

  void SetActiveProject( Project *prj );
  void SetActiveDisassembler( Project *prj );
  void SetActiveEditor( Project *prj );
  void SetActiveListing( Project *prj );
  void SetActivePerformance( Project *prj );
  bool SaveProject( const ProjectData &pdata );

  void ConnectAllModules( void );
  void DisconnectAllModules( void );

  ProjectData *FindProject( const QString &name );

  void Refresh( void );
  void RefreshActions( void );
  void ReloadActiveProject( void );
  
  void RecreateLatestProjectsList( const ProjectData &pdata );

public slots:
  void slotNewProject( void );
  void slotOpenProject( void );
  void slotSaveProject( void );
  void slotCloseProject( void );
  void slotAssembleProject( void );
  void slotProjectSettings( void );
  void slotToggleRegisters( void );
  void slotToggleSFR( void );
  void slotToggleMemory( void );
  void slotToggleStack( void );
  void slotTogglePerformance( void );
  void slotToggleIOPorts( void );
  void slotToggleSerial( void );
  void slotToggleInterrupts( void );
  void slotToggleDebugger( void );

  void slotToggleConsole( void );
  void slotToggleProjects( void );
  
  void slotAboutSimikon( void );
  void slotInstructionSetHelp( void );

  void slotStartSimulator( void );
  void slotStop( void );


  void slotOpenProject0( void );
  void slotOpenProject1( void );
  void slotOpenProject2( void );
  void slotOpenProject3( void );
  void slotOpenProject4( void );
  void slotOpenProject5( void );
  void slotOpenProject6( void );
  void slotOpenProject7( void );
  void slotOpenProject8( void );
  void slotOpenProject9( void );
  
private slots:
  void SlotChangeActiveProject( Project * );
  void SlotStartDebug( void );
  void SlotModuleClosed( void );

  void SlotAssemblerProjectAssembled( void );
  void SlotAssemblerSyntaxError( void );

  void SlotPageSelected( QWidget * );

signals:
  void signalStatusMessage( const char * );

private:
  STabCtl            *stabctl;

  Project            *activeProject;
  QListView          *project;

  SimikonRegisters   *simikonRegisters;
  SimikonSFR         *simikonSFR;
  SimikonMemory      *simikonMemory;
  SimikonStack       *simikonStack;
  SimikonDebugger    *simikonDebugger;
  SimikonPerformance *simikonPerformance;
  SimikonSerial      *simikonSerial;
  SimikonInterrupts  *simikonInterrupts;
  SimikonIOPorts     *simikonIOPorts;
  SimikonAbout       *simikonAbout;
  SimikonHelp        *simikonHelp;

  Simikon8052        *controller;
  SimikonAS31        *assembler;
  
  ProjectSettings    *projectSettings;

  ModuleData         module[ NO_OF_MODULES ];

  bool assembling;
  bool mode; // true: debugger
             // false: simulator
};

#endif // SIMIKON_PROJECT_HH
