/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "config.h"
#include "SimikonProperties.hh"
#include "SimikonUtils.hh"
#include "DataStructures.hh"
#include "LeakTracker.hh"

// WIDGET_INCLUDES
#include "stabctl.h"

// QT_INCLUDES
#include <qlayout.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qlabel.h>
#include <qfiledialog.h>
#include <qdir.h>
#include <qstringlist.h>
#include <qframe.h>

// MISC_INCLUDES
#include <iostream.h>

// PIXMAPS
#include "locale.xpm"
#include "folder.xpm"
//#include "file.xpm"

enum {
  ENTRY_LOCALE_E,
  ENTRY_PROJECTS_DIR_E,
  ENTRY_HELP_DIR_E
};

static const char *entry_keys[] = {
  "locale",
  "projects_dir",
  "help_dir"
};

static const char *default_entry_values[] = {
  DEFAULT_LOCALE,
  DEFAULT_PROJECTS_DIR,
  DEFAULT_HELP_DIR
};

enum {
  CHECK_SPLASH_E,
  CHECK_PROJECT_E,
  CHECK_REMEMBER_WINDOWS_E,
  CHECK_REMEMBER_MODULES_E,
  CHECK_REMEMBER_PROJECTS_E,
  CHECK_LOCALE_E,
  CHECK_GENERATE_LISTING_E
};

static const char *check_keys[] = {
  "show_splash",
  "open_last_project",
  "remember_widget_states",
  "remember_module_state",
  "remember_latest_projects",
  "use_locale",
  "generate_listing"
};

static const char *default_check_values[] = {
  "true",
  "false",
  "false",
  "true",
  "true",
  "false",
  "true"
};

enum {
  SPIN_REMEMBER_PROJECTS_E
};

static const char *spin_keys[] = {
  "remember_latest_projects_no"
};

static const char *default_spin_values[] = {
  "5"
};

enum {
  COMBO_LIBRARY_E,
  COMBO_MODEL_E
};

enum {
  WIDGET_STATUSBAR_E,
  WIDGET_TOOLBAR_E,
  WIDGET_REGISTERS_E,
  WIDGET_DISSEMBLER_E,
  WIDGET_MEMORY_E
};

static const char *widget_keys[] = {
  "show_statusbar",
  "show_toolbar",
  "show_console",
  "show_project"
};

static const char *default_widget_values[] = {
  "true",
  "true",
  "true",
  "true"
};

static const char *latest_project_keys[] = {
  "Project0", "Project1", "Project2", "Project3", "Project4", 
  "Project5", "Project6", "Project7", "Project8", "Project9",
  "last_project"
};

static const char *latest_project_values[] = {
  "", "", "", "", "", "", "", "", "", "", ""
};

static const char *int_keys[] = {
  "window_width",
  "window_height"
};

static const int default_int_values[] = {
  490,
  520
};

#define NO_OF_ENTRY_SETTINGS sizeof( entry_keys ) / sizeof( entry_keys[0] )
#define NO_OF_CHECK_SETTINGS sizeof( check_keys ) / sizeof( check_keys[0] )
#define NO_OF_WIDGET_SETTINGS sizeof( widget_keys ) / sizeof( widget_keys[0] )
#define NO_OF_SPIN_SETTINGS sizeof( spin_keys ) / sizeof( spin_keys[0] )
#define NO_OF_INT_SETTINGS sizeof( int_keys ) / sizeof( int_keys[0] )


static QLineEdit *entry[ NO_OF_ENTRY_SETTINGS ];
static QCheckBox *check[ NO_OF_CHECK_SETTINGS ];
static QSpinBox  *spin[ NO_OF_SPIN_SETTINGS ];

extern SimikonSettings *simikonSettings;

/*===========================================================================*/
/* class: SimikonProperties                                                  */
/*===========================================================================*/

SimikonProperties::SimikonProperties( QWidget *parent, const char *name )
  : QDialog( parent, name, true )
{
  setCaption( "Simikon Properties" );

  SS_PUT_DEFAULTS( entry_keys, default_entry_values, NO_OF_ENTRY_SETTINGS );
  SS_PUT_DEFAULTS( check_keys, default_check_values, NO_OF_CHECK_SETTINGS );
  SS_PUT_DEFAULTS( spin_keys, default_spin_values, NO_OF_SPIN_SETTINGS );
  SS_PUT_DEFAULTS( widget_keys, default_widget_values, NO_OF_WIDGET_SETTINGS );
  SS_PUT_DEFAULTS( latest_project_keys, latest_project_values, 10 );
  SS_PUT_DEFAULTS( int_keys, default_int_values, NO_OF_INT_SETTINGS );

  QString tmp1 = SS_GET( "projects_dir" );
  QString tmp2 = fullFilePath( tmp1, userHome() );

  if ( tmp1 != tmp2 )
    SS_PUT( "projects_dir", tmp2.ascii() );


  QBoxLayout *topLayout = new QVBoxLayout( this );

  // STabCtl
  STabCtl *stabctl = new STabCtl( this, "STabCtl" );
  topLayout->addWidget( stabctl, 10 );

  InitGeneralTab( stabctl );
  InitPathsTab( stabctl );
  InitAssemblerTab( stabctl );

  QFrame *frame = new QFrame( this );
  frame->setFrameStyle( QFrame::WinPanel | QFrame::Raised );
  frame->setLineWidth( 2 );
  topLayout->addWidget( frame );

  // Buttons
  QWidget *buttons = new QWidget( frame, "Buttons" );
  QBoxLayout *hboxLayout = new QHBoxLayout( frame, 2 );
  hboxLayout->addWidget( buttons );

  InitButtons( buttons );
  SetProperties();

  setFixedSize( 320, 200 );

}

SimikonProperties::~SimikonProperties()
{

}


void
SimikonProperties::InitGeneralTab( STabCtl *stabctl )
{
  QWidget    *generalTab = new QWidget( stabctl, "PropertiesGeneralTab" );
  QBoxLayout *topLayout  = new QVBoxLayout( generalTab, 2 );

  QBoxLayout *hboxLayout;

  QGridLayout *layout = new QGridLayout( topLayout, 2, 3, 1 );

  // Locale
  check[ CHECK_LOCALE_E ] = new QCheckBox( tr( "Locale: " ), generalTab );
  connect( check[ CHECK_LOCALE_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotLocaleCheckToggled( bool ) ) );
  connect( check[ CHECK_LOCALE_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );

  entry[ ENTRY_LOCALE_E ] = new QLineEdit( generalTab );
  entry[ ENTRY_LOCALE_E ]->setEnabled( false );
  connect( entry[ ENTRY_LOCALE_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );

  localeButton = new QPushButton( generalTab );
  connect( localeButton, SIGNAL( clicked() ), 
           this, SLOT( SlotLocaleButtonClicked() ) );
  localeButton->setPixmap( QPixmap( locale_xpm ) );
  localeButton->setEnabled( false );

  layout->addWidget( check[ CHECK_LOCALE_E ], 0, 0 );
  layout->addWidget( entry[ ENTRY_LOCALE_E ], 0, 1 );
  layout->addWidget( localeButton, 0, 2 );

  // Show Splash Screen
  check[ CHECK_SPLASH_E ] =
    new QCheckBox( tr( "Show Splash Screen" ), generalTab );
  connect( check[ CHECK_SPLASH_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  topLayout->addWidget( check[ CHECK_SPLASH_E ] );

  // Remember view menu setup
  check[ CHECK_REMEMBER_WINDOWS_E ] = 
    new QCheckBox( tr( "Remember View Menu Setup On Exit" ), generalTab );
  connect( check[ CHECK_REMEMBER_WINDOWS_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  topLayout->addWidget( check[ CHECK_REMEMBER_WINDOWS_E ] );

  // Remember active modules
  check[ CHECK_REMEMBER_MODULES_E ] =
    new QCheckBox( tr( "Remember Active Modules" ), generalTab );
  connect( check[ CHECK_REMEMBER_MODULES_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  topLayout->addWidget( check[ CHECK_REMEMBER_MODULES_E ] );

  // Open last project on startup
  check[ CHECK_PROJECT_E ] = 
    new QCheckBox( tr( "Open Last Project On Startup" ), generalTab );
  connect( check[ CHECK_PROJECT_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  topLayout->addWidget( check[ CHECK_PROJECT_E ] );
 
  // Remember last used projects
  hboxLayout = new QHBoxLayout( topLayout );
 
  check[ CHECK_REMEMBER_PROJECTS_E ] = 
    new QCheckBox( tr( "Remember Latest Opened Projects" ), generalTab );
  connect( check[ CHECK_REMEMBER_PROJECTS_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  connect( check[ CHECK_REMEMBER_PROJECTS_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotRemeberProjectsCheckToggled( bool ) ) );
  hboxLayout->addWidget( check[ CHECK_REMEMBER_PROJECTS_E ] );

  spin[ SPIN_REMEMBER_PROJECTS_E ] =
    new QSpinBox( 1, 10, 1, generalTab );
  connect( spin[ SPIN_REMEMBER_PROJECTS_E ], SIGNAL( valueChanged( int ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  spin[ SPIN_REMEMBER_PROJECTS_E ]->setEnabled( false );
  hboxLayout->addWidget( spin[ SPIN_REMEMBER_PROJECTS_E ] );

  stabctl->insertPage( generalTab, tr( "General" ) );
}

void
SimikonProperties::InitPathsTab( STabCtl *stabctl )
{
  QWidget     *pathsTab  = new QWidget( stabctl, "PropertiesPathsTab" );
  QBoxLayout  *topLayout = new QVBoxLayout( pathsTab, 2 );
  QGridLayout *layout    = new QGridLayout( topLayout, 4, 3, 1 );

  QLabel      *label;
  QPushButton *button;

  // Project dir
  entry[ ENTRY_PROJECTS_DIR_E ] = new QLineEdit( pathsTab );
  connect( entry[ ENTRY_PROJECTS_DIR_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  label = new QLabel( entry[ ENTRY_PROJECTS_DIR_E ], tr( "&Projects:  " ), 
                      pathsTab );
  button = new QPushButton( pathsTab );
  connect( button, SIGNAL( clicked() ), 
           this, SLOT( SlotProjectButtonClicked() ) );
  button->setPixmap( QPixmap( folder_xpm ) );

  layout->addWidget( label, 0, 0 );
  layout->addWidget( entry[ ENTRY_PROJECTS_DIR_E ], 0, 1 );
  layout->addWidget( button, 0, 2 );

  // Help dir
  entry[ ENTRY_HELP_DIR_E ] = new QLineEdit( pathsTab );
  connect( entry[ ENTRY_HELP_DIR_E ], SIGNAL( textChanged( const QString & ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  label = new QLabel( entry[ ENTRY_HELP_DIR_E ], tr( "&Help: " ),
                      pathsTab );
  button = new QPushButton( pathsTab );
  connect( button, SIGNAL( clicked() ),
           this, SLOT( SlotHelpButtonClicked() ) );
  button->setPixmap( QPixmap( folder_xpm ) );

  layout->addWidget( label, 1, 0 );
  layout->addWidget( entry[ ENTRY_HELP_DIR_E ], 1, 1 );
  layout->addWidget( button, 1, 2 );

  stabctl->insertPage( pathsTab, tr( "Paths" ) );
}

void
SimikonProperties::InitAssemblerTab( STabCtl *stabctl )
{
  QWidget     *assemblerTab = new QWidget( stabctl, "PropertiesAssemblerTab" );
  QBoxLayout  *topLayout    = new QVBoxLayout( assemblerTab, 2 );

  // Generate Listing file
  check[ CHECK_GENERATE_LISTING_E ] =
    new QCheckBox( tr( "Generate Assembler Listing File" ), assemblerTab );
  connect( check[ CHECK_GENERATE_LISTING_E ], SIGNAL( toggled( bool ) ),
           this, SLOT( SlotPropertiesChanged( void ) ) );
  topLayout->addWidget( check[ CHECK_GENERATE_LISTING_E ] );
  
  // Start address

  // stratch space
  topLayout->addStretch( 1 );

  stabctl->insertPage( assemblerTab, tr( "Assembler" ) );
}

void
SimikonProperties::InitButtons( QWidget *parent )
{
  QBoxLayout *buttonsLayout = new QHBoxLayout( parent, 2 );

  buttonsLayout->addStretch( 10 );

  okButton = new QPushButton( tr( "   OK   " ), parent );
  connect( okButton, SIGNAL( clicked() ), 
           this, SLOT( SlotOkButtonClicked() ) );
  buttonsLayout->addWidget( okButton );

  QPushButton *cancelButton = new QPushButton( tr( "Cancel" ), parent );
  connect( cancelButton, SIGNAL( clicked() ), 
           this, SLOT( SlotCancelButtonClicked() ) );
  buttonsLayout->addWidget( cancelButton );
}

void
SimikonProperties::SetProperties( void )
{
  for ( uint i = 0; i < NO_OF_ENTRY_SETTINGS; i++ )
    entry[i]->setText( SS_GET( entry_keys[i] ) );

  for ( uint i = 0; i < NO_OF_CHECK_SETTINGS; i++ )
    check[i]->setChecked( SS_GET_BOOL( check_keys[i] ) );

  for ( uint i = 0; i < NO_OF_SPIN_SETTINGS; i++ )
    spin[i]->setValue( SS_GET_INT( spin_keys[i] ) );
}

void
SimikonProperties::ApplyProperties( void )
{
  for ( uint i = 0; i < NO_OF_ENTRY_SETTINGS; i++ )
    SS_PUT( entry_keys[i], entry[i]->text().ascii() );

  for ( uint i = 0; i < NO_OF_CHECK_SETTINGS; i++ )
    SS_PUT( check_keys[i], check[i]->isChecked() );

  for ( uint i = 0; i < NO_OF_SPIN_SETTINGS; i++ )
    SS_PUT( spin_keys[i], spin[i]->value() );
}

void
SimikonProperties::SlotOkButtonClicked( void )
{
  ApplyProperties();
  accept();
}

void
SimikonProperties::SlotCancelButtonClicked( void )
{
  SetProperties();
  reject();
}

void
SimikonProperties::SlotLocaleButtonClicked( void )
{
  QString localePath = QFileDialog::getOpenFileName( QString::null, "*.qm" );

  if ( !localePath.isEmpty() ) {
    entry[ ENTRY_LOCALE_E ]->setText( localePath );
  }
}

void
SimikonProperties::SlotHelpButtonClicked( void )
{
  QString helpDir = QFileDialog::getExistingDirectory();

  if ( !helpDir.isEmpty() ) {
    entry[ ENTRY_HELP_DIR_E ]->setText( helpDir );
  }
}

void
SimikonProperties::SlotLocaleCheckToggled( bool state )
{
  entry[ ENTRY_LOCALE_E ]->setEnabled( state );
  localeButton->setEnabled( state );
}

void
SimikonProperties::SlotRemeberProjectsCheckToggled( bool state )
{
  spin[ SPIN_REMEMBER_PROJECTS_E ]->setEnabled( state );
}

void
SimikonProperties::SlotProjectButtonClicked( void )
{
  QString dirPath = QFileDialog::getExistingDirectory();

  if ( !dirPath.isEmpty() ) {
    entry[ ENTRY_PROJECTS_DIR_E ]->setText( dirPath );
  }
}

void
SimikonProperties::SlotPropertiesChanged( void )
{
  if ( ( check[ CHECK_LOCALE_E ]->isChecked() && 
         entry[ ENTRY_LOCALE_E ]->text().isEmpty() ) ||
         entry[ ENTRY_PROJECTS_DIR_E ]->text().isEmpty() ) {

    okButton->setEnabled( false );
  } else {
    okButton->setEnabled( true );
  }
}
