/*
 * Copyright (C) 1999, 2000 Alen Kovac <akovac@jagor.srce.hr>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity 
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *   
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// SIMIKON_INCLUDES
#include "config.h"
#include "SimikonSettings.hh"
#include "SimikonUtils.hh"
#include "LeakTracker.hh"

// MISC_INCLUDES
#include <iostream.h>

// QT_INCLUDES
#include <qfile.h>
#include <qtextstream.h>

/*===========================================================================*/
/* class: SimikonSettings                                                    */
/*===========================================================================*/

SimikonSettings::SimikonSettings()
{
  settings = new QDict<QString>;
  settings->setAutoDelete( true );

  settingsFile += userHome();
  settingsFile += "/";
  settingsFile += SIMIKON_SETTINGS_FILE;

  isLoaded = false;

  SimikonSettings::ParseSettings();

  isDirty = false;
}

SimikonSettings::~SimikonSettings()
{
  if ( isDirty ) {
    SimikonSettings::SaveSettings();
  }

  delete settings;
}


void
SimikonSettings::putDefaults( const char **key, const char **data, uint n )
{
  if ( !isLoaded )
    for ( uint i = 0; i < n; i++ )
      putEntry( key[i], data[i] );
}

void
SimikonSettings::putDefaults( const char **key, const int data[], uint n )
{
  if ( !isLoaded )
    for ( uint i = 0; i < n; i++ )
      putEntry( key[i], data[i] );
}

void
SimikonSettings::putEntry( const char *key, const char *value )
{
  settings->replace( QString( key ), new QString( value ) );
	
  emit signalEntryChanged( key, value );

  isDirty = true;
}

void
SimikonSettings::putEntry( const char *key, bool value )
{
  ( value ) ? putEntry( key, "true" ) : putEntry( key, "false" );
}

void
SimikonSettings::putEntry( const char *key, int value )
{
  putEntry( key, QString::number( value ).ascii() );
}

const char *
SimikonSettings::getEntry( const char *key )
{
  QString *data = settings->find( key );

  if ( !data ) {
#ifdef DEBUG
    cout << "WARNING: SimikonSettings::getEntry(): "
         << "Can't find data name: " << key << endl;
#endif
    return 0;
  }
  
  return data->ascii();
}

bool
SimikonSettings::getBoolEntry( const char *key )
{
  return ( !strncmp( getEntry( key ), "true", 4 ) ) ? true : false;
}

int
SimikonSettings::getIntEntry( const char *key )
{
  return QString( getEntry( key ) ).toInt();
}

void
SimikonSettings::ParseSettings( void )
{
  QFile fp( settingsFile );
	
  if ( !fp.open( IO_ReadOnly ) ) {
#ifdef DEBUG
    cout << "WARNING: SimikonSettings::ParseSettings(): "
         << "Can't open file for reading! \n";
#endif
    return;
  }

  QTextStream tfp( &fp );
  QString line;

  while ( !tfp.eof() ) {
    line = tfp.readLine();

    char *keyToken, *valueToken;

    keyToken = strtok( (char *) line.ascii(), "':' ' '" );
    valueToken = strtok( NULL, "' '" );

    putEntry( keyToken, valueToken );
  }

  fp.close();

  isLoaded = true;
}

void
SimikonSettings::SaveSettings( void )
{
  QFile fp( settingsFile );

  if ( !fp.open( IO_WriteOnly ) ) {
    cout << "WARNING: SimikonSettings::SaveSettings(): "
         << "Can't open file for writing! \n";
    return;
  }

  QDictIterator<QString> it( *settings );

  while ( it.current() ) {
    QString tmpSettingsEntry;

    tmpSettingsEntry += it.currentKey();
    tmpSettingsEntry += ": ";
    tmpSettingsEntry += *( QString * ) it.current();
    tmpSettingsEntry += "\n";

    fp.writeBlock( tmpSettingsEntry.ascii(), tmpSettingsEntry.length() );

    ++it;
  };
	
  fp.close();
}
